// opcode.h
// no dependencies

/*
    MWA 13 July 2023
    A deficiency of this file is there is no overflow checking for
    opcode enums. So if a privileged instruction collides with XOOB,
    or a nonprivileged instruction reaches OC_MIN_PRIV, we haven't
    done anything to warn the maintainer. I don't think we're close to
    an event of this kind happening, but nonetheless....
*/

// These bitmasks describe an ALU opcode's signedness.
enum signage {
    left_signed = 4,
    right_signed = 2,
    dest_signed = 1,
};

// Some opcodes have 2, 4, or 8 signedness variants.
// These macros streamline making lists of these opcodes.
// One convention is, no trailing comma (caller supplies that).
#define us2(x) x ## _u, x ## _s
#define us4(x) x ## _uu, x ## _us, x ## _su, x ## _ss
#define us8(x) x ## _uuu, x ## _uus, x ## _usu, x ## _uss, \
               x ## _suu, x ## _sus, x ## _ssu, x ## _sss

// Here are opcodes for instructions that employ the ALU.
enum {
    // 48 additive instructions
    us8(oc_A),      // add
    us8(oc_S),      // subtract
    us8(oc_RS),     // reverse subtract
    us8(oc_AC),     // add with carry
    us8(oc_SC),     // subtract with carry
    us8(oc_RSC),    // reverse subtract with carry

    // 16 bitwise boolean instructions
    oc_IGF,         // ignorant false
    oc_IGT,         // ignorant true
    oc_XL,          // exactly left
    oc_XR,          // exactly right
    oc_NL,          // not left
    oc_NR,          // not right
    oc_AND,         // and
    oc_NAND,        // nand
    oc_OR,          // or
    oc_NOR,         // nor
    oc_XOR,         // xor
    oc_XNOR,        // xnor
    oc_LANR,        // left and not right
    oc_RANL,        // right and not left
    oc_LONR,        // left or not right
    oc_RONL,        // right or not left

    // 16 max & min instructions
    us8(oc_MAX),    // maximum
    us8(oc_MIN),    // minimum

    // Bound can check a signed or unsigned array index.
    us2(oc_BOUND),  // bound (array index validation)

    // Compares specify signedness of left and right operands.
    us4(oc_CMP),    // compare (as in magnitude)

    // The arithmetic shifts indicate signedness of the destination
    // and left operand. The right operand requires special encoding.
    us4(oc_ASL),    // arithmetic shift left
    us4(oc_ASR),    // arithmetic shift right

    // various other ALU instructions
    oc_LSL,         // logical shift left
    oc_LSR,         // logical shift right
    oc_ROL,         // rotate left
    oc_ML,          // multiply low
    oc_MH,          // multiply high
    oc_MHNS,        // multiply high no sign
    oc_MHL,         // split-register MHNS (left) and ML (right)
    oc_MHL0,        // MHL using tribble 0 of L six times
    oc_MHL1,        // MHL using tribble 1 of L six times
    oc_MHL2,        // MHL using tribble 2 of L six times
    oc_MHL3,        // MHL using tribble 3 of L six times
    oc_MHL4,        // MHL using tribble 4 of L six times
    oc_MHL5,        // MHL using tribble 5 of L six times
    oc_DSL,         // double shift left
    oc_SWIZ,        // swizzle
    oc_NUDGE,       // nudge
    oc_CRF,         // clear range flag
    oc_HAM2,        // 2-instruction popcount, second instruction
    oc_TXOR,        // transposing xor
    oc_A_w,         // add with wrap
    oc_AC_w,        // add with carry with wrap
    oc_S_w,         // subtract with wrap
    oc_SC_w,        // subtract with carry with wrap
    oc_RS_w,        // reverse subtract with wrap
    oc_RSC_w,       // reverse subtract with carry with wrap
    oc_PIT,         // permute inside tribbles (gamma layer)
    oc_PAT,         // permute across tribbles (beta layer)
    oc_PAIT,        // permute across, then inside tribbles (same right operand)
    oc_MIX,         // forward substitution-permutation network
    oc_XIM,         // inverse substitution-permutation network
    oc_PVIO,        // paravirtualized I/O
    oc_UNARY_A,     // simple unary, alpha layer
    oc_UNARY_B,     // simple unary, beta  layer
    oc_UNARY_G,     // simple unary, gamma layer

    // The "stacked unary" operations give alpha, beta, and gamma the
    // same right operand. Several variants of T flag input processing,
    // theta, and zeta are needed, which is why so many opcodes here.
    oc_STUN_A,
    oc_STUN_B,
    oc_STUN_C,
    oc_STUN_D,
    oc_STUN_E,
    oc_STUN_F,
    oc_STUN_G,
    oc_STUN_H,
    oc_STUN_I,      // This one is reserved for future use.
    oc_STUN_J,      // also reserved
    oc_STUN_K,      // also reserved
    oc_NO_MORE_ALU, // (< this value are ALU instructions)
};

// opcodes for the remaining (non-ALU) instructions
enum {

    // Opcodes with branch destinations are clustered for assembler's benefit.
    oc_J            // jump
        = oc_NO_MORE_ALU,
    oc_J_lt,        // jump conditionally (10 cases) 
    oc_J_le,
    oc_J_eq,
    oc_J_ne,
    oc_J_ge,
    oc_J_gt,
    oc_J_r,
    oc_J_nr,
    oc_J_t,
    oc_J_nt,
    oc_C,           // call
    oc_MIN_BRANCH_NPRIV = oc_J,
    oc_MAX_BRANCH_NPRIV = oc_C,

    oc_R,           // return
    oc_REVERT,      // revert
    oc_NOP,         // no operation
    oc_NOLIST,      // NOP that doesn't appear in listings
    oc_IMP,         // load immediate positive
    oc_IMN,         // load immediate negative 
    oc_IMH,         // load immediate high 18 bits
    oc_IMB,         // load immediate with high duplicating low
    oc_LD,          // ordinary read from RAM
    oc_ADDLD,       // ordinary read from RAM via ADD (with constraints)
    oc_STO,         // ordinary write to  RAM
    oc_STO2,        // double ordinary write to RAM transposed
    oc_LDSTO,       // ordinary transposed reg/memory swap
    oc_YIELD,       // yield via hardware (dc.d130)

    // Opcodes for internal use by the assember.
    // They don't appear in finished object code.
    oc_asm_addr,    // IMP with a label or address.

    // The numbering gap here will be implemented as oc_NOP.

    // privileged instructions
    oc_MIN_PRIV     // reserving 64 highest opcodes as privileged
        = 448,

    // Opcodes with branch destinations are clustered for assembler's benefit.
    oc_CALI =       // call stack initialize
        oc_MIN_PRIV,
    oc_PCALL,       // privileged call + priv (yield to superuser)
    oc_NPCALL,      // privileged call + npriv (yield to user)
    oc_MIN_BRANCH_PRIV = oc_CALI,
    oc_MAX_BRANCH_PRIV = oc_NPCALL,

    oc_HALT,        // halt
    oc_WCM,         // privileged write code memory
    oc_RCM1,        // privileged read code memory
    oc_RCM2,        // (two unseparated instructions, may not context switch)
    oc_WDM,         // privileged write data memory (bypass page table)
    oc_WDM2,        // privileged double wriite data memory (bypass page table)
    oc_RDM,         // privileged read data memory (bypass page table)
    oc_ADDRDM,      // privileged read data memory via ADD (bypass page table)
    oc_RWDM,        // privileged trans. reg/memory swap (bypass page table)
    oc_WPT,         // privileged write page table
    oc_RPT,         // privileged read page table
    oc_XANY,        // privileged execute any instruction via reg.
    oc_JANY,        // privileged jump to any location via reg.
    oc_USER,        // privileged set user id
    oc_NPRIV,       // privileged enter user mode
    oc_PRIV,        // privileged enter privileged mode
    oc_SETUP,       // privileged enter neither mode (page table & stack setup)
    oc_PEEK,        // privileged copy user reg to superuser reg while npriv
    oc_POKE,        // privileged copy superuser reg to user reg while priv
    oc_HIJACK,      // privileged call-no-branch + priv (yield to superuser)
    oc_TIMER,       // privileged timer setpoint serial I/O
    oc_MEMSET,      // privileged memset via simulator

    // The numbering gap here will be implemented as "privileged" oc_NOP.

    oc_XOOB = 511   // privileged executable out-of-band command
};

// These "opcode category groups" show the general operation,
// but don't include signage information. They have the numeric
// value of the lowest opcode in the group they represent.
enum {
    ocat_A = oc_A_uuu,
    ocat_S = oc_S_uuu,
    ocat_RS = oc_RS_uuu,
    ocat_AC = oc_AC_uuu,
    ocat_SC = oc_SC_uuu,
    ocat_RSC = oc_RSC_uuu,
    ocat_MAX = oc_MAX_uuu,
    ocat_MIN = oc_MIN_uuu,
    ocat_BOUND = oc_BOUND_u,
    ocat_CMP = oc_CMP_uu,
    ocat_ASL = oc_ASL_uu,
    ocat_ASR = oc_ASR_uu,
};
