/*
    Cross-assembler API
    MWA 26 May 2023 ff
*/

// Cleanup everything such that the assembler could restart if desired.
// 'as_trace' is left as it was for diagnostics.
// Always succeeds.
void as_dealloc(void)
{
    ascoll_dealloc();
    aslex_lnum = 0;
}

// Read source code from an open stream and assemble it.
// Default listing options are used.
// Returns 0 for success.
int as_assemble_from_stream(FILE *f, char *line_prompt)
{
    char buf[16384];
    int lex_errors = 0;
    int parens = 0;
    int maybe_ok = 1;                   // not ok yet

    if (sizeof(long) != 8 || sizeof(void *) != 8)
        asfail("object sizes not as expected");

    if (workspace) {
        gpa("xa: need as_dealloc() before assembling again");
        return 1;
    }

    fold_const(0, 0);                   // claim register 0 for zero constant
    consts->reg_no = 0;                 // disavow "immediate only" status

    // The assembler first emits an instruction to call the glossary. It's
    // a NOP here instead of a call, because we don't want resolve_branches()
    // to touch it. We'll know the address to call later.
    push_code((ulong) oc_NOP << 27);

    // Insert main:: scope on line zero, and advance line number to 1.
    aslex("main::\n", &parens);

    // read-print-eval loop
    while (1) {
        if (line_prompt) {
            printf("%s", line_prompt);
            fflush(stdout);
        }
        if (!fgets(buf, sizeof(buf), f))
            break;
        lex_errors |= aslex(buf, &parens);
    }
    if (line_prompt)
        printf("\n");       // Get to beginning of line after EOF

    // Guarantee that the last statement has an end-of-line to simplify parsing.
    aslex("\n", &parens);

    // Mark the end of the input to protect the assembler.
    push_eof_token();

    if (parens) {
        gpa("%eend of file within multiline comment\n", aslex_lnum);
        ++lex_errors;
    }

    do {
        if (lex_errors)
            break;
        if (assyn(toks, pass1))
            break;
        if (betwixt())
            break;
        if (assyn(toks, pass2))
            break;
        resolve_branches();
        gen_glossary();
        maybe_ok = 0;                   // ok now
    } while (0);

    if (maybe_ok)
        gpa("\n%eassembly failed due to above error(s)\n", aslex_lnum);
    else {
        list();
        list2();
        list2b();
    }

    return maybe_ok;
}

// Read source code from a closed file and assemble it.
// Returns 0 for success.
int as_assemble_from_filename(char *fname)
{
    FILE *f = fopen(fname, "r");
    int maybe_ok = 0;

    if (!f) {
        gpa("xa: open of file '%s' failed\n", fname);
        return 1;
    }

    maybe_ok = as_assemble_from_stream(f, NULL);
    fclose(f);

    return maybe_ok;
}

// Copy object code produced by the assembler to an array for simulation.
// Returns 0 for success.
int as_output_program_code(ulong *write_here, int max_words)
{
    Code *c;

    ifnot (c = code) {
        gpa("xa: no code to output\n");
        return 1;
    }

    do {
        c = c -> next;
        if (max_words <= 0) {
            gpa("xa: code output was truncated to fit\n");
            return 1;
        }
        *write_here++ = c->instr;
        --max_words;
    } while (c != code);

    return 0;                           // success
}


static void out_word(FILE *f, long word)
{
    char buf[5];

    buf[0] = word >> 32 &  15;
    buf[1] = word >> 24 & 255;
    buf[2] = word >> 16 & 255;
    buf[3] = word >>  8 & 255;
    buf[4] = word       & 255;

    fwrite(buf, 1, 5, f);
}

// Write object code to executable file.
// Returns 0 for success.
int as_write_executable_file(char *outfile)
{
    #define as_exe_d36 68719268038L     // ...d36`t
    #define as_exe_len 68719301527L     // ...len`t
    #define as_exe_boo 68719261208L     // ...boo`t
    #define as_exe_cod 68719265293L     // ...cod`t
    #define as_exe_eof 68719273487L     // ...eof`t

    Code *c;
    FILE *f = fopen(outfile, "w");

    ifnot (c = code) {
        gpa("xa: no code to output\n");
        return 1;
    }

    if (!f) {
        gpa("xa: cannot open %s\n", outfile);
        return 1;
    }

    // Program header. See "Executable format for Dauug|36 binaries" in doc.
    out_word(f, as_exe_d36);
    out_word(f, n_code);
    out_word(f, as_exe_len);
    if (boot_code_offset) {
        out_word(f, boot_code_offset);
        out_word(f, as_exe_boo);
    }
    else
        out_word(f, as_exe_cod);

    do {
        c = c -> next;
        out_word(f, c->instr);
    } while (c != code);

    // End of program. See "Executable format for Dauug|36 binaries" in doc.
    out_word(f, as_exe_eof);

    if (ferror(f) || fclose(f)) {
        gpa("xa: error writing to %s\n", outfile);
        return 1;
    }

    return 0;                           // success

    #undef as_exe_d36
    #undef as_exe_len
    #undef as_exe_boo
    #undef as_exe_cod
    #undef as_exe_eof
}

// Call to assemble a program to start at a code location other than 0.
void as_set_boot_code_offset(int bco)
{
    boot_code_offset = bco;
}

// Call to assemble a program to start at a code location other than 0.
void as_set_code_page_size(int s)
{
    page_size_mask = s ? s - 1 : 0;
}

// Call to change verbosity.
void as_set_verbosity(int v)
{
    verbose = v;
}
