/*
    Cross-assembler's collections
    MWA 11 May 2023 ff
*/

enum {
    as_default_workspace_size = 5 * 65536,  // bytes
};

enum TokenTag {
    as_const,
    as_lnum,
    as_ident,
    as_sym,
    as_eof,
};

enum RegFlags {
    as_unsigned = 1,
    as_signed = 2,
    as_kept = 4,
};

enum ConstFlags {
    as_immed_only = 1024,
};

/*
    To the degree possible, assembler data is stored as circularly-linked
    association lists. This is to reduce its source and object code footprint.
    The lists are accessed via pointers to their LAST element.
*/
typedef struct Assoc {
    ulong key1;                 // key1 & key2 can be used for searching
    struct Assoc *next;
    ulong key2;
} Assoc;

/*
    Here are specializations on Assoc.
*/
typedef struct Token {
    ulong tag;                  // enum Token, but note ulong
    struct Token *next;
    ulong payload;
} Token;

typedef struct Const {
    ulong number;
    struct Const *next;
    ulong minus;                // secondary search key: 'number' is signed -ve
    ulong reg_no;               // will be as_immed_only until marked in-use
} Const;

typedef struct Str {
    ulong len;
    struct Str *next;
    ulong byte[0];              // convention: interprets 'char' as unsigned
} Str;

typedef struct Scope {
    Str *name;
    struct Scope *next;
    ulong lnum;
    ulong code_addr;
} Scope;

typedef struct Reg {
    Str *name;
    struct Reg *next;
    Scope *scope;               // secondary search key
    ulong flags;                // enum RegFlags
    ulong reg_no;               // 0 - 511
} Reg;

typedef struct Label {
    Str *name;
    struct Label *next;
    Scope *scope;               // secondary search key
    ulong lnum;
    ulong code_addr;
} Label;

typedef struct Keep {
    Str *name;
    struct Keep *next;
    Scope *scope;               // secondary search key
    ulong lnum;                 // for error reporting
    long force_reg;             // -1 (none) or forced register number
} Keep;

typedef struct Code {
    ulong instr;
    struct Code *next;
} Code;

static Token *toks;             // list of tokens to parse
static Const *consts;           // list of numeric constants that exist
static Str   *strs;             // list of identifier & label names
static Reg   *regs;             // list of registers that are defined
static Label *labels;           // list of labels that are defined
static Scope *scopes;           // list of scopes that are defined
static Keep  *keeps;            // names of registers that are kept
static Code *code;              // final assembled code

static char *workspace;         // all memory allocation goes here
static char *next_workspace;
static char *end_workspace;

static int n_code;              // used for branch resolution
static int blank_saved;         // fully managed herein; no need to cleanup
static int workspace_size       // how much workspace to allocate
    = as_default_workspace_size;
static int page_size_mask;      // (OS page size for user programs) - 1

static int mem_histogram = 0;   // indicate how and why memory is consumed

// This offset is added to all branch destinations throughout the assembler.
// If nonzero, the executable file will use a ...boo`t command instead of a
// ...cod`t command to place the code at this offset.
static int boot_code_offset;

// This adds tokenization and strings to the listing if set.
static int verbose;

static void *getmem(size_t size)
{
    void *v;

    if (!workspace) {
        workspace = next_workspace = calloc(1, workspace_size);
        end_workspace = workspace + workspace_size;
    }

    if (next_workspace + size > end_workspace)
        asfail("out of memory");

    v = next_workspace;
    next_workspace += size;
    return v;
}

/*
    Here is a macro to allocate memory for a pointer to fixed-size object.
    It's not okay for Str, but fine for Token, Const, Scope, Reg, Label.

    Usage:

        Label *p;               // p is uninitialized.
        appear(p);              // p now points to zeroed new storage.
*/
static void *appear_(size_t size, char *purpose)
{
    if (mem_histogram)
        printf("appear %s %zi\n", purpose, size);
    return getmem(size);
}

#define appear(p, purpose) p = appear_(sizeof(*p), purpose)

/*
    Find an item in one of the association lists.
    There are functions here for one or two keys.
    Many keys have to be cast from pointer to ulong.
*/
static void *as_search(void *list, ulong key1)
{
    Assoc *a = list;

    if (!a)
        return a;

    do {
        a = a->next;
        if (a->key1 == key1)
            return a;
    } while (a != list);

    return NULL;
}

static void *as_search2(void *list, ulong key1, ulong key2)
{
    Assoc *a = list;

    if (!a)
        return a;

    do {
        a = a->next;
        if (a->key1 == key1 && a->key2 == key2)
            return a;
    } while (a != list);

    return NULL;
}

/*
    Unconditionally add something to an association list. Bring your own memory.
    Ordinarily, you'll want to have confirmed it wasn't there first.

    Usage:

        Token *toks = NULL;

        while (at_will) {
            Token *t = [alloc and fill in the next token];
            as_append(&toks, t);
        }

    WARNING: 'listv' is a void * to C, but it's the address of a pointer.
*/
static void as_append(void *listv, void *elv)
{
    Assoc **list = listv;
    Assoc *el = elv;

    if (!*list) {
        *list = el->next = el;
        return;
    }
    el->next = (*list)->next;
    *list = (*list)->next = el;
}

/*
    Add a folded constant to 'consts'.
    'number' must be in 36-bit unsigned range.
    'minus' must be 0 or 1.
*/
static Const *fold_const(ulong number, int minus)
{
    Const *a;

    if (number & ~0777777777777)
        asfail("fold_const() out of range");

    if (a = as_search2(consts, number, minus))
        return a;

    appear(a, "Const");                 // not found: alloc, fill, append
    a->number = number;
    a->minus = minus;
    a->reg_no = as_immed_only;
    as_append(&consts, a);
    return a;
}

/*
    Compare the entire Str a to a char slice [b, b_).
    If you have an ASCIZ string instead of a slice, pass NULL as b_.
    Returns 1 iff an exact match, otherwise 0.
*/
static int same_str(Str *a, char *b, char *b_)
{
    ulong *c;

    if (!b_)
        b_ = strchr(b, '\0');

    if (a->len != b_ - b)
        return 0;                       // lengths do not match

    for (c = a->byte; b != b_; ++c, ++b)
        if (*c != (unsigned char) *b)
            return 0;                   // characters do not match

    return 1;                           // exact match
}

/*
    Add a folded string to 'strs'. Takes char slice [cs, cs_) as input.
    This process expands unsigned bytes to unsigned longs. It's inefficient,
    but it's pragmatic considering the purpose of this assembler.
*/
static Str *fold_str(char *cs, char *cs_)
{
    Str *s;
    ulong *u;
    size_t len;

    if (!cs_)                           // unlikely, but support for now
        cs_ = strchr(cs, '\0');

    if (s = strs)
        do {
            s = s->next;
            if (same_str(s, cs, cs_))
                return s;               // found
        } while (s != strs);

    // since not found, allocate storage for string of this length
    len = sizeof(*s) + (cs_ - cs) * sizeof(ulong);
    if (mem_histogram)
        printf("appear Str %zi\n", len);
    s = getmem(len);
    if (!s)
        asfail("fold_str() out of memory");

    // fill in length and bytes
    s->len = cs_ - cs;
    for (u = s->byte; cs != cs_; ++u, ++cs)
        *u = (unsigned char) *cs;

    // append to association list and return
    as_append(&strs, s);
    return s;
}

/*
    Write a Str to a stream.
*/
static void say_str(Str *s, FILE *f)
{
    int i;

    for (i=0; i<s->len; i++)
        fputc(s->byte[i], f);
}

/*
    Diagnostic: Print a Token.
*/
static void say_token(Token *t)
{
    Const *c;
    Str *s;

    if (!t)
        printf("$NULL");

    else switch (t->tag) {
        case as_const:
            c = (Const *) t->payload;
            pprint(c->number, c->minus, 0);
            break;

        case as_lnum:
            printf("%lu:", t->payload);
            break;

        case as_ident:
            s = (Str *) t->payload;
            printf("["); 
            say_str(s, stdout);
            printf("]");
            break;

        case as_sym:
            s = (Str *) t->payload;
            say_str(s, stdout);
            break;

        case as_eof:
            printf("$eof");
            break;

        default:
            printf("unknown (%5lu, %4lu)", t->tag, t->payload);
    }
}

// Push a constant on the token list.
static void push_const_token(ulong n, int minus)
{
    Token *t;
    Const *c;

    if (minus)
        n = (1L << 36) - n;

    c = fold_const(n, minus);

    appear(t, "Token");
    t->tag = as_const;
    t->payload = (ulong) c;
    as_append(&toks, t);
}

// Push a line number on the token list.
static void push_lnum_token(long lnum)
{
    Token *t;

    appear(t, "Token");
    t->tag = as_lnum;
    t->payload = lnum;
    as_append(&toks, t);
}

// Push an identifier on the token list.
static void push_ident_token(char *s, char *s_)
{
    Token *t;
    Str *str = fold_str(s, s_);

    appear(t, "Token");
    t->tag = as_ident;
    t->payload = (ulong) str;
    as_append(&toks, t);
}

// Push a symbol on the token list.
// We just pretend it's an identifier and retag it afterward.
static void push_sym_token(char *s, char *s_)
{
    push_ident_token(s, s_);
    toks->tag = as_sym;                 // toks points to *last* element
}

// Push EOF on the token list.
static void push_eof_token(void)
{
    Token *t;

    appear(t, "Token");
    t->tag = as_eof;
    t->payload = 0;
    as_append(&toks, t);
}

// Push an instruction into the code.
static void push_code(ulong instr)
{
    Code *c;

more:
    appear(c, "Code");
    c->instr = instr;
    as_append(&code, c);
    ++n_code;

    // For user programs, the last instruction on every code page
    // needs to be a NOP. Append one if we're at that location.
    if (page_size_mask &&
            (n_code + boot_code_offset & page_size_mask) == page_size_mask) {
        instr = (ulong) oc_NOP << 27;
        goto more;
    }
}

/*
    This is a General Print function for the Assembler. It can print:

    \n  newline (possibly with un-red escape sequence)

    %i  signed long (cast to pointer)
    %u  unsigned long (cast to pointer)
    %s  Asciz
    %:  char slice (uses two arguments)
    %$  Str
    %O  Scope
    %t  token
    %e  red line number where error appears (place at beginning)

    Every arg past the format string needs to be four bytes, no matter what.
*/
static void gpa(char *fmt, ...)
{
    int red = 0;
    long l;
    ulong u;
    char *s, *s_;
    Str *str;
    Scope *scope;
    Token *t;
    va_list ap;

    va_start(ap, fmt);

    for (; *fmt; ++fmt) {
        if (*fmt == '\n') {
            fputs(red ? "\x1b[m\n" : "\n", stdout);
            red = 0;
            continue;
        }

        if (*fmt != '%') {
            fputc(*fmt, stdout);
            continue;
        }

        switch (*++fmt) {
            case 'i':
                l = va_arg(ap, long);
                printf("%li", l);
                break;

            case 'u':
                u = va_arg(ap, ulong);
                printf("%lu", u);
                break;

            case 's':
                s = va_arg(ap, char *);
                printf("%s", s);
                break;

            case ':':
                s = va_arg(ap, char *);
                s_ = va_arg(ap, char *);
                if (!s_)
                    s_ = strchr(s, '\0');
                printf("%.*s", (int) (s_ - s), s);
                break;

            case '$':
                str = va_arg(ap, Str *);
                say_str(str, stdout);
                break;

            case 'O':
                scope = va_arg(ap, Scope *);
                say_str(scope->name, stdout);
                break;

            case 't':
                t = va_arg(ap, Token *);
                say_token(t);
                break;

            case 'e':
                red = 1;
                u = va_arg(ap, ulong);
                printf("\x1b[1;31;40mline %lu: ", u);
                break;

            default:
                fputc('?', stdout);
        }
    }

    va_end(ap);
}

// Print a blank line between distinct sections of assembler listing.
static void blank(int level)
{
    if (level == blank_saved)
        return;
    blank_saved = level;
    printf("\n");
}

/*
    Diagnostic: show the token list.
*/
static void list(void)
{
    Scope *scope;
    Str *str;
    Const *con;
    Label *label;
    Reg *reg;
    Token *t;

    blank_saved = 1;

    // Scopes
    scope = scopes;
    if (scope) {
        blank(1);
        printf("Labels and registers by scope:\n");
        do {
            scope = scope->next;
            blank((int) (ulong) scope);     // kludge
            gpa("  %O:: line %i, code %i\n",
                    scope, scope->lnum, scope->code_addr + boot_code_offset);

            // Labels within scope
            label = labels;
            if (label) {
                do {
                    label = label->next;
                    if (label->scope != scope)
                        continue;
                    blank(3);
                    gpa("    %$: line %i, code %i\n", label->name,
                            label->lnum, label->code_addr + boot_code_offset);
                } while (label != labels);
            }

            // Regs within scope
            reg = regs;
            if (reg) {
                do {
                    reg = reg->next;
                    if (reg->scope != scope)
                        continue;
                    blank(4);
                    printf("    %3li  %c%c  ", reg->reg_no, reg->flags & as_kept ?
                            'k' : ' ', reg->flags & as_unsigned ? 'u' :
                            reg->flags & as_signed ? 's' : '?');
                    gpa("%$\n", reg->name);
                } while (reg != regs);
            }
        } while (scope != scopes);
    }

    // Consts
    con = consts;
    if (con) {
        blank(5);
        printf("36-bit constants:\n");
        do {
            con = con->next;
            blank(6);
            if (con->reg_no == as_immed_only)
                printf("  imm  ");
            else
                printf("  %3li  ", con->reg_no);

            pprint(con->number, con->minus, 15);
            printf("\n");
        } while (con != consts);
    }

    if (verbose) {
        // Strings
        str = strs;
        if (str) {
            blank(7);
            printf("Strings:\n");
            do {
                str = str->next;
                blank(8);
                gpa("  %$\n", str);
            } while (str != strs);
        }

        // Tokens
        t = toks;
        if (t) {
            blank(9);
            printf("Tokenization:\n     ");
            do {
                t = t->next;
                blank(10);
                if (t->tag == as_lnum)
                    printf("\n  %3lu", t->payload);
                else
                    gpa(" %t", t);
            } while (t != toks);
            printf("\n");
        }
    }
}

// See if a token is a particular symbol.
static int same_sym(Token *t, char *cs, char *cs_)
{
    if (t->tag != as_sym)
        return 0;
    return same_str((Str *) t->payload, cs, cs_);
}

// See if a token is a particular identifier.
// XXX This could be merged with same_sym; only the tag is different.
static int same_ident(Token *t, char *cs, char *cs_)
{
    if (t->tag != as_ident)
        return 0;
    return same_str((Str *) t->payload, cs, cs_);
}

// conversion of opcode names to opcode numbers
static struct {
    int opc;
    char *name;
} oclist[] = {
    #include "oclist.h"
    -1, NULL,
};

// You only need this if you have a string, because opcode.h is also here.
static int name_to_opcode(char *name, char *name_)
{
    int i, len;

    if (!name_)
        name_ = strchr(name, '\0');
    len = name_ - name;

    for (i=0; oclist[i].name; i++) {
        if (strlen(oclist[i].name) != len)
            continue;
        if (!memcmp(name, oclist[i].name, len))
            return oclist[i].opc;
    }

    return -1;
}

// You only need this if you have a string, because opcode.h is also here.
// The assembler version should merge name_to_opcode() with str_to_opcode().
static int str_to_opcode(Str *s)
{
    int i;

    for (i=0; oclist[i].name; i++) {
        if (same_str(s, oclist[i].name, NULL))
            return oclist[i].opc;
    }

    return -1;
}

static char *opcode_to_name(int opcode)
{
    int i;

    for (i=0; oclist[i].opc >= 0; i++)
        if (opcode == oclist[i].opc)
            return oclist[i].name;

    return "?";
}

// List code after pass 2.
static void list2(void)
{
    Code *c;
    char *name;
    int opc, instr;
    int offset = -1;

    ifnot (c = code)
        return;                         // not generated due to error

    printf("\nCode has %i instructions:\n\n", n_code);

    do {
        c = c->next;
        ++offset;
        opc = c->instr >> 27;
        instr = c->instr & 0777777777;
        name = opcode_to_name(opc);
        printf("%4i  %-7s ", offset + boot_code_offset, name);

        // branch instructions should show full number
        if (opc >= oc_MIN_BRANCH_NPRIV && opc <= oc_MAX_BRANCH_NPRIV ||
                opc >= oc_MIN_BRANCH_PRIV && opc <= oc_MAX_BRANCH_PRIV) {
            printf("%11i\n", instr);
            continue;
        }

        switch (opc) {
            // immediates should show 8 bits & 16 bits
            case oc_IMH:
            case oc_IMP:
            case oc_IMN:
            case oc_IMB:
                printf("%3i ", instr >> 18 & 511);
                pprint(instr & 0777777, 0, 7);
                printf("\n");
            break;

            // format everything else as dest, left, right
            default:
                printf("%3i %3i %3i\n", instr >> 18 & 511,
                        instr >> 9 & 511, instr & 511);
        }
    } while (c != code);

    printf("\nMemory allocated was %zi words.\n",
            (next_workspace - workspace) / 4);
}

// List object code after pass 2. Teaching purpose.
// Likely obsolete after November 2023.
static void list2b(void)
{
    Code *c;

    ifnot (c = code)
        return;                         // not generated due to error

    printf("\nOctal dump of code (without offsets):\n\n");

    do {
        c = c->next;

        printf("%03lo.%03lo.%03lo.%03lo\n", c->instr >> 27 & 511,
                c->instr >> 18 & 511, c->instr >> 9 & 511, c->instr & 511);
    } while (c != code);

    printf("\n");
}

// Cleanup everything such this module allocated and touched.
// Always succeeds.
static void ascoll_dealloc(void)
{
    free(workspace);
    workspace = next_workspace = end_workspace = NULL;
    workspace_size = as_default_workspace_size;

    toks = NULL;
    consts = NULL;
    strs = NULL;
    regs = NULL;
    labels = NULL;
    scopes = NULL;
    keeps = NULL;
    code = NULL;
    n_code = 0;
}
