/*
    Cross-assembler's "glossary" scope generation
    MWA 19 May 2023 ff

    The "glossary" is a scope at the end of a program that initializes
    constant registers. The assembler generates and places it after any
    code that's assembled, and inserts a call to it before any other code
    that's assembled. So the code as a whole looks like:

            call 'glossary  ; The assembler adds this instruction.

            ; Your program goes here.

        'glossary::         ; The assembler adds from here to the end.
            imh ...         ; first nonzero const (complex)
            iml ...
            or ...
            imn ...         ; second nonzero const (simple negative)
                            ; more consts as needed
            iml ...         ; zero const (was used as a temp for complex consts)
            return

    To avoid namespace pollution, the scope 'glossary isn't tabulated by
    name. So you can't access it by mistake. In some instances (operating
    system stuff like registers wiped, new user forked, etc.), your code
    may need to call the glossary scope again yourself. To do this, put
    the scope name you'd like to use right at the end of your code.

    This module doesn't insert the original call 'glossary instruction;

*/

static void gen_glossary(void)
{
    Const *c, *d;
    ulong instr;

    // The NOP at the beginning of the generated code needs to be
    // a CALL to this block. Since we need the present value of
    // n_code for this and we're about to change it, now is the time.
    code->next->instr = (ulong) oc_C << 27 |  n_code + boot_code_offset;

    ifnot (c = consts)
        return;                          // impossible (zero reg from asmain.c)

    // loop over constants
    do {
        c = c->next;
        if (c->reg_no == as_immed_only)
            continue;
        if (!c->number)
            continue;                   // generate 0 constant last

        /*
            A register may appear twice in 'consts' due to signedness.
            Kludge workaround: look at the constants list up to this
            point to deduplicate.
        */
        for (d = consts->next; d != c; d = d->next)
            if (d->reg_no == c->reg_no)
                goto is_dupe;

        // need code generated to load const 'c'
        if (instr = make_imm(c->number)) {
            // easy case
            instr |= c->reg_no << 18;
            push_code(instr);
            continue;
        }

        // hard case
        if (c->number > 0777777777777L)
            asfail("> 36 bits in gen_glossary");
        instr = (ulong) oc_IMH << 27 | c->number >> 18;       // reg. 0 as temp
        push_code(instr);
        instr = (ulong) oc_IMP << 27 | c->reg_no << 18 | c->number & 0777777;
        push_code(instr);
        instr = (ulong) oc_OR << 27 | c->reg_no << 18 | c->reg_no << 9;
        push_code(instr);

is_dupe:;
    } while (c != consts);

    // register 0 as constant 0, then return
    push_code((ulong) oc_IMP << 27);
    push_code((ulong) oc_R << 27);
}
