/*
    Cross-assembler's LEXer
    MWA 8 May 2023 ff

    This module tries to keep as much complexity as possible in the lexen[]
    array of strings, so that when the lexer is ported to other languages,
    that complexity requires at most a format change as opposed to any kind
    of logic rewritten.

    lexen[] is interpreted by a small amount of super-simple code. The
    benefit to this is that when the assembler moves to self-hosting, only a
    tiny amount of assembly language will need to be written and audited
    against its own binary object code.

    There is a performance price to all this, but once a bootstrapped
    assembler is available, performance is easy to improve. The goal,
    however, is to limit us to just three assemblers that need maintained
    and security audits:

    (1) the cross assembler (currently written in C), tiniest case
    (2) self-hosted assembler, identical to cross assembler, binary audited
    (3) most recent self-hosted assembler, must assemble with (1) and (2)
*/

/*
            whitespace is ignored
    !       issue CS as error messasge in the event this rule does not match
    (       advance parser past series of (s
    )       advance parser past same-length series of )s
    +       no constant ~ (minus) seen
    -       constant ~ (minus) seen
    :       change state to that named in CS
    :Ohio   on line by itself, introduces rules for state called Ohio
    =       advance parser if input matches CS exactly
    ?       advance parser 1 character if input matches any char in CS
    [abc]   set constant sequence (CS) of characters
    ]       advance parser past right bracket ] if present
    {       mark beginning of string slice
    }       mark end of string slice
    b       advance parser past any byte other than \0
    c       emit token as numeric constant from number scan
    e       peek for end of input buffer: continue rule if \0 appears here
    f       fail (rule does not match)
    i       emit token as identifier from slice
    l       emit token as incremented line number
    L       increment line number, but do not emit a token for it
    m       remember this place (only one deep)
    M       return to remembered place (whether or not rule matches)
    n       scan CS to number in base 10, ignoring any _
    N       scan slice to number in present radix, ignoring any _
    q       quit lexing
    r       set radix from scanned number, confirming within 1 <= r <= 64
    s       emit token as symbol from CS
    S       emit token as symbol from slice
*/

static char *lexen[] = {
    // Entry point for lexer: skip whitespace and detect end of file.
    ":ws",
    "[ \t\r]? [ws]:",   // Return courtesy Ananya and Windows Subsystem for Linux
    "[\n]= l [ws]:",
    "[;]= :",
    "( [parens]:",      // start multiline comment
    "[)]= [) is not within a comment]! f",
    "e q",
    "[~]:",

    // Flush to end of line after semicolon.
    ":;",
    "[\n]= l [ws]:",
    "e q",
    "b [;]:",

    // Go through multiline comment, being sure to keep track of line number.
    // number and making sure the comment does terminate.
    ":parens",
    ") [ws]:",
    "e q",              // aslex() caller keeps track of unterminated comment
    "[\n]= L [parens]:",
    "b [parens]:",

    // This hook picks up constant minus, denoted as ~.
    ":~",
    "[~]= - [seek]:",
    "+ [seek]:",

    // Lexer now points to something. This will be the rescan point after 
    // any radix is determined. The first character will tell us a little.
    ":seek",
    "m { [_]= :",
    "[0123456789]? [base10]:",
    "[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.]? [base64]:",
    "[nanid]:",

    // Numbers require digits, not just underscores.
    ":_",
    "[_]= :",
    "[`]= [numeral cannot be all separators]! f",
    "[0123456789]? [base10]:",
    "[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.]? [base64]:",
    "} [cannot attach ~ to identifier]! i [ws]:",   // _-only identifier done

    // At least one base 10 digit and possible underscores, nothing else yet.
    ":base10",
    "[0123456789_]? [base10]:",
    "[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.]? [base64]:",
    "[`]= :",
    "} [number too large for machine]! [10]n r N [number too negative]! c [ws]:",   // base 10 number

    // At least one base 64 digit and possible underscores
    ":base64",
    "[0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'._]? [base64]:",
    "[`]= :",
    "} [cannot attach ~ to identifier]! i [ws]:",   // no radix, so identifier

    // Lexer sees backtick. It needs a decimal radix or letter abbreviation.
    ":`",
    "[u]= [1]n r [check.end]:",
    "[b]= [2]n r [check.end]:",
    "[o]= [8]n r [check.end]:",
    "[d]= [10]n r [check.end]:",
    "[h]= [16]n r [check.end]:",
    "[t]= [64]n r [check.end]:",
    "[10] n r { [0123456789]? } [radix.digits]:",
    "[illegal or missing radix after `]! f",

    // Lexer saw ` and a digit. Get any additional digits, and convert to radix.
    ":radix.digits",
    "} [0123456789]? [radix.digits]:",
    "[` radix must be in range 1 through 64]! N r [check.end]:",

    // Ensure no higher digits or ` or _ are in the radix.
    ":check.end",
    "[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'._`]? [illegal character following radix]! f",
    "[num.with.radix]:",

    // The thing being lexed is a number with radix. Mark it for scanning.
    ":num.with.radix",
    "M { [marking.num]:",

    //  Finish marking and scan the number.
    ":marking.num",
    "[0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'._]? } [marking.num]:",
    "[digit too large for radix or number too large for machine]! N [flush.radix]:",

    // Skip over the ` and radix, because we already validated and scanned them.
    // Then emit a numeric constant token before returning to start.
    ":flush.radix",
    "[`0123456789ubodht]? [flush.radix]:",
    "[number too negative]! c [ws]:",

    // The thing being lexed is Not a Number or IDentifier.
    // Look for end of line/end of file or symbol.
    ":nanid",
    
    "[++]= s [ws]:",                    // multiple-character tokens
    "[+=]= s [ws]:",
    "[--]= s [ws]:",
    "[-=]= s [ws]:",
    "[~--]= s [ws]:",
    "[~-]= s [ws]:",
    "[!&]= s [ws]:",
    "[!|]= s [ws]:",
    "[!^]= s [ws]:",
    "[<=]= s [ws]:",
    "[>=]= s [ws]:",
    "[==]= s [ws]:",
    "[!=]= s [ws]:",
    "[::]= s [ws]:",

    "[+-&|^!<>=:]{?} S [ws]:",          // 1-char tokens that language uses

    "[~@#$%*{}[\\\",/?]{?} S [ws]:",    // 1-char tokens language doesn't use
    "{]}= S [ws]:",                     // but user macros might choose to

    "[unrecognized token]! f",          // something weird like UTF-8

    // Flush to whitespace after tokenizing error so more can be checked.
    ":flush.ws",
    "m[\n]= M [ws]:",           // found (newline)
    "[;]= M [ws]:",             // found (semicolon introduces comment)
    "e [ws]:",                  // found (EOF)
    "[(]= M [ws]:",             // found (parenthesis introduces comment)
    "b [flush.ws]:",            // not found (regular byte)

    NULL,
};

static const char *tetrasexagesimal = "0123456789abcdefghijklmnopqrstuvwxyz"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ'.";

static int aslex_lnum;          // cleaned up by as_dealloc()

static char **find_state(char **language, char *cs, char *cs_)
{
    char **r, *s;
    int len = cs_ - cs;

    if (0)
        gpa("state: %:\n", cs, cs_);

    for (r = language; s = *r; ++r) {
        if (*s != ':')
            continue;       // a rule, not a label
        ++s;
        if (strlen(s) != len)
            continue;       // label has wrong length
        if (memcmp(s, cs, len))
            continue;       // label does not match
        break;              // label matches
    }

    if (!*r)
        printf("ASLEX - offending rule: {%.*s}\n", len, cs);

    return r;               // *r may be NULL if not found
}

static long scan_u36(char *s, char *s_, int radix)
{
    int d;
    long n = 0;
    char *found;

    for (; s != s_; ++s) {
        if (*s == '_')
            continue;
        found = strchr(tetrasexagesimal, *s);
        if (!found) {
            printf("ASLEX - bad digit in scan_u36");
            return -1;
        }
        d = found - tetrasexagesimal;
        if (d >= radix)
            return -1;
        n = radix * n + d;
        if (n >= 1L << 36)
            return -1;
    }

    return n;
}

static int count_char_dups(char *s, int c)
{
    int n = 0;
    while (*s++ == c)
        ++n; 
    return n;
}

// Lex the program in 'src'. Returns 0 iff no errors.
// *parens should be 0 when lexing the first line. It's used as state
// information for comments that span aslex() calls.
static int aslex(
    char *src,              // start of line to be parsed
    int *parens             // last length of ( sequence seen
) {
    char *s = src;          // current parser location
    char *rule;             // remainder of present rule being tried
    char **rulep;           // pointer to present rule
    int c;                  // present character of rule being tried

    char *slice = src;      // marked slice of source using braces
    char *slice_ = src;     // end thereof

    char *cs = "ws";        // marked Constant Sequence in single quotes
    char *cs_ = NULL;       // end thereof

    char *msg = NULL;       // error message on reserve
    char *msg_ = NULL;      // end thereof

    int radix = 0;          // numeric constant radix

    char *remember = src;   // remembered place
    char *rewind = src;     // parser position at start of current rule
    int rewind_l = 0;       // line number as of rewind

    int errors = 0;         // number of errors noted

    long num = 0;           // scanned number

    int minus = 0;          // true iff ~ was lexed

    if (*parens)            // consider unterminated multiline comment
        cs = "parens";

new_state:
    if (!cs_)
        cs_ = strchr(cs, '\0');
    rulep = find_state(lexen, cs, cs_);
    if (as_trace)
        gpa("state: %:\n", cs, cs_);

    while (1) {

        if (as_trace)
            printf("s: <%.10s>\n", s);

        // If there is another rule for this state, fetch it.
        rule = *++rulep;

        if (!rule) {
            printf("ASLEX - rule not found\n");
            goto fatal;
        }

        if (*rule == ':') {
            printf("ASLEX - state out of rules\n");
            goto fatal;
        }

        if (as_trace)
            printf("    rule: %s\n", rule);

        rewind = s;
        rewind_l = aslex_lnum;
        msg = msg_ = NULL;

        cs = cs_ = rule;    // safety for malformed rules

        // Go through the steps of this rule.
        for (; c = *rule; ++rule) {

            switch (c) {

                case ' ':
                case '\t':
                    break;

                case '!':
                    msg = cs;
                    msg_ = cs_;
                    break;

                case '(':
                    if (*parens = count_char_dups(s, '('))
                        s += *parens;
                    else
                        goto nope;
                    break;

                case ')':
                    if (*parens == count_char_dups(s, ')'))
                        s += *parens;
                    else
                        goto nope;
                    break;

                case '+':
                    minus = 0;
                    break;

                case '-':
                    minus = 1;
                    break;

                case ':':
                    goto new_state;

                case '=':
                    if (memcmp(cs, s, cs_ - cs))
                        goto nope;
                    s += cs_ - cs;
                    break;

                case '?':
                    if (memchr(cs, *s, cs_ - cs))
                        ++s;
                    else
                        goto nope;
                    break;

                case '[':
                    cs = rule + 1;                  // char after opening [
                    cs_ = strchr(cs, ']');          // location of closing ]
                    if (!cs_) {
                        printf("ASLEX - unmatched [\n");
                        goto fatal;
                    }
                    rule = cs_;
                    break;

                case ']':
                    if (*s == ']')
                        ++s;
                    else
                        goto nope;
                    break;

                case '{':
                    slice = s;
                    break;

                case '}':
                    slice_ = s;
                    break;

                case 'b':
                    if (*s)
                        ++s;
                    else
                        goto nope;
                    break;

                case 'c':
                    if (minus && num > 1L << 35)
                        goto nope;
                    push_const_token(num, minus);
                    break;

                case 'e':
                    if (*s != '\0')
                        goto nope;
                    break;

                case 'f':
                    goto nope;

                case 'i':
                    if (minus)
                        goto nope;
                    push_ident_token(slice, slice_);
                    break;

                case 'l':
                    ++aslex_lnum;
                    push_lnum_token(aslex_lnum);
                    break;

                case 'L':
                    ++aslex_lnum;
                    break;

                case 'm':
                    remember = s;
                    break;

                case 'M':
                    s = remember;
                    break;

                case 'n':
                    num = scan_u36(cs, cs_, 10);
                    if (num < 0)
                        goto nope;
                    break;

                case 'N':
                    num = scan_u36(slice, slice_, radix);
                    if (num < 0)
                        goto nope;
                    break;

                case 'q':
                    goto quit;

                case 'r':
                    if (num < 1 || num > 64)
                        goto nope;
                    radix = num;
                    break;

                case 's':
                    push_sym_token(cs, cs_);
                    break;

                case 'S':
                    push_sym_token(slice, slice_);
                    break;

                default:
                    printf("ASLEX - ASCII %i in rule\n", c);
                    goto fatal;
            }
        }

        printf("ASLEX - rule without outcome\n");
        goto fatal;

nope:
        if (msg) {
            gpa("%e%:\n", aslex_lnum, msg, msg_);
            ++errors;
            cs = "flush.ws";
            cs_ = NULL;
            goto new_state;
        }
        s = rewind;
        aslex_lnum = rewind_l;
    }

quit:
    return errors;

fatal:
    return 1 + errors;
}
