/*
    Cross-assembler's SYNtax and code generation
    MWA 12 May 2023 ff

    This parses an array of tokens and ultimately generates code.
    See comments at top of aslex.c for rationale for this mini-interpreter.

    This is a two-pass assembler. The C code doesn't change at all or
    consider which pass we're in, but the rules do change. The two sets
    of rules are named 'pass1' and 'pass2'.
*/

/*
            whitespace is ignored
    !       issue CS as error messasge in the event this rule does not match
    $       note and advance parser past line number (as_lnum)
    +       create unsigned register within scope (must not already exist)
    -       create signed register within scope (must not already exist)
    .       advance parser past any token (wildcard; excludes as_eof)
    :       change state to that named in CS
    :Ohio   on line by itself, introduces rules for state called Ohio
    >       rule can't backtrack, so prepare a default error message
    ?       trace: display CS and current parser position
    [abc]   set constant sequence (CS) of characters
    0       generate instruction word using opcode looked up from CS
    1       shift (first) most recent register into instruction word
    2       shift (second) most recent register into instruction word
    3       shift (third) most recent register into instruction word
    7       emit IMP + dest reg + (label or scope line number)
    8       parse existing scope and set instruction word to its line number
    9       shift 0 into instruction word (9 means like, unused)
    c       advance parser past constant (as_const)
    C       mark const as needing a register
    d       test const and mark as needing a register if it can't IMM
    e       peek for end of tokens: continue rule if as_eof appears here
    f       fail (rule does not match) and jump to rule [const]
    g       set (global) scope (already known to exist) and code offset
    G       create (global) scope (must not already exist)
    i       set 'id' and advance parser past identifier (as_ident)
    I       attempt to generate IMM for constant assignment
    J       like 'K' except specifies a forced register number
    k       advance parser past keyword (as_ident specified in CS)
    K       add register within scope to kept list (must not already be kept)
    l       parse existing label and set instruction word to its line number
    m       set label code offset
    L       create label within scope (must not already exist)
    o       advance parser past non-lvalue register (o stands for 'o'utright)
    O       advance parser past lvalue register
    P       advance parser past opcode emit IMM for it
    q       quit parsing
    s       advance parser past symbol (as_sym specified in CS)
    S       limit next O or o to 'S'igned registers
    U       limit next O or o to 'U'nsigned registers
    W       limit next O or o to 'W'rapped registers
    x       register number (0-511) or opcode for 'emit'
    X       issue opcode and three registers for 'emit'
    z       shift zero constant's register number into instruction word
*/

/*
    SOME NOTES

    We're going to drop the IMM keyword and just use "x = 14".
*/

// Pass 1 collects valid label and register names with scopes and signages.
// It also collects constants that need registers and marks kept registers.
// I've placed some constructs in rough order of frequency anticipated.
static char *pass1[] = {
    // Many lines are parsed during pass 2. For now, just collect their consts.
    // MWA This rule name is hardcoded for flushing lines after errors. Keep it.
    ":$",
    "c C [$]:",
    "$ [bol]:",
    ". [$]:",

    // The parser starts each line here.
    ":bol",
    "$ [bol]:",
    "[u.]k :",                      // u.
    "[unsigned]k [u.]:",            // unsigned
    "[s.]k :",                      // s.
    "[signed]k [s.]:",              // signed
    "i [:]s L [bol]:",              // label:
    "i [::]s $ G [bol]:",           // scope::
    "[keep]k :",                    // keep
    "[emit]k i x x x $ [bol]:",     // emit (opcode, 3 reg. numbers)
    "e q",                          // end of input
    "i [::]s i [=]s c $ d [bol]:",  // x::y = 12345
    "i [=]s c $ d [bol]:",          // z = 54321
    "[$]:",                         // (anything else)

    // unsigned registers: specify 0 or more
    ":u.",
    "$ [bol]:",
    "i + [u.]:",
    ". [invalid character in register name]! f",

    // signed registers: specify 0 or more
    ":s.",
    "$ [bol]:",
    "i - [s.]:",
    ". [invalid character in register name]! f",

    // kept registers: specify 0 or more
    ":keep",
    "$ [bol]:",
    "i c J [keep]:",
    "i K [keep]:",
    ". [invalid character in register name]! f",

    NULL,
};

// Pass two assembles the program, but leaves branches for later resolution.
static char *pass2[] = {
    // Lines already finished by pass 1 can be flushed.
    ":$",
    "$ [bol]:",
    ". [$]:",

    // The parser starts each line here.
    ":bol",
    "$ [bol]:",
    "[u.]k [$]:",                   // u. (flush)
    "[unsigned]k [$]:",             // unsigned (flush)
    "[s.]k [$]:",                   // s. (flush)
    "[signed]k [$]:",               // signed (flush)
    "i [:]s m [bol]:",              // label: (set code offset)
    "i [::]s $ g [bol]:",           // scope:: (set current scope & code offset)
    "[keep]k [$]:",                 // keep (flush)
    "e q",                          // end of input

    // For better or worse, the keyword statements have to parse
    // before the assignment statements. This is to protect the
    // keyword from false "register does not exist" alarms.

    // bound
    "[bound]k So > [<]s o $ [bound.s]9120 [bol]:",
    "[bound]k > Uo [<]s o $ [bound.u]9120 [bol]:",

    // cmp
    "[cmp]k Uo [-]s Uo $ [cmp.uu]9210 [bol]:",
    "[cmp]k Uo [-]s So $ [cmp.us]9210 [bol]:",
    "[cmp]k So [-]s Uo $ [cmp.su]9210 [bol]:",
    "[cmp]k > So [-]s So $ [cmp.ss]9210 [bol]:",

    // instructions with no operands
    "[crf]k > $ 9990 [bol]:",
    "[halt]k > $ 9990 [bol]:",
    "[hijack]k > $ 9990 [bol]:",      // not intended for assembler programs
    "[memset]k > $ 9990 [bol]:",
    "[nolist]k > $ 9990 [bol]:",
    "[nop]k > $ 9990 [bol]:",
    "[npriv]k > $ 9990 [bol]:",
    "[priv]k > $ 9990 [bol]:",
    "[r.]k [via]k > i 8 $ [j]0 [bol]:",
    "[return]k [via]k > i 8 $ [j]0 [bol]:",
    "[r.]k > $ [r]9990 [bol]:",
    "[return]k > $ [r]9990 [bol]:",
    "[revert]k > $ 9990 [bol]:",
    "[setup]k > $ 9990 [bol]:",
    "[yield]k > $ 9990 [bol]:",
    "[rjinx]k > $ [r]9990 [bol]:",

    // [keyword] everywhere (left, right, dest)
    "[timer]k > O $ 1110 [bol]:",

    // [keyword] left = right
    "[wdm]k > o [=]s o $ [wdm]9210 [bol]:",
    "[wdm2]k > o [=]s o $ [wdm2]9210 [bol]:",
    "[wpt]k > o [=]s o $ [wpt]9210 [bol]:",
    "[sto]k > o [=]s o $ [sto]9210 [bol]:",
    "[sto2]k > o [=]s o $ [sto2]9210 [bol]:",

    // [keyword] right = left
    "[wcm]k > o [=]s o $ [wcm]9120 [bol]:",

    // [keyword] dest = left
    "[poke]k > O [=]s o $ [poke]2190 [bol]:",

    // [keyword] left
    "[xany]k > o $ 9190 [bol]:",

    // [keyword] right
    "[jany]k > o $ 9910 [bol]:",
    "[rcm1]k > o $ 9910 [bol]:",
    "[user]k > o $ 9910 [bol]:",

    // [keyword] dest
    "[rcm2]k > O $ 1990 [bol]:",

    // [keyword] scope
    "[c.]k > i 8 $ [c]0 [bol]:",
    "[call]k > i 8 $ [c]0 [bol]:",
    "[pcall]k > i 8 $ 0 [bol]:",
    "[npcall]k > i 8 $ 0 [bol]:",

    // [keyword] label
    "[lcall]k > i l $ [c]0 [bol]:",
    "[cali]k > i l $ 0 [bol]:",
    "[j.]k i > l $ [j]0 [bol]:",
    "[j.]k [<]s > i l $ [j.lt]0 [bol]:",
    "[j.]k [<=]s > i l $ [j.le]0 [bol]:",
    "[j.]k [==]s > i l $ [j.eq]0 [bol]:",
    "[j.]k [!=]s > i l $ [j.ne]0 [bol]:",
    "[j.]k [>=]s > i l $ [j.ge]0 [bol]:",
    "[j.]k [>]s > i l $ [j.gt]0 [bol]:",
    "[j.]k [+]s [r]k > i l $ [j.r]0 [bol]:",
    "[j.]k [-]s [r]k > i l $ [j.nr]0 [bol]:",
    "[j.]k [+]s > [t]k i l $ [j.t]0 [bol]:",
    "[j.]k [-]s > [t]k i l $ [j.nt]0 [bol]:",
    "[jump]k i > l $ [j]0 [bol]:",
    "[jump]k [<]s > i l $ [j.lt]0 [bol]:",
    "[jump]k [<=]s > i l $ [j.le]0 [bol]:",
    "[jump]k [==]s > i l $ [j.eq]0 [bol]:",
    "[jump]k [!=]s > i l $ [j.ne]0 [bol]:",
    "[jump]k [>=]s > i l $ [j.ge]0 [bol]:",
    "[jump]k [>]s > i l $ [j.gt]0 [bol]:",
    "[jump]k [+]s [r]k > i l $ [j.r]0 [bol]:",
    "[jump]k [-]s [r]k > i l $ [j.nr]0 [bol]:",
    "[jump]k [+]s > [t]k i l $ [j.t]0 [bol]:",
    "[jump]k [-]s > [t]k i l $ [j.nt]0 [bol]:",
    "[jinx]k i > l $ [j]0 [bol]:",

    // [keyword] opcode reg_no reg_no reg_no
    "[emit]k > [emit requires opcode and 3 register numbers]! x x x x $ X [bol]:",

    // Unary operations come next.

    // dest = [keyword] left
    "O [=]s [ld]k o $ 2190 [bol]:",
    "O [=]s [rdm]k o $ 2190 [bol]:",
    "O [=]s [rpt]k o $ 2190 [bol]:",
    "O [=]s [peek]k o $ 2190 [bol]:",

    // dest = [keyword] right
    "O [=]s [not]k o $ [nr]2910 [bol]:",
    "O [=]s [!]s o $ [nr]2910 [bol]:",

    // dest = "opcode" [opcode]
    "O [=]s [opcode]k i $ > [invalid opcode on previous line]! P [bol]:",

    // dest = ::[scope]
    // dest = :[label]:
    // This replaces the old assembler's "cursor" hack.
    "O [=]s [::]s i 8 > $ 7 [bol]:",
    "O [=]s [:]s i l > $ 7 [bol]:",

    // unary +
    // This is same as assignment except for [+]s.
    // This doesn't match for automatic IMM optimization, so it's also
    // a way you can suppress that for testing etc.
    "UO [=]s [+]s Uo [a.uuu]21z0 [bol]:",
    "UO [=]s [+]s So [a.suu]21z0 [bol]:",
    "SO [=]s [+]s Uo [a.uus]21z0 [bol]:",
    "SO [=]s [+]s So [a.sus]21z0 [bol]:",
    "WO [=]s [+]s o [a.w]21z0 [bol]:",

    // unary ++
    "UO [=]s [++]s Uo [ac.uuu]21z0 [bol]:",
    "UO [=]s [++]s So [ac.suu]21z0 [bol]:",
    "SO [=]s [++]s Uo [ac.uus]21z0 [bol]:",
    "SO [=]s [++]s So [ac.sus]21z0 [bol]:",
    "WO [=]s [++]s o [ac.w]21z0 [bol]:",

    // unary -
    "UO [=]s [-]s Uo [s.uuu]2z10 [bol]:",
    "UO [=]s [-]s So [s.usu]2z10 [bol]:",
    "SO [=]s [-]s Uo [s.uus]2z10 [bol]:",
    "SO [=]s [-]s So [s.uss]2z10 [bol]:",
    "WO [=]s [-]s o [s.w]2z10 [bol]:",

    // unary --
    "UO [=]s [--]s Uo [sc.uuu]2z10 [bol]:",
    "UO [=]s [--]s So [sc.usu]2z10 [bol]:",
    "SO [=]s [--]s Uo [sc.uus]2z10 [bol]:",
    "SO [=]s [--]s So [sc.uss]2z10 [bol]:",
    "WO [=]s [--]s o [sc.w]2z10 [bol]:",

    // IMM generation (optimization)
    "O [=]s c $ I [bol]:",

    // assignment (needs to follow keywords)
    "UO [=]s Uo $ [a.uuu]21z0 [bol]:",
    "UO [=]s So $ [a.suu]21z0 [bol]:",
    "SO [=]s Uo $ [a.uus]21z0 [bol]:",
    "SO [=]s So $ [a.sus]21z0 [bol]:",

    // All clear for binary functions; that is, statements in the form
    // [lvalue] = [something] [operation] [something].

    // add
    "UO [=]s Uo [+]s Uo $ [a.uuu]3210 [bol]:",
    "SO [=]s Uo [+]s Uo $ [a.uus]3210 [bol]:",
    "UO [=]s Uo [+]s So $ [a.usu]3210 [bol]:",
    "SO [=]s Uo [+]s So $ [a.uss]3210 [bol]:",
    "UO [=]s So [+]s Uo $ [a.suu]3210 [bol]:",
    "SO [=]s So [+]s Uo $ [a.sus]3210 [bol]:",
    "UO [=]s So [+]s So $ [a.ssu]3210 [bol]:",
    "SO [=]s So [+]s So $ [a.sss]3210 [bol]:",
    "UO [+=]s Uo $ [a.uuu]2210 [bol]:",
    "UO [+=]s So $ [a.usu]2210 [bol]:",
    "SO [+=]s Uo $ [a.sus]2210 [bol]:",
    "SO [+=]s So $ [a.sss]2210 [bol]:",


    // subtract
    "UO [=]s Uo [-]s Uo $ [s.uuu]3210 [bol]:",
    "SO [=]s Uo [-]s Uo $ [s.uus]3210 [bol]:",
    "UO [=]s Uo [-]s So $ [s.usu]3210 [bol]:",
    "SO [=]s Uo [-]s So $ [s.uss]3210 [bol]:",
    "UO [=]s So [-]s Uo $ [s.suu]3210 [bol]:",
    "SO [=]s So [-]s Uo $ [s.sus]3210 [bol]:",
    "UO [=]s So [-]s So $ [s.ssu]3210 [bol]:",
    "SO [=]s So [-]s So $ [s.sss]3210 [bol]:",
    "UO [-=]s Uo $ [s.uuu]2210 [bol]:",
    "UO [-=]s So $ [s.usu]2210 [bol]:",
    "SO [-=]s Uo $ [s.sus]2210 [bol]:",
    "SO [-=]s So $ [s.sss]2210 [bol]:",

    // reverse subtract
    "UO [=]s Uo [~-]s Uo $ [rs.uuu]3210 [bol]:",
    "SO [=]s Uo [~-]s Uo $ [rs.uus]3210 [bol]:",
    "UO [=]s Uo [~-]s So $ [rs.usu]3210 [bol]:",
    "SO [=]s Uo [~-]s So $ [rs.uss]3210 [bol]:",
    "UO [=]s So [~-]s Uo $ [rs.suu]3210 [bol]:",
    "SO [=]s So [~-]s Uo $ [rs.sus]3210 [bol]:",
    "UO [=]s So [~-]s So $ [rs.ssu]3210 [bol]:",
    "SO [=]s So [~-]s So $ [rs.sss]3210 [bol]:",

    // add with carry
    "UO [=]s Uo [++]s Uo $ [ac.uuu]3210 [bol]:",
    "SO [=]s Uo [++]s Uo $ [ac.uus]3210 [bol]:",
    "UO [=]s Uo [++]s So $ [ac.usu]3210 [bol]:",
    "SO [=]s Uo [++]s So $ [ac.uss]3210 [bol]:",
    "UO [=]s So [++]s Uo $ [ac.suu]3210 [bol]:",
    "SO [=]s So [++]s Uo $ [ac.sus]3210 [bol]:",
    "UO [=]s So [++]s So $ [ac.ssu]3210 [bol]:",
    "SO [=]s So [++]s So $ [ac.sss]3210 [bol]:",

    // subtract with carry
    "UO [=]s Uo [--]s Uo $ [sc.uuu]3210 [bol]:",
    "SO [=]s Uo [--]s Uo $ [sc.uus]3210 [bol]:",
    "UO [=]s Uo [--]s So $ [sc.usu]3210 [bol]:",
    "SO [=]s Uo [--]s So $ [sc.uss]3210 [bol]:",
    "UO [=]s So [--]s Uo $ [sc.suu]3210 [bol]:",
    "SO [=]s So [--]s Uo $ [sc.sus]3210 [bol]:",
    "UO [=]s So [--]s So $ [sc.ssu]3210 [bol]:",
    "SO [=]s So [--]s So $ [sc.sss]3210 [bol]:",

    // reverse subtract with carry
    "UO [=]s Uo [~--]s Uo $ [rsc.uuu]3210 [bol]:",
    "SO [=]s Uo [~--]s Uo $ [rsc.uus]3210 [bol]:",
    "UO [=]s Uo [~--]s So $ [rsc.usu]3210 [bol]:",
    "SO [=]s Uo [~--]s So $ [rsc.uss]3210 [bol]:",
    "UO [=]s So [~--]s Uo $ [rsc.suu]3210 [bol]:",
    "SO [=]s So [~--]s Uo $ [rsc.sus]3210 [bol]:",
    "UO [=]s So [~--]s So $ [rsc.ssu]3210 [bol]:",
    "SO [=]s So [~--]s So $ [rsc.sss]3210 [bol]:",

    // wrapping + - ~- ++ -- ~--
    "WO [=]s o [+]s o $ [a.w]3210 [bol]:",
    "WO [=]s o [-]s o $ [s.w]3210 [bol]:",
    "WO [=]s o [~-]s o $ [rs.w]3210 [bol]:",
    "WO [=]s o [++]s o $ [ac.w]3210 [bol]:",
    "WO [=]s o [--]s o $ [sc.w]3210 [bol]:",
    "WO [=]s o [~--]s o $ [rsc.w]3210 [bol]:",

    // bitwise boolean by keyword
    "O [=]s o [igf]k o $ 3210 [bol]:",
    "O [=]s o [igt]k o $ 3210 [bol]:",
    "O [=]s o [xl]k o $ 3210 [bol]:",
    "O [=]s o [xr]k o $ 3210 [bol]:",
    "O [=]s o [nl]k o $ 3210 [bol]:",
    "O [=]s o [nr]k o $ 3210 [bol]:",
    "O [=]s o [and]k o $ 3210 [bol]:",
    "O [=]s o [nand]k o $ 3210 [bol]:",
    "O [=]s o [or]k o $ 3210 [bol]:",
    "O [=]s o [nor]k o $ 3210 [bol]:",
    "O [=]s o [xor]k o $ 3210 [bol]:",
    "O [=]s o [xnor]k o $ 3210 [bol]:",
    "O [=]s o [lanr]k o $ 3210 [bol]:",
    "O [=]s o [ranl]k o $ 3210 [bol]:",
    "O [=]s o [lonr]k o $ 3210 [bol]:",
    "O [=]s o [ronl]k o $ 3210 [bol]:",

    // bitwise boolean using symbols, no ! on arguments
    "O [=]s o [&]s o $ [and]3210 [bol]:",
    "O [=]s o [|]s o $ [or]3210 [bol]:",
    "O [=]s o [^]s o $ [xor]3210 [bol]:",
    "O [=]s o [!&]s o $ [nand]3210 [bol]:",
    "O [=]s o [!|]s o $ [nor]3210 [bol]:",
    "O [=]s o [!^]s o $ [xnor]3210 [bol]:",

    // bitwise boolean using symbols, ! on left argument
    "O [=]s [!]s o [&]s o $ [ranl]3210 [bol]:",
    "O [=]s [!]s o [|]s o $ [ronl]3210 [bol]:",
    "O [=]s [!]s o [^]s o $ [xnor]3210 [bol]:",
    "O [=]s [!]s o [!&]s o $ [lonr]3210 [bol]:",
    "O [=]s [!]s o [!|]s o $ [lanr]3210 [bol]:",
    "O [=]s [!]s o [!^]s o $ [xor]3210 [bol]:",

    // bitwise boolean using symbols, ! on right argument
    "O [=]s o [&]s [!]s o $ [lanr]3210 [bol]:",
    "O [=]s o [|]s [!]s o $ [lonr]3210 [bol]:",
    "O [=]s o [^]s [!]s o $ [xnor]3210 [bol]:",
    "O [=]s o [!&]s [!]s o $ [ronl]3210 [bol]:",
    "O [=]s o [!|]s [!]s o $ [ranl]3210 [bol]:",
    "O [=]s o [!^]s [!]s o $ [xor]3210 [bol]:",

    // bitwise boolean using symbols, ! on both arguments
    "O [=]s [!]s o [&]s [!]s o $ [nor]3210 [bol]:",
    "O [=]s [!]s o [|]s [!]s o $ [nand]3210 [bol]:",
    "O [=]s [!]s o [^]s [!]s o $ [xor]3210 [bol]:",
    "O [=]s [!]s o [!&]s [!]s o $ [or]3210 [bol]:",
    "O [=]s [!]s o [!|]s [!]s o $ [and]3210 [bol]:",
    "O [=]s [!]s o [!^]s [!]s o $ [xnor]3210 [bol]:",

    // maximum
    "UO [=]s Uo [max]k Uo $ [max.uuu]3210 [bol]:",
    "SO [=]s Uo [max]k Uo $ [max.uus]3210 [bol]:",
    "UO [=]s Uo [max]k So $ [max.usu]3210 [bol]:",
    "SO [=]s Uo [max]k So $ [max.uss]3210 [bol]:",
    "UO [=]s So [max]k Uo $ [max.suu]3210 [bol]:",
    "SO [=]s So [max]k Uo $ [max.sus]3210 [bol]:",
    "UO [=]s So [max]k So $ [max.ssu]3210 [bol]:",
    "SO [=]s So [max]k So $ [max.sss]3210 [bol]:",

    // minimum
    "UO [=]s Uo [min]k Uo $ [min.uuu]3210 [bol]:",
    "SO [=]s Uo [min]k Uo $ [min.uus]3210 [bol]:",
    "UO [=]s Uo [min]k So $ [min.usu]3210 [bol]:",
    "SO [=]s Uo [min]k So $ [min.uss]3210 [bol]:",
    "UO [=]s So [min]k Uo $ [min.suu]3210 [bol]:",
    "SO [=]s So [min]k Uo $ [min.sus]3210 [bol]:",
    "UO [=]s So [min]k So $ [min.ssu]3210 [bol]:",
    "SO [=]s So [min]k So $ [min.sss]3210 [bol]:",

    // asl
    "UO [=]s Uo [asl]k o $ [asl.uu]3210 [bol]:",
    "SO [=]s Uo [asl]k o $ [asl.us]3210 [bol]:",
    "UO [=]s So [asl]k o $ [asl.su]3210 [bol]:",
    "SO [=]s So [asl]k o $ [asl.ss]3210 [bol]:",

    // asr
    "UO [=]s Uo [asr]k o $ [asr.uu]3210 [bol]:",
    "SO [=]s Uo [asr]k o $ [asr.us]3210 [bol]:",
    "UO [=]s So [asr]k o $ [asr.su]3210 [bol]:",
    "SO [=]s So [asr]k o $ [asr.ss]3210 [bol]:",

    // simple 3-register ALU cases without signage
    "O [=]s o [lsl]k o $ 3210 [bol]:",
    "O [=]s o [lsr]k o $ 3210 [bol]:",
    "O [=]s o [rol]k o $ 3210 [bol]:",
    "O [=]s o [ml]k o $ 3210 [bol]:",
    "O [=]s o [mh]k o $ 3210 [bol]:",
    "O [=]s o [mhns]k o $ 3210 [bol]:",
    "O [=]s o [mhl]k o $ 3210 [bol]:",
    "O [=]s o [mhl0]k o $ 3210 [bol]:",
    "O [=]s o [mhl1]k o $ 3210 [bol]:",
    "O [=]s o [mhl2]k o $ 3210 [bol]:",
    "O [=]s o [mhl3]k o $ 3210 [bol]:",
    "O [=]s o [mhl4]k o $ 3210 [bol]:",
    "O [=]s o [mhl5]k o $ 3210 [bol]:",
    "O [=]s o [dsl]k o $ 3210 [bol]:",
    "O [=]s o [swiz]k o $ 3210 [bol]:",
    "O [=]s o [nudge]k o $ 3210 [bol]:",
    "O [=]s o [pit]k o $ 3210 [bol]:",
    "O [=]s o [pat]k o $ 3210 [bol]:",
    "O [=]s o [pait]k o $ 3210 [bol]:",
    "O [=]s o [mix]k o $ 3210 [bol]:",
    "O [=]s o [xim]k o $ 3210 [bol]:",
    "O [=]s o [ham2]k o $ 3210 [bol]:",
    "O [=]s o [txor]k o $ 3210 [bol]:",
    "O [=]s o [un.a]k o $ 3210 [bol]:",
    "O [=]s o [un.b]k o $ 3210 [bol]:",
    "O [=]s o [un.g]k o $ 3210 [bol]:",
    "O [=]s o [stun.a]k o $ 3210 [bol]:",
    "O [=]s o [stun.b]k o $ 3210 [bol]:",
    "O [=]s o [stun.c]k o $ 3210 [bol]:",
    "O [=]s o [stun.d]k o $ 3210 [bol]:",
    "O [=]s o [stun.e]k o $ 3210 [bol]:",
    "O [=]s o [stun.f]k o $ 3210 [bol]:",
    "O [=]s o [stun.g]k o $ 3210 [bol]:",
    "O [=]s o [stun.h]k o $ 3210 [bol]:",
    "O [=]s o [stun.i]k o $ 3210 [bol]:",
    "O [=]s o [stun.j]k o $ 3210 [bol]:",
    "O [=]s o [stun.k]k o $ 3210 [bol]:",

    // dest = left [keyword] right
    "O [=]s o [addld]k o $ 3210 [bol]:",
    "O [=]s o [addrdm]k o $ 3210 [bol]:",
    "O [=]s o [ldsto]k o $ 3210 [bol]:",
    "O [=]s o [rwdm]k o $ 3210 [bol]:",
    "O [=]s o [pvio]k o $ 3210 [bol]:",

    ". [syntax error]! f",          // XXX work in progress

    NULL,
};

/*
    The target architecture doesn't support recursion, but there is a largish
    construct that appears up to three times for many opcodes. Specifically:

    (<wrap>|<signed>|<s.>|<unsigned>|<u.>|)((scope::|)reg)|const

    In the above, | denotes alternatives within parentheses.
    "scope" and "reg" are identifiers, and "const" is a number.

    The cleanest way to implement this is in outright code.
*/
static
char *outright(     // returns error message / "" (success) / NULL (no match)
    Token **tok,    // input/output: parser position
    Scope *cscope,  // current scope
    int lvalue,     // allows <wrap>, prohibits const
    int signage,    // if nonzero, require 'u'nsigned, 's'igned, or 'w'rap
    int *reg_no,    // output: register number
    Reg **reg_out,  // output: register (only safe for lvalues)
    int *usw_out    // output: signage (only safe for lvalues)
) {
    Token *t = *tok;
    Scope *scope = cscope;  // parsed scope
    Reg *reg;               // parsed register
    Str *id;                // parsed identifier
    int usw = 0;            // char 'u', 's', 'w'
    Const *con;             // parsed constant

    if (!same_sym(t, "<", NULL))
        goto no_cast;
    t = t->next;
    if (same_ident(t, "wrap", NULL)) {
        usw = 'w';
        if (!lvalue)
            return "<wrap> only valid on lvalues";
    }
    else if (same_ident(t, "s.", NULL))
        usw = 's';
    else if (same_ident(t, "signed", NULL))
        usw = 's';
    else if (same_ident(t, "u.", NULL))
        usw = 'u';
    else if (same_ident(t, "unsigned", NULL))
        usw = 's';
    else
        return "bad cast after <";
    t = t->next;
    if (!same_sym(t, ">", NULL))
        return "missing > after cast";
    t = t->next;

no_cast:
    if (!lvalue && t->tag == as_const) {
        con = (Const *) t->payload;
        if (!usw)
            usw = con->minus ? 's' : 'u';
        if (signage && signage != usw)
            return NULL;                // signage specified, but doesn't match
        *reg_no = con->reg_no;
        t = t->next;
        *tok = t;
        return "";                      // successfully parsed constant!
    }

    if (t->tag != as_ident)
        return usw ? "need value after cast" : NULL;
    id = (Str *) t->payload;            // probably a register name
    t = t->next;

    if (same_sym(t, "::", NULL)) {
        // id was actually a scope name
        ifnot (scope = as_search(scopes, (ulong) id))
            return "scope does not exist";
        t = t->next;
        if (t->tag != as_ident)
            return "need register name after scope";
        id = (Str *) t->payload;
        t = t->next;
    }

    ifnot (reg = as_search2(regs, (ulong) id, (ulong) scope))
        return "register does not exist";

    if (scope != cscope && !(reg->flags & as_kept))
        return "register is out of scope and not kept";

    if (!usw)
        usw = reg->flags & as_signed ? 's' : 'u';
    if (signage && signage != usw)
        return NULL;                    // signage specified, but doesn't match
    *reg_no = reg->reg_no;
    *reg_out = reg;
    *usw_out = usw;
    *tok = t;
    return "";                          // successfully parsed register!
}

// Parse the program in 'src'. Returns 0 iff no errors.
static int assyn(Token *src, char **language)
{
    char *rule;             // remainder of present rule being tried
    char **rulep;           // pointer to present rule
    int c;                  // present character of rule being tried

    char *cs = "bol";       // marked Constant Sequence in single quotes
    char *cs_ = NULL;       // end thereof

    char *msg = NULL;       // error message on reserve
    char *msg_ = NULL;      // end thereof

    int errors = 0;         // number of errors noted

    ulong lnum = 0;         // line number (dead reckoned from tokens)
    Const *conp = NULL;     // temporary use

    Token *t = src->next;   // current parser position
    Token *rewind = t;      // parser position at start of current rule

    Scope *scope = NULL;    // line 0 will initialize this
    Str *id = NULL;         // some identifier; line 0 will initialize
    Label *label = NULL;    // temporary use (pass1), parsed label (pass2)
    Reg *reg = NULL;        // temporary use (pass1), last reg parsed (pass2)
    Keep *keep = NULL;      // temporary use
    Scope *call_me = NULL;  // temporary use

    int signage_want = 0;   // if nonzero, outright() looks for 's', 'u', 'w'
    int signage_eff = 0;    // from last outright()

    int reg_no1 = 0;        // three-stack of registers from outright() calls
    int reg_no2 = 0;
    int reg_no3 = 0;

    int temp;               // for local use within any "case _:"
    char *m;                // temporary use

    ulong instr = 0;        // instruction being generated by [opcode]3210

    FILE *call_log = NULL;  // file for generating a call graph
    FILE *kept_vars = NULL; // file listing kept variables

    // for documentation and diagnostics
    call_log = fopen(language == pass2 ? "/tmp/call.log" : "/dev/null", "w");
    kept_vars = fopen(language == pass1 ? "/tmp/kept.vars" : "/dev/null", "w");

new_state:
    if (!cs_)
        cs_ = strchr(cs, '\0');
    rulep = find_state(language, cs, cs_);
    if (as_trace)
        gpa("state: %:\n", cs, cs_);

    while (1) {

        if (as_trace)
            gpa("  token: %t\n", t);

        // If there is another rule for this state, fetch it.
        rule = *++rulep;

        if (!rule) {
            printf("ASSYN BUG - rule not found\n");
            goto fatal;
        }

        if (*rule == ':') {
            printf("ASSYN BUG - state out of rules\n");
            goto fatal;
        }

        if (as_trace)
            printf("    rule: %s\n", rule);

        rewind = t;
        msg = msg_ = NULL;

        cs = cs_ = rule;    // safety for malformed rules

        // Go through the steps of this rule.
        for (; c = *rule; ++rule) {

            switch (c) {

                case '\t':
                case ' ':
                    break;

                case '!':
                    msg = cs;
                    msg_ = cs_;
                    break;

                case '$':
                    if (t->tag != as_lnum)
                        goto nope;
                    lnum = t->payload;
                    t = t->next;
                    break;

                case '+':
                    if (!as_search2(regs, (ulong) id, (ulong) scope)) {
                        appear(reg, "Reg");         // create new register
                        reg->name = id;
                        reg->scope = scope;
                        reg->flags = as_unsigned;
                        as_append(&regs, reg);
                    }
                    else {
reg_exists:
                        // TODO Try to accommodate this message in bytecode,
                        //      even though parsing continues on same line.
                        gpa("%eregister %O::%$ already exists\n", lnum, scope, id);
                        ++errors;
                    }
                    break;

                case '-':
                    if (!as_search2(regs, (ulong) id, (ulong) scope)) {
                        appear(reg, "Reg");         // create new register
                        reg->name = id;
                        reg->scope = scope;
                        reg->flags = as_signed;
                        as_append(&regs, reg);
                    }
                    else
                        goto reg_exists;
                    break;

                case '.':
                    if (t->tag == as_eof)
                        goto nope;
                    t = t->next;
                    break;

                case ':':
                    goto new_state;

                case '[':
                    cs = rule + 1;                  // char after opening [
                    cs_ = strchr(cs, ']');          // location of closing ]
                    if (!cs_) {
                        printf("ASSYN BUG - unmatched [\n");
                        goto fatal;
                    }
                    rule = cs_;
                    break;

                case '>':
                    msg = "line starts ok, but fails before end";
                    msg_ = NULL;
                    break;

                case '?':
                    gpa("trace: %: <%t>\n", cs, cs_, t);
                    break;

                case '0':
                    temp = name_to_opcode(cs, cs_);
                    if (temp < 0) {
                        msg = "internal error (unknown opcode)";
                        msg_ = NULL;
                        goto nope;
                    }
                    instr = instr & 0777777777 | (ulong) temp << 27;
                    push_code(instr);
                    break;

                case '1':
                    instr = instr << 9 | reg_no1;
                    break;

                case '2':
                    instr = instr << 9 | reg_no2;
                    break;

                case '3':
                    instr = instr << 9 | reg_no3;
                    break;

                case '7':
                    instr |= (ulong) oc_asm_addr << 27 | (ulong) reg_no1 << 18;
                    push_code(instr);
                    break;

                case '8':
                    call_me = as_search(scopes, (ulong) id);
                    if (!call_me) {
                        msg = "scope not defined";
                        msg_ = NULL;
                        goto nope;
                    }

                    // In /tmp/call.log, write lines containing calling
                    // and called scope names so I can document complex
                    // programs.
                    if (strncmp("::", cs, 2)) {
                        say_str(scope->name, call_log);
                        fputc(' ', call_log);
                        say_str(call_me->name, call_log);
                        fputc('\n', call_log);
                    }
                    instr = call_me->lnum;
                    break;

                // Although 'z' and '9' presently do the same thing and
                // put zeros in the instruction word, 'z' semantically means
                // the zero constant's register number, which may change.
                // Warning: aspre.c uses register 0 as a temporary before
                // making it the zero constant, so keep that synchronized.
                case '9': case 'z':
                    instr = instr << 9;             // leave operand empty
                    break;

                case 'c':
                    if (t->tag != as_const)
                        goto nope;
                    conp = (Const *) t->payload;
                    t = t->next;
                    break;

                case 'C':
                    conp->reg_no = 0;               // remove 'as_immed_only'
                    break;

                case 'd':
                    if (!make_imm(conp->number))    // can we not IMM this?
                        conp->reg_no = 0;           // remove 'as_immed_only'
                    break;

                case 'e':
                    if (t->tag != as_eof)
                        goto nope;
                    break;

                case 'f':
                    goto nope;

                case 'g':
                    scope = as_search(scopes, (ulong) id);
                    if (!scope)
                        asfail("scope missing in pass2");
                    scope->code_addr = n_code;
                    break;

                case 'G':
                    if (as_search(scopes, (ulong) id)) {
                        gpa("%escope %$ already exists\n", lnum, id);
                        ++errors;
                        goto nope;
                    }
                    appear(scope, "Scope");         // create new scope
                    scope->name = id;
                    scope->lnum = lnum;
                    as_append(&scopes, scope);
                    break;

                case 'i':
                    if (t->tag != as_ident)
                        goto nope;
                    id = (Str *) t->payload;
                    t = t->next;
                    break;

                case 'I':
                    if ((signage_eff == 's') != conp->minus
                            && conp->number & 1L << 35) {

                        // This is the ONLY situation where we're one too
                        // high on the reported line number, and it's because
                        // we made sure the line ends before applying the IMM
                        // optimization to assignment. This isn't something
                        // that's going to become huge like Pinocchio's nose,
                        // and I think it's okay in this bizarre case (who
                        // writes constants that don't fit?) to indicate in
                        // the message that the previous line is the culprit.
                        // Because the "solution" would add much churn.
                        msg = "constant on previous line does not fit in register";
                        msg_ = NULL;
                        goto nope;
                    }
                    ifnot (instr = make_imm(conp->number))
                        goto nope;                  // add zero instead
                    instr |= reg_no1 << 18;
                    push_code(instr);
                    break;

                case 'k':
                    if (!same_ident(t, cs, cs_))
                        goto nope;
                    t = t->next;
                    break;

                case 'K':
                    conp = NULL;
                case 'J':
                    temp = -1;                      // no reg. # forced
                    if (conp) {
                        if (conp->minus || conp->number < 1
                                || conp->number > 511) {
                            gpa("%e invalid register number\n", lnum);
                            ++errors;
                        }
                        else
                            temp = conp->number;
                    }
                    if (!as_search2(keeps, (ulong) id, (ulong) scope)) {
                        // For documentation generation.
                        say_str(scope->name, kept_vars);
                        fputs("::", kept_vars);
                        say_str(id, kept_vars);
                        fputc('\n', kept_vars);
                        fflush(kept_vars);

                        appear(keep, "Keep");       // new entry, register fmt
                        keep->name = id;
                        keep->scope = scope;
                        keep->lnum = lnum;
                        keep->force_reg = temp;
                        as_append(&keeps, keep);
                    }
                    else {
                        // Parsing continues on same line.
                        gpa("%eregister %O::%$ already kept\n", lnum, scope, id);
                        ++errors;
                    }
                    break;

                case 'l':
                    label = as_search2(labels, (ulong) id, (ulong) scope);
                    if (!label) {
                        msg = "label not defined";
                        msg_ = NULL;
                        goto nope;
                    }
                    instr = label->lnum;
                    break;

                case 'L':
                    if (as_search2(labels, (ulong) id, (ulong) scope)) {
                        gpa("%elabel %O::%$ already exists\n", lnum, scope, id);
                        ++errors;
                        goto nope;
                    }
                    appear(label, "Label");         // create new label
                    label->name = id;
                    label->scope = scope;
                    label->lnum = lnum;
                    as_append(&labels, label);
                    break;

                case 'm':
                    label = as_search2(labels, (ulong) id, (ulong) scope);
                    if (!label)
                        asfail("label missing in pass2");
                    label->code_addr = n_code;
                    break;

                case 'O':
                case 'o':
                    m = outright(&t, scope, c == 'O', signage_want,
                            &temp, &reg, &signage_eff);
                    signage_want = 0;

                    // Four cases from outright() need treated:
                    if (m) {
                        if (*m) {
                            // 1. returned string - it's an error message
                            msg = m, msg_ = NULL;
                            goto nope;
                        }
                        // 2. returned "" - register was successfully parsed
                        reg_no3 = reg_no2, reg_no2 = reg_no1, reg_no1 = temp;
                        break;
                    }
                    // 3. returned NULL, and a default error msg was set
                    // 4. returned NULL - not an error, try next rule
                    goto nope;

                case 'p':
                    push_code((ulong) oc_C << 27 | 0777777777);
                    break;

                case 'P':
                    temp = str_to_opcode((Str *) id);
                    if (temp < 0)
                        goto nope;
                    push_code((ulong) oc_IMH << 27 | reg_no1 << 18 | temp << 9);
                    break;

                case 'q':
                    goto quit;

                case 's':
                    if (t->tag != as_sym)
                        goto nope;
                    if (!same_str((Str *) t->payload, cs, cs_))
                        goto nope;                    
                    t = t->next;
                    break;

                case 'S':
                    signage_want = 's';
                    break;

                case 'U':
                    signage_want = 'u';
                    break;

                case 'W':
                    signage_want = 'w';
                    break;

                case 'x':
                    if (t->tag == as_ident) {
                        id = (Str *) t->payload;
                        temp = str_to_opcode((Str *) id);
                        if (temp < 0)
                            goto nope;
                    }
                    else if (t->tag == as_const) {
                        conp = (Const *) t->payload;
                        if (conp->minus || conp->number > 511)
                            goto nope;
                        temp = conp->number;
                    }
                    else
                        goto nope;
                    instr = instr << 9 | temp;
                    t = t->next;
                    break;

                case 'X':
                    push_code(instr & 0777777777777L);
                    break;

                default:
                    printf("ASSYN BUG - ASCII %i in rule\n", c);
                    goto fatal;
            }
        }

        printf("ASSYN BUG - rule without outcome\n");
        goto fatal;

nope:
        if (msg) {
            gpa("%e%:\n", lnum, msg, msg_);
            ++errors;
            cs = "$";
            cs_ = NULL;
            goto new_state;
        }
        t = rewind;
    }

fatal:
    ++errors;                           // no longer zero

quit:
    if (call_log)
        fclose(call_log);
    if (kept_vars)
        fclose(kept_vars);
    return errors;                      // possibly zero
}

/*
    Assign a sequential or a given register number.

    Input 'want':

        0 through 511       assign this specific register number
        -1                  assign next available register number
        -2                  forget all assignments (start fresh)

    There is no possibility that sequential assignment will preclude
    a later specific assignment, because betwixt() calls this with
    the kept, specific assignments first.

    Note that specific numbers are permitted to duplicate.
    That's why we offer specific numbers to be assigned.
*/
static int assign_reg(int want)
{
    static int next = 0;
    static char have[512];

    // Start fresh if want == -2.
    if (want == -2) {
        memset(have, 0, sizeof(have));
        next = 0;
        return 0;                       // return value ignored
    }

    // Assign next open number if want == -1.
    if (want == -1) {
        while (next < 512 && have[next])
            ++next;
        want = next++;
    }

    // Ensure a valid register number. If not valid, we pin 'next'
    // to an out-of-range result that betwixt() will check at end.
    if (want >= 512)
        return next = 2048;

    // Assign the requested number and mark it as taken.
    have[want] = 1;
    return want;
}

/*
    Between passes 1 and 2:

    - flag kept registers (esp. detecting names that don't exist)
    - assign registers to non-immediate constants
    - assign register numbers

    Returns 0 iff success.
*/
static int betwixt(void)
{
    Keep *k = keeps;
    Reg *r;
    Const *c, *d;
    int errors = 0;

    assign_reg(-2);                     // Initialize pool of register numbers.

    if (k)
        do {
            k = k->next;
            r = as_search2(regs, (ulong) k->name, (ulong) k->scope);
            if (!r) {
                gpa("line %i: register %$ does not exist\n", k->lnum, k->name);
                errors = 1;
            }
            else {
                r->flags |= as_kept;
                if (k->force_reg > 0)   // Specifically requested register #?
                    r->reg_no = assign_reg(k->force_reg);
            }
        } while (k != keeps);

    // It's important that we do constants first, because the assembler
    // depends on the 0 constant being in register 0.
    if (consts) {

        // positive constants go first
        c = consts;
        do {
            c = c->next;
            if (c->minus || c->reg_no == as_immed_only)
                continue;
            c -> reg_no = assign_reg(-1);
        } while (c != consts);

        // negative constants go now
        c = consts;
        do {
            c = c->next;
            if (!c->minus || c->reg_no == as_immed_only)
                continue;

            // Registers with the same BITS, like ......`t and -1, can share.
            d = as_search2(consts, c->number, 0);
            if (d && d->reg_no != as_immed_only)
                c->reg_no = d->reg_no;
            else
                c -> reg_no = assign_reg(-1);

        } while (c != consts);
    }

    if (r = regs)
        do {
            r = r->next;
            if (!r -> reg_no)
                r -> reg_no = assign_reg(-1);
        } while (r != regs);

    if (assign_reg(-1) > 512) {
        printf("  Error(registers exhausted)\n");
        errors = 1;
    }

    return errors;
}

/*
    After pass 2, resolve branches to scopes and labels.
*/
static void resolve_branches(void)
{
    Code *c = code;
    Label *label;
    Scope *scope;
    int opcode;
    int addr;
    int dest;

    if (!c)
        asfail("resolve code");             // code should have existed

    do {
        c = c->next;
        opcode = c->instr >> 27;
        addr = c->instr & 0777777777;

        // CALL and unconditional JUMP can be to either scopes or labels,
        // although their syntax varies. This means both namespaces are
        // searched here. But there is an ambiguity problem. FIXME TODO

        // scopes
        if (opcode == oc_J || opcode == oc_C || opcode >= oc_MIN_BRANCH_PRIV
                && opcode <= oc_MAX_BRANCH_PRIV && opcode != oc_CALI) {

            // addr is a line number, but needs to be a code address.
            if (scope = scopes)
                do {
                    scope = scope->next;
                    if (scope->lnum == addr) {
                        c->instr = (ulong) opcode << 27
                                | scope->code_addr + boot_code_offset;
                        goto updated;
                    }
                } while (scope != scopes);
        }

        // labels
        if (opcode >= oc_MIN_BRANCH_NPRIV && opcode <= oc_MAX_BRANCH_NPRIV
                || opcode == oc_CALI) {

            // addr is a line number, but needs to be a code address.
            if (label = labels)
                do {
                    label = label->next;
                    if (label->lnum == addr) {
                        c->instr = (ulong) opcode << 27
                                | label->code_addr + boot_code_offset;
                        goto updated;
                    }
                } while (label != labels);

            asfail("resolve label");        // label wasn't found
        }

        // for the operating system: support non-constant branches
        if (opcode == oc_asm_addr) {

            // contains scope/label line number, but needs to be code address
            dest = c->instr >> 18 & 511;    // split addr into dest & line #
            addr &= 0777777;                // bad choice: this is a line #

            if (scope = scopes)
                do {
                    scope = scope->next;
                    if (scope->lnum == addr) {
                        addr = scope->code_addr;    // now it's a code addr.
yip:                    c->instr = (ulong) oc_IMP << 27 | dest << 18
                                | addr + boot_code_offset;
                        goto updated;
                    }
                } while (scope != scopes);

            if (label = labels)
                do {
                    label = label->next;
                    if (label->lnum == addr) {      // now it's a code addr.
                        addr = label->code_addr;
                        goto yip;
                    }
                } while (label != labels);

            asfail("resolve oc_asm_addr");  // should not occur
        }

updated:;
    } while (c != code);        
}
