/*
    decoder.c
    Using the instruct.c database, fill in the instruction decoder SRAMs.
    The various CPU cycle types are implemented below.
*/

/*
    The decoder RAMs are 512 instructions * 4 steps, and are 72 bits
    (2 36-bit RAMs) wide. The following constants are too large for
    some compilers' enum keyword, so they are macros here. By or-ing
    these together appropriately, we form instruction decoder words.
    For what these all mean, refer to the architecture documentation.
*/
#define V(bit, val) (val ## L << bit)

// Lockout A is intriguing from a security perspective in the sense that
// either alpha or ff m can bypass the page table. So some combinations are
// "dual use".
#define lo_a_A  V(11, 0)                // dual use
#define lo_a_m  V(11, 3)                // dual use
#define lo_a_P  V(11, 1)
#define lo_a_Ah V(11, 2)                // dual use

#define lo_b_B    V(13, 0)              // XXX see question on spreadsheet here
#define lo_b_Mw   V(13, 2)
#define lo_b_M    V(13, 1)
#define priv_lo_b_io   V(13, 3)

#define lo_c_C    V(15, 0)
#define priv_lo_c_Cw   V(15, 2)
#define priv_lo_c_xany V(15, 1)
#define lo_c_dest V(15, 3)

#define lo_e_E   V(0, 1)                // BEWARE: lo_e_* must go to D1, never D0.
#define lo_e_Ews V(0, 0)
#define lo_e_Ewe V(0, 2)

#define lo_g_G    V(17, 0)
#define lo_g_LR   V(17, 2)
#define lo_g_Gn   V(17, 1)
#define lo_g_LRh  V(17, 3)

#define priv_lo_ip_priv   V(19, 0)      // FIXME bad choice: privileged by omission
#define lo_ip_return V(19, 2)
#define lo_ip_next   V(19, 1)
#define lo_ip_jump   V(19, 1)

#define lo_s_S  V(34, 1)
#define lo_s_Sh V(34, 3)
#define lo_s_Sw V(34, 0)

#define lo_z_Zh   V(26, 0)
#define lo_z_Z    V(26, 1)
#define lo_z_revert V(26, 2)

#define lo_t_T    V(22, 0)
#define lo_t_Th   V(22, 1)

#define priv_save_inst V(31, 1)

#define alpha_beta_gamma V(0, 1)

#define theta_zeta V(6, 1)

#define t_flip_allow V(28, 1)

#define priv_p_write V(21, 1)           // this bit gets inverted

#define priv_clock_ffu V(23, 1)

#define priv_clock_timer V(33, 1)

#define priv_priv_mode V(24, 1)
#define priv_setup_mode V(24, 2)
#define priv_npriv_mode V(24, 3)

#define hijack_decoder V(30, 1)

#define priv_preset_call_depth V(2, 1)  // BEWARE: must go to D1, never D0.

#define priv_clock_ffw V(5, 1)          // BEWARE: must go to D1, never D0.

#define priv_skip_ip_increment V(3, 1)  // BEWARE: must go to D1, never D0.

#define beta_hold V(7, 1)               // BEWARE: must go to D1, never D0.

#define advance_M0_M1 V(4, 1)           // BEWARE: must go to D1, never D0.

// Below are for the electrical simulator only.
#define priv_virt_memset V(33, 1)       // BEWARE: must go to D1, never D0.
#define strobe_pvio V(34, 1)            // BEWARE: must go to D1, never D0.
#define priv_halt_cpu V(35, 1)          // BEWARE: must go to D1, never D0.

/*
    Given a pointer to a 4-word instruction slot within decoder SRAM D0,
    fill in the firmware for that opcode. This is the more-full of the
    two decoder RAMs.

    Many instructions have lo_ip_jump at click 3 in order to have
    ff a's clock low. This lets us snatch the desired code addresss in
    ff a via a rising clock in click 0 for instructions like wcm (write
    code memory).
*/
static void fill_inst_0(long *m, struct instruction *in)
{
    int i;

    // We change this value mid-cycle in branch instruction in order
    // to stand down on the branch decision. Otherwise the next instruction
    // (even a NOP) will branch.
    int in_zeta = in->zeta;

    // Here we fill in the decoder bits for control and sequencing.
    switch (in->cycle) {
        default:
            ng("!the_insts insane");

        case cy_alu:    // ALU operation
        case cy_pvi:    // PVIO (differs from ALU in D1)
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_c:      // call
            // Same as cy_cali for D0. This derives from cy_j.
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * zeta_j;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | lo_s_Sw;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_j:      // jump
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * in_zeta;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | lo_s_S;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_ld:     // load memory (honors page table)
            // This derives from cy_rdm.
            m[0] = lo_a_P | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_P | lo_b_M | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_lds:    // exchange memory with transposed register
            // This derives from cy_rdm.
            // Warning: Alpha read on click 1 with beta control signals.
            //          Depends on alpha_xl == beta_id (see slots.h).
            //          Sanity check written into (firmware.c).
            // Warning: Beta read on click 2 with gamma control signals.
            //          Depends on beta_tr == gamma_add_more (see slots.h).
            //          Sanity check written into (firmware.c).
            m[0] = lo_a_P | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_P | lo_b_M | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_now:    // no-write ALU operation: cy_alu except lo_g_Gn
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_r:      // return
            // This derives from cy_c.
            m[0] = lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest | lo_ip_return | lo_s_S;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_rvt:    // revert
            // This derives from cy_r.
            // There isn't enough time to attach priv/nonpriv condition to
            // this instruction, which might have saved a CPU cycle.
            m[0] = lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_z_revert | lo_g_LRh | lo_c_dest | lo_ip_return | lo_s_S;
            m[2] = lo_z_Z | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_st2:    // transposed double memory write (honors page table)
            // This derives from cy_wd2.
            // Warning: Alpha read on click 1 with beta control signals.
            //          Depends on alpha_cmp_us == beta_tr (see slots.h).
            //          Sanity check written into (firmware.c).
            // Warning: Beta read on click 2 with gamma control signals.
            //          Depends on beta_tr == gamma_add_more (see slots.h).
            //          Sanity check written into (firmware.c).
            m[0] = lo_a_P | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_P | lo_b_Mw | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_sto:    // store memory (honors page table)
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_P | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_yld:    //  set dc.d130 to trigger control decoder hijack
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S | hijack_decoder;
            break;

        case cy_priv_csi:    // cali
            // Same as cy_c for D0. This derives from cy_j.
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * zeta_j;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | lo_s_Sw;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_priv_hij:    //  decoder hijack
            // This derives from cy_call, but skips the branch. That means
            // adjusting lo_ip and removing zeta's branch force. After that,
            // the CPU goes priv.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_z_Z | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_next | lo_s_Sw | priv_priv_mode;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_priv_hlt:    //  halt: cy_now except priv_halt_cpu via D1
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_imm:    // load immediate / privileged read code memory 2
            // Warning: Alpha read on click 1 with beta control signals, but
            //          the result is ignored.
            m[0] = lo_a_m | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_A | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_ja:     //  jump to any location
            // This derives from cy_j.
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * in_zeta;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | priv_lo_ip_priv | lo_s_S;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_priv_npc:    // npcall = call + npriv
            // This derives from cy_c.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * zeta_j;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | lo_s_Sw | priv_npriv_mode;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_priv_pca:    // pcall = call + priv
            // This derives from cy_c.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_z_Z | lo_g_LRh 
                 | lo_c_dest | lo_s_S | theta_zeta * zeta_j;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | lo_s_Sw | priv_priv_mode;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            in_zeta = zeta_nop;
            break;

        case cy_priv_pek:    //  copy user register to superuser register
            // This derives from cy_alu.
            // For clock pulse separation, mode changes are never on click 3.
            // CPU must already be in *npriv* mode to use PEEK.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S | priv_priv_mode;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_G | lo_c_C | lo_s_S | priv_npriv_mode;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_pok:    //  copy superuser register to user register
            // This derives from cy_alu.
            // For clock pulse separation, mode changes are never on click 3.
            // CPU must already be in *priv* mode to use POKE.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S | priv_npriv_mode;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_G | lo_c_C | lo_s_S | priv_priv_mode;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_prv:     //  enter privileged mode
            // This derives from cy_now.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T | priv_priv_mode
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;
 
        case cy_priv_pts:     //  enter neither mode (setup)
            // This derives from cy_now.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T | priv_setup_mode
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_rcm:    //  read code memory 1
            // This derives from cy_wcm.
            m[0] = priv_lo_ip_priv | lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_C | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_rdm:    //  read data memory (bypasses page table)
            // This derives from cy_alu.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_M | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_rpt:    //  read page table
            // This derives from cy_alu.
            m[0] = lo_a_P | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_P | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_rwm:    //  exchange memory with transposed register
                        // (bypasses page table)
            // This derives from cy_rdm.
            // Warning: Alpha read on click 1 with beta control signals.
            //          Depends on alpha_xl == beta_id (see slots.h).
            //          Sanity check written into (firmware.c).
            // Warning: Beta read on click 2 with gamma control signals.
            //          Depends on beta_tr == gamma_add_more (see slots.h).
            //          Sanity check written into (firmware.c).
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_A | lo_b_M | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_tim:    //  timer setpoint serial I/O
            // This derives from cy_alu and uses unsigned register doubling.
            // The bit pattern to enter is 10_XXXX_XXXX_XXXX.
            // Output is via beta[17] as shown:
            //
            // imm bits 10_XXXX_XXXX_XXXX`2
            // serial.io.loop:
            //     timer bits
            //     j >= serial.io.loop
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S | priv_clock_timer;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_G | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_uid:     //  set user id
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S | priv_clock_ffu;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;
 
        case cy_priv_usr:     //  enter user mode
            // This derives from cy_now.
            // For clock pulse separation, mode changes are never on click 3.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T | priv_npriv_mode
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;
 
        case cy_priv_wcm:    //  write code memory
            // This derives from cy_now.
            m[0] = priv_lo_ip_priv | lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | priv_lo_c_Cw | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_wd2:    //  transposed double memory write
            // This derives from cy_wdm.
            // Warning: Alpha read on click 1 with beta control signals.
            //          Depends on alpha_cmp_us == beta_tr (see slots.h).
            //          Sanity check written into (firmware.c).
            // Warning: Beta read on click 2 with gamma control signals.
            //          Depends on beta_tr == gamma_add_more (see slots.h).
            //          Sanity check written into (firmware.c).
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_A | lo_b_Mw | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_wdm:    //  write data memory (bypasses page table)
            // This derives from cy_rdm.
            // Warning: Alpha read on click 1 with beta control signals.
            //          Depends on alpha_xl == beta_id (see slots.h).
            //          Sanity check written into (firmware.c).
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_A | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_b_Mw | lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_wpt:    //  write page table
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S | priv_p_write;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_z_Zh | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | lo_s_S;
            break;

        case cy_priv_xa:     //  execute any instruction
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | priv_lo_c_xany | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;        

        case cy_priv_set:    // memset via simulator
            // This derives from cy_now.
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh | lo_t_T
                 | lo_g_LRh | lo_c_dest | lo_ip_next | lo_s_S;
            m[2] = lo_a_Ah | lo_z_Zh | lo_t_Th | lo_g_Gn | lo_c_C | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Z | lo_g_LR | lo_c_dest | lo_s_S;
            break;
    }

    // Now we fill in operations for the ALU SRAMs.
    // We also indicate how the incoming T flag is seen by the ALU.
    // The cycle when t_flip_allow is inserted affects how alu.c is written,
    // so watch for that if you make changes.
    // XXX There is probably a simpler way than this / 2 scheme for
    //     understanding how theta works and its slots. Same for gamma.
    //     An issue is that in gamma's case, the hardware overwrites
    //     bit 0 of the operation with the carry decision, but with
    //     theta, there are 5 bits in the control decoder (not 6), so
    //     we shift them left to insert the adjusted T flag.
    m[0] |= alpha_beta_gamma * in->alpha;
    m[1] |= alpha_beta_gamma * in->beta  | theta_zeta * (in->theta / 2)
                | t_flip_allow * in->t_vision;
    m[2] |= alpha_beta_gamma * in->gamma;
    m[3] |= theta_zeta * in_zeta;

/*
    Not only does ff m retains its value during context switches, but it
    is directly readable by RCM2 (privileged) and IMB/IMH/IMN/IMP
    (nonprivileged). The IM* instructions are made safe for nonprivileged
    users by the fact that every opcode except RCM1 (which is privileged)
    copies the next instruction being fetched to ff m during click 3. So
    if a normal program without access to RCM2 uses ff m via IMB/IMH/IMN/
    IMP, it will see its own presently executing instruction.

    For this to work out:

    1. priv_save_inst must be mandatory on click 3 for all instructions,
       except for RCM1.

    2. Never follow RCM1 with IMB/IMH/IMN/IMP in privileged code, because 
       ff m will contain the wrong data. Follow RCM1 with RCM2.

    3. RCM2 only works if it's immediately preceded by RCM1 without any
       possibility of a context switch.
*/
    if (in->cycle == cy_priv_rcm)
        m[2] |= priv_save_inst;         //  use
    else
        m[3] |= priv_save_inst;         // nonprivileged use

    // priv_p_write needs to be inverted.
    for (i=0; i<4; i++)
        m[i] ^= priv_p_write;
}

/*
    Given a pointer to a 4-word instruction slot within decoder SRAM D1,
    fill in the firmware for that opcode. This is the less-full of the
    two decoder RAMs.
*/
static void fill_inst_1(long *m, struct instruction *in)
{
    int i;

    switch (in->cycle) {
        default:
            ng("!the_insts insane");

        case cy_alu:        // ALU operation
        case cy_j:          // jump
        case cy_ld:         // load memory (honors page table)
        case cy_now:        // no-write ALU operation
        case cy_yld:        //  yield via hardware
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_c:          // call
            m[0] = lo_e_Ews;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_r:          // return
        case cy_rvt:        // revert
            m[0] = lo_e_Ewe;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_st2:        // transposed double memory write
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E | advance_M0_M1;
            m[3] = lo_e_E;
            break;

        case cy_sto:        // store memory (honors page table)
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E | beta_hold;
            m[3] = lo_e_E;
            break;


        case cy_priv_rcm:   // read code memory 1
        case cy_lds:        // transposed register/memory exchange
        case cy_priv_uid:   // set user id
        case cy_priv_rdm:   // read data memory (bypasses page table)
        case cy_priv_wpt:   // write page table
        case cy_priv_rpt:   // read page table
        case cy_priv_ja:    // jump to any location
        case cy_priv_prv:   // enter privileged mode
        case cy_priv_usr:   // enter user mode
        case cy_priv_pts:   // enter neither mode (setup)
        case cy_priv_pek:   // copy user register to superuser register
        case cy_priv_pok:   // copy superuser register to user register
        case cy_priv_rwm:   // transposed register/memory exchange
        case cy_priv_tim:   // timer setpoint serial I/O
        case cy_priv_imm:   // load immediate / privileged read code memory 2
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_priv_pca:   // pcall
        case cy_priv_npc:   // npcall
        case cy_priv_hij:   // decoder hijack
            m[0] = lo_e_Ews;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_priv_wd2:   // transposed double memory write
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E | advance_M0_M1;
            m[3] = lo_e_E;
            break;

        case cy_priv_xa:    // execute any instruction
            m[0] = lo_e_E | priv_clock_ffw;
            m[1] = lo_e_E;
            m[2] = lo_e_E | priv_skip_ip_increment;
            m[3] = lo_e_E;
            break;

        case cy_priv_wdm:   // write data memory (bypasses page table)
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E | beta_hold;
            m[3] = lo_e_E;
            break;

        case cy_priv_hlt:   // halt: cy_now except priv_halt_cpu via D1
            m[0] = lo_e_E;
            m[1] = lo_e_E | priv_halt_cpu;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_pvi:        // pvio: cy_alu except strobe_pvio via D1
            m[0] = lo_e_E;
            m[1] = lo_e_E | strobe_pvio;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_priv_wcm:   // write code memory
            m[0] = lo_e_E | priv_clock_ffw;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_priv_csi:   // call stack initialize
            m[0] = lo_e_Ews | priv_preset_call_depth;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_priv_set:   // memset via simulator
            m[0] = lo_e_E | priv_virt_memset;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;
    }

    // priv_skip_ip_increment needs to be inverted.
    for (i=0; i<4; i++)
        m[i] ^= priv_skip_ip_increment;
}

static struct instruction *find_inst(int opcode)
{
    int i;
    for (i = 0; i < CARD(the_insts); i++)
        if (the_insts[i].opcode == opcode)
            return the_insts + i;
    ng("!find_inst missing opcode %i\n", opcode);
    return NULL;
}

/*
    This loads instruction decoder SRAM D0 or D1.
    The caller already allocated memory.
*/
static void load_decoder(long *mem, int rows, int chip)
{
    void (*filler)(long *mem, struct instruction *in) =
            chip ? fill_inst_1 : fill_inst_0;

    static struct instruction *nop, *revert, *hijack, *inst;
    int i, j, opcode;

    if (rows != 4 * 4 * 512)
        ng("!decoder.c and firmware.c conflict");

    // Find the NOP instruction if we haven't yet.
    if (!nop) {
        nop = find_inst(oc_NOP);
        revert = find_inst(oc_REVERT);
        hijack = find_inst(oc_HIJACK);
    }

    // Pre-fill entire table with NOP, because some opcodes aren't used
    // and we can't let them do insecure things.
    for (i = 0; i < 2048; i++)
        filler(mem + 4 * i, nop);

    // Fill in entries that correspond to opcodes. There are four quadrants:
    for (i = 0; i < CARD(the_insts); i++) {

        opcode = the_insts[i].opcode;

        // Quadrant 0 is the normal instruction.
        inst = the_insts + i;
        filler(mem + 4 * opcode, inst);

        // Quadrant 1 is the normal instruction, but missing the IP increment.
        // It's the first instruction executed via hijacking.
        inst = the_insts + i;
        filler(mem + 4 * opcode + 2048, inst);
        if (chip) {
            for (j=0; j<4; j++)
                mem[4 * opcode + 2048 + j] &= ~priv_skip_ip_increment;
        }

        // Quadrant 3 is the second instruction executed via hijacking,
        // because the counter circuit minimized using a Gray code.
        // This instruction, like CALL, pushes registers and flags on
        // the user stack, but it doesn't bother to branch. It also
        // leaves the CPU in priv mode. It doesn't matter now that the
        // instruction pointer diverged from where we want to resume,
        // because it's about to be forgotten.
        inst = the_insts + i;
        filler(mem + 4 * opcode + 6144, hijack);

        // Quadrant 2 REVERTs to the superuser program. It happens third.
        inst = the_insts + i;
        filler(mem + 4 * opcode + 4096, revert);
    }
}

#undef V
