/*
    instruct.c
    Opcode database from which the instruction decoder is fed
*/

/*
    These are the architecture's CPU cycle types. Each opcode is exactly
    one of these types, but some types are used by more than one opcode.
    The cycle type is used to compute what the instruction decoder needs
    to do in terms of handling the CPU's SRAMs, flip-flops, and some of
    the glue logic. Most of this decoding leads to output enables,
    clock enables, etc.

    The implementation of each cycle type is in decoder.c.
*/
enum {
    // nonprivileged
    cy_alu,     // ALU operation
    cy_c,       // call
    cy_j,       // jump
    cy_ld,      // load memory (honors page table)
    cy_lds,     // load AND store (honors page table)
    cy_now,     // no-write ALU operation
    cy_r,       // return
    cy_rvt,     // revert
    cy_st2,     // double store memory (honors page table)
    cy_sto,     // store memory (honors page table)
    cy_yld,     // yield (strobes dc.d130)
    cy_pvi,     // paravirtualized I/O (strobes dc.d145)

    // privileged
    cy_priv_csi,     // call stack initialize
    cy_priv_hij,     // call-no-branch + priv (yield to superuser)
    cy_priv_hlt,     // halt
    cy_priv_imm,     // load immediate
    cy_priv_ja,      // jump to any location
    cy_priv_npc,     // npcall = call + npriv
    cy_priv_pca,     // pcall = call + priv
    cy_priv_pek,     // copy user reg to superuser reg while npriv
    cy_priv_pok,     // copy superuser reg to user reg while priv
    cy_priv_prv,     // enter mode
    cy_priv_pts,     // enter neither mode (page table & stack setup)
    cy_priv_rcm,     // read code memory
    cy_priv_rdm,     // read data memory (bypasses page table)
    cy_priv_rpt,     // read page table
    cy_priv_rwm,     // read AND write data memory (bypasses page table)
    cy_priv_set,     // memset within simulator
    cy_priv_tim,     // timer setpoint serial I/O
    cy_priv_uid,     // set user id
    cy_priv_usr,     // enter user mode
    cy_priv_wcm,     // write code memory
    cy_priv_wd2,     // double write data memory (bypasses page table)
    cy_priv_wdm,     // write data memory (bypasses page table)
    cy_priv_wpt,     // write page table
    cy_priv_xa,      // execute any instruction
};

/*
    The ALU can do four things with the incoming T flag.
    This is not the new value of T, but how the existing T is seen.
*/
enum {
    t_allow = 2,                // This bit is ANDed with the incoming T.
    t_flip = 1,                 // The AND result is XORed with this bit.

    t_0 = 0,                    // replace T with 0
    t_1 = t_flip,               // replace T with 1
    t_id = t_allow,             // leave T as-is
    t_inv = t_flip | t_allow,   // invert T
};

/*
    These are the architecture's opcodes. Most columns here represent
    address inputs to the ALU's many SRAMs. The instruction decoder SRAM
    is computed from this table.
*/
static struct instruction {
    // These apply to all instructions.
    char *name;         // assembler name (footnote N)
    int opcode;         // opcode (oc_*)
    int cycle;          // cycle sequence (cy_*)

    // These apply only to ALU instructions.
    int alpha;          // ALU first operation (alpha_*)
    int beta;           // ALU second operation (beta_*)
    int gamma;          // ALU third operation (gamma_*)
    int theta;          // ALU theta operation (theta_*)
    int zeta;           // ALU zeta operation (flag scheme; zeta_*)
    int t_vision;       // ALU's vision of incoming T flag
} the_insts[] = {
#define X(name, op, cy, a, b, g, th, z, t) name, oc_ ## op, cy_ ## cy, \
        alpha_ ## a, beta_ ## b, gamma_ ## g, theta_ ## th, zeta_ ## z, t_ ## t,
//    name       opcode  cycle  alpha beta gamma theta zeta    T
    X("a.uuu"  , A_uuu  , alu, add_uu, id , add, add, add_u  , 0  )
    X("a.uus"  , A_uus  , alu, add_uu, id , add, add, add_s  , 0  )
    X("a.usu"  , A_usu  , alu, add_us, id , add, add, add_u  , 0  )
    X("a.uss"  , A_uss  , alu, add_us, id , add, add, add_s  , 0  )
    X("a.suu"  , A_suu  , alu, add_su, id , add, add, add_u  , 0  )
    X("a.sus"  , A_sus  , alu, add_su, id , add, add, add_s  , 0  )
    X("a.ssu"  , A_ssu  , alu, add_ss, id , add, add, add_u  , 0  )
    X("a.sss"  , A_sss  , alu, add_ss, id , add, add, add_s  , 0  )

    X("s.uuu"  , S_uuu  , alu, sub_uu, id , add, add, add_u  , 1  )
    X("s.uus"  , S_uus  , alu, sub_uu, id , add, add, add_s  , 1  )
    X("s.usu"  , S_usu  , alu, sub_us, id , add, add, add_u  , 1  )
    X("s.uss"  , S_uss  , alu, sub_us, id , add, add, add_s  , 1  )
    X("s.suu"  , S_suu  , alu, sub_su, id , add, add, add_u  , 1  )
    X("s.sus"  , S_sus  , alu, sub_su, id , add, add, add_s  , 1  )
    X("s.ssu"  , S_ssu  , alu, sub_ss, id , add, add, add_u  , 1  )
    X("s.sss"  , S_sss  , alu, sub_ss, id , add, add, add_s  , 1  )

    X("rs.uuu" , RS_uuu , alu, rev_uu, id , add, add, add_u  , 1  )
    X("rs.uus" , RS_uus , alu, rev_uu, id , add, add, add_s  , 1  )
    X("rs.usu" , RS_usu , alu, rev_us, id , add, add, add_u  , 1  )
    X("rs.uss" , RS_uss , alu, rev_us, id , add, add, add_s  , 1  )
    X("rs.suu" , RS_suu , alu, rev_su, id , add, add, add_u  , 1  )
    X("rs.sus" , RS_sus , alu, rev_su, id , add, add, add_s  , 1  )
    X("rs.ssu" , RS_ssu , alu, rev_ss, id , add, add, add_u  , 1  )
    X("rs.sss" , RS_sss , alu, rev_ss, id , add, add, add_s  , 1  )

    X("ac.uuu" , AC_uuu , alu, add_uu, id , add, add, add_u  , id )
    X("ac.uus" , AC_uus , alu, add_uu, id , add, add, add_s  , id )
    X("ac.usu" , AC_usu , alu, add_us, id , add, add, add_u  , id )
    X("ac.uss" , AC_uss , alu, add_us, id , add, add, add_s  , id )
    X("ac.suu" , AC_suu , alu, add_su, id , add, add, add_u  , id )
    X("ac.sus" , AC_sus , alu, add_su, id , add, add, add_s  , id )
    X("ac.ssu" , AC_ssu , alu, add_ss, id , add, add, add_u  , id )
    X("ac.sss" , AC_sss , alu, add_ss, id , add, add, add_s  , id )

    X("sc.uuu" , SC_uuu , alu, sub_uu, id , add, add, add_u  , inv)
    X("sc.uus" , SC_uus , alu, sub_uu, id , add, add, add_s  , inv)
    X("sc.usu" , SC_usu , alu, sub_us, id , add, add, add_u  , inv)
    X("sc.uss" , SC_uss , alu, sub_us, id , add, add, add_s  , inv)
    X("sc.suu" , SC_suu , alu, sub_su, id , add, add, add_u  , inv)
    X("sc.sus" , SC_sus , alu, sub_su, id , add, add, add_s  , inv)
    X("sc.ssu" , SC_ssu , alu, sub_ss, id , add, add, add_u  , inv)
    X("sc.sss" , SC_sss , alu, sub_ss, id , add, add, add_s  , inv)

    X("rsc.uuu", RSC_uuu, alu, rev_uu, id , add, add, add_u  , inv)
    X("rsc.uus", RSC_uus, alu, rev_uu, id , add, add, add_s  , inv)
    X("rsc.usu", RSC_usu, alu, rev_us, id , add, add, add_u  , inv)
    X("rsc.uss", RSC_uss, alu, rev_us, id , add, add, add_s  , inv)
    X("rsc.suu", RSC_suu, alu, rev_su, id , add, add, add_u  , inv)
    X("rsc.sus", RSC_sus, alu, rev_su, id , add, add, add_s  , inv)
    X("rsc.ssu", RSC_ssu, alu, rev_ss, id , add, add, add_u  , inv)
    X("rsc.sss", RSC_sss, alu, rev_ss, id , add, add, add_s  , inv)

    // FIXME MWA XXX Check T on these wrapping insts. 12Sep2022
    X("a.w"    , A_w    , alu, add_uu, id , add, add, wrap   , 0  )
    X("ac.w"   , AC_w   , alu, add_uu, id , add, add, wrap   , inv)
    X("s.w"    , S_w    , alu, sub_uu, id , add, add, wrap   , 1  )
    X("sc.w"   , SC_w   , alu, sub_uu, id , add, add, wrap   , id )
    X("rs.w"   , RS_w   , alu, rev_uu, id , add, add, wrap   , 1  )
    X("rsc.w"  , RSC_w  , alu, rev_uu, id , add, add, wrap   , id )

    X("igf"    , IGF    , alu, xl    , id , kil, t  , logic  , 1  )
    X("igt"    , IGT    , alu, xr    , not, xor, t  , logic  , 1  )
    X("xl"     , XL     , alu, xl    , id , mux, t  , logic  , 0  )
    X("xr"     , XR     , alu, xl    , id , mux, t  , logic  , 1  )
    X("nl"     , NL     , alu, xl    , not, add, t  , logic  , 0  )
    X("nr"     , NR     , alu, xr    , not, add, t  , logic  , 0  )
    X("and"    , AND    , alu, and   , id , add, t  , logic  , 0  )
    X("nand"   , NAND   , alu, and   , not, add, t  , logic  , 0  )
    X("or"     , OR     , alu, or    , id , add, t  , logic  , 0  )
    X("nor"    , NOR    , alu, or    , not, add, t  , logic  , 0  )
    X("xor"    , XOR    , alu, xl    , id , xor, t  , logic  , 1  )
    X("xnor"   , XNOR   , alu, xl    , not, xor, t  , logic  , 1  )
    X("lanr"   , LANR   , alu, or    , id , xor, t  , logic  , 1  )
    X("ranl"   , RANL   , alu, and   , id , xor, t  , logic  , 1  )
    X("lonr"   , LONR   , alu, and   , not, xor, t  , logic  , 1  )
    X("ronl"   , RONL   , alu, or    , not, xor, t  , logic  , 1  )

    X("max.uuu", MAX_uuu, alu, cmp_uu, id , mux, lt , add_u  , 0  )
    X("max.uus", MAX_uus, alu, cmp_uu, id , mux, lt , add_s  , 0  )
    X("max.usu", MAX_usu, alu, cmp_us, id , mux, lt , add_u  , 0  )
    X("max.uss", MAX_uss, alu, cmp_us, id , mux, lt , add_s  , 0  )
    X("max.suu", MAX_suu, alu, cmp_su, id , mux, lt , add_u  , 0  )
    X("max.sus", MAX_sus, alu, cmp_su, id , mux, lt , add_s  , 0  )
    X("max.ssu", MAX_ssu, alu, cmp_ss, id , mux, lt , add_u  , 0  )
    X("max.sss", MAX_sss, alu, cmp_ss, id , mux, lt , add_s  , 0  )

    X("min.uuu", MIN_uuu, alu, cmp_uu, id , mux, gt , add_u  , 1  )
    X("min.uus", MIN_uus, alu, cmp_uu, id , mux, gt , add_s  , 1  )
    X("min.usu", MIN_usu, alu, cmp_us, id , mux, gt , add_u  , 1  )
    X("min.uss", MIN_uss, alu, cmp_us, id , mux, gt , add_s  , 1  )
    X("min.suu", MIN_suu, alu, cmp_su, id , mux, gt , add_u  , 1  )
    X("min.sus", MIN_sus, alu, cmp_su, id , mux, gt , add_s  , 1  )
    X("min.ssu", MIN_ssu, alu, cmp_ss, id , mux, gt , add_u  , 1  )
    X("min.sss", MIN_sss, alu, cmp_ss, id , mux, gt , add_s  , 1  )

    X("bound.u", BOUND_u, now, sub_uu, id , add, add, bound  , 1  )
    X("bound.s", BOUND_s, now, sub_us, id , add, add, bound  , 1  )

    X("cmp.uu" , CMP_uu , now, sub_uu, id , add, add, cmp    , 1  )
    X("cmp.us" , CMP_us , now, sub_us, id , add, add, cmp    , 1  )
    X("cmp.su" , CMP_su , now, sub_su, id , add, add, cmp    , 1  )
    X("cmp.ss" , CMP_ss , now, sub_ss, id , add, add, cmp    , 1  )

    X("asl.uu" , ASL_uu , alu, e36   , shl, rol, t  , asl_uu , 0  )
    X("asl.us" , ASL_us , alu, e35   , shl, rol, t  , asl_us , 0  )
    X("asl.su" , ASL_su , alu, e36   , shl, rol, t  , asl_su , 0  )
    X("asl.ss" , ASL_ss , alu, e35   , shl, rol, t  , asl_ss , 0  )

    X("asr.uu" , ASR_uu , alu, asr   , shr, rol, t  , asr_uu , 0  )
    X("asr.us" , ASR_us , alu, asr   , shr, rol, t  , asr_us , 0  )
    X("asr.su" , ASR_su , alu, asr   , shr, rol, c5 , asr_su , 0  )
    X("asr.ss" , ASR_ss , alu, asr   , shr, rol, c5 , asr_ss , 0  )

//    name       opcode  cycle  alpha beta gamma theta zeta    T
    X("lsl"    , LSL    , alu, xl    , shl, rol, t  , logic  , 0  )
    X("lsr"    , LSR    , alu, xl    , shr, rol, t  , logic  , 0  )
    X("rol"    , ROL    , alu, xl    , rol, rol, t  , logic  , 0  )

    X("ml"     , ML     , alu, ml    , id , add, t  , asr_uu , 0  )
    X("mh"     , MH     , alu, mh    , rtl, add, t  , asr_us , 0  )
    X("mhns"   , MHNS   , alu, mh    , id , add, t  , asr_uu , 0  )
    X("mhl"    , MHL    , alu, xl    , id , mhl, t  , nop    , 1  )
    X("mhl0"   , MHL0   , alu, xl    , sw0, mhl, t  , nop    , 1  )
    X("mhl1"   , MHL1   , alu, xl    , sw1, mhl, t  , nop    , 1  )
    X("mhl2"   , MHL2   , alu, xl    , sw2, mhl, t  , nop    , 1  )
    X("mhl3"   , MHL3   , alu, xl    , sw3, mhl, t  , nop    , 1  )
    X("mhl4"   , MHL4   , alu, xl    , sw4, mhl, t  , nop    , 1  )
    X("mhl5"   , MHL5   , alu, xl    , sw5, mhl, t  , nop    , 1  )
    X("dsl"    , DSL    , alu, dsl   , rtl, add, dsl, logic  , id )

    X("swiz"   , SWIZ   , alu, xl    , swz, add, t  , logic  , 0  )
    X("nudge"  , NUDGE  , alu, nud   , id , mux, 2r , logic  , 1  )
    X("ham2"   , HAM2   , alu, xl    , ham, ham, t  , logic  , 1  )
    X("txor"   , TXOR   , alu, xl    , txo, add, t  , logic  , 0  )

    X("pit"    , PIT    , alu, xl    , id , pit, t  , logic  , 0  )
    X("pat"    , PAT    , alu, xl    , pat, add, t  , logic  , 0  )
    X("pait"   , PAIT   , alu, xl    , pat, pit, t  , logic  , 0  )

    X("mix"    , MIX    , alu, mix   , mix, mix, t  , logic  , 0  )
    X("xim"    , XIM    , alu, xim   , xim, xim, t  , logic  , 1  )

    X("un.a"   , UNARY_A, alu, uny   , id , add, t  , logic  , 0  )
    X("un.b"   , UNARY_B, alu, xl    , uny, add, t  , logic  , 0  )
    X("un.g"   , UNARY_G, alu, xl    , id , uny, t  , logic  , 0  )
    X("stun.a" , STUN_A , alu, stu   , stu, stu, add, add_u  , 1  )
    X("stun.b" , STUN_B , alu, stu   , stu, stu, gt , logic  , 1  )
    X("stun.c" , STUN_C , alu, stu   , stu, stu, t  , asl_uu , 0  )
    X("stun.d" , STUN_D , alu, stu   , stu, stu, 2r , logic  , 1  )
    X("stun.e" , STUN_E , alu, stu   , stu, stu, 2l , logic  , 0  )
    X("stun.f" , STUN_F , alu, stu   , stu, stu, rev, t_adj  , 1  )
    X("stun.g" , STUN_G , alu, stu   , stu, stu, t  , t_adj  , 0  )
    X("stun.h" , STUN_H , alu, stu   , stu, stu, t  , t_adj  , id )
    X("stun.i" , STUN_I , alu, stu   , stu, stu, add, t_adj  , 1  )
    X("stun.j" , STUN_J , alu, stu   , stu, stu, t  , logic  , 0  )
    X("stun.k" , STUN_K , alu, stu   , stu, stu, t  , logic  , 1  )

    X("c"      , C      , c  , xl    , id , kil, t  , nop    , 1  )
    X("j"      , J      , j  , xl    , id , kil, t  , j      , 1  )
    X("j.lt"   , J_lt   , j  , xl    , id , kil, t  , j_lt   , 1  )
    X("j.le"   , J_le   , j  , xl    , id , kil, t  , j_le   , 1  )
    X("j.eq"   , J_eq   , j  , xl    , id , kil, t  , j_eq   , 1  )
    X("j.ne"   , J_ne   , j  , xl    , id , kil, t  , j_ne   , 1  )
    X("j.ge"   , J_ge   , j  , xl    , id , kil, t  , j_ge   , 1  )
    X("j.gt"   , J_gt   , j  , xl    , id , kil, t  , j_gt   , 1  )
    X("j.r"    , J_r    , j  , xl    , id , kil, t  , j_r    , 1  )
    X("j.nr"   , J_nr   , j  , xl    , id , kil, t  , j_nr   , 1  )
    X("j.t"    , J_t    , j  , xl    , id , kil, t  , j_t    , 1  )
    X("j.nt"   , J_nt   , j  , xl    , id , kil, t  , j_nt   , 1  )

    X("crf"    , CRF    , now, xl    , id , kil, t  , clear_r, 1  )
    X("r"      , R      , r  , xl    , id , kil, t  , nop    , 1  )
    X("revert" , REVERT , rvt, xl    , id , kil, t  , nop    , 1  )
    X("nop"    , NOP    , now, xl    , id , kil, t  , nop    , 1  )
    X("nolist" , NOLIST , now, xl    , id , kil, t  , nop    , 1  )
    X("yield"  , YIELD  , yld, xl    , id , kil, t  , nop    , 1  )
    X("pvio"   , PVIO   , pvi, xl    , id , mux, t  , nop    , 1  )

    // XXX alpha_imm (== alpha_boo) is never seen by circuitry, but *IS*
    //     seen by the logical simulation (as an identity function).
    //
    // cy_priv_imm is able to view ff m's state from prior instructions,
    // so it's designated as a privileged cycle type. Its use here for
    // IMP, IMN, IMH, IMB precludes such viewing, so these are safe
    // nonprivileged instructions. You can't see why by looking at the
    // IM* and RCM2 instructions in this table; the answer is in decoder.c
    // where RCM1 does a save_inst one click earlier than all others.
    X("imp"    , IMP    , priv_imm, xl    , imm, imp, t  , logic  , 0  )
    X("imn"    , IMN    , priv_imm, xl    , imm, imn, t  , logic  , 1  )
    X("imh"    , IMH    , priv_imm, xl    , imm, imh, t  , logic  , 0  )
    X("imb"    , IMB    , priv_imm, xl    , imm, add, t  , logic  , 0  )

    X("ld"     , LD     , ld , xl    , id , add, t  , logic  , 0  )
    X("addld"  , ADDLD  , ld , add_uu, id , add, t  , logic  , 0  )
    X("sto"    , STO    , sto, xr    , id , add, t  , nop    , 0  )
    X("sto2"   , STO2   , st2, xl    , tr , 999, t  , nop    , 0  )
    X("ldsto"  , LDSTO  , lds, xl    , tr , 999, t  , logic  , 0  )

    // Here is the growing list of privileged instructions:
    X("halt"   , HALT   , priv_hlt, xl    , id , kil, t  , nop    , 0  )
    X("wcm"    , WCM    , priv_wcm, xl    , id , add, t  , nop    , 0  )
    X("rcm1"   , RCM1   , priv_rcm, xl    , id , add, t  , nop    , 0  )
    X("rcm2"   , RCM2   , priv_imm, xl    , id , add, t  , logic  , 0  )
    X("wdm"    , WDM    , priv_wdm, xr    , id , add, t  , nop    , 0  )
    X("wdm2"   , WDM2   , priv_wd2, xl    , tr , 999, t  , nop    , 0  )
    X("rdm"    , RDM    , priv_rdm, xl    , id , add, t  , logic  , 0  )
    X("addrdm" , ADDRDM , priv_rdm, add_uu, id , add, t  , logic  , 0  )
    X("rwdm"   , RWDM   , priv_rwm, xl    , tr , 999, t  , logic  , 0  )

    X("wpt"    , WPT    , priv_wpt, xr    , id , add, t  , nop    , 0  )
    X("rpt"    , RPT    , priv_rpt, xl    , id , add, t  , nop    , 0  )
    X("cali"   , CALI   , priv_csi, xl    , id , kil, t  , nop    , 1  )
    X("xany"   , XANY   , priv_xa , xl    , id , kil, t  , nop    , 0  )
    X("jany"   , JANY   , priv_ja , xl    , id , kil, t  , j      , 0  )
    X("user"   , USER   , priv_uid, xr    , rtl, kil, t  , nop    , 0  )
    X("priv"   , PRIV   , priv_prv, xl    , id , kil, t  , nop    , 0  )
    X("npriv"  , NPRIV  , priv_usr, xl    , id , kil, t  , nop    , 0  )
    X("setup"  , SETUP  , priv_pts, xl    , id , kil, t  , nop    , 0  )
    X("peek"   , PEEK   , priv_pek, xl    , id , mux, t  , logic  , 0  )
    X("poke"   , POKE   , priv_pok, xl    , id , mux, t  , nop    , 0  )
    X("pcall"  , PCALL  , priv_pca, xl    , id , kil, t  , nop    , 1  )
    X("npcall" , NPCALL , priv_npc, xl    , id , kil, t  , nop    , 1  )
    X("hijack" , HIJACK , priv_hij, xl    , id , kil, t  , nop    , 1  )
    X("timer"  , TIMER  , priv_tim, add_uu, id , add, add, logic  , 0  )
    X("memset" , MEMSET , priv_set, xl    , id , kil, t  , nop    , 1  )

    X("xoob"   , XOOB   , now     , xl    , id , kil, t  , nop    , 1  )

#undef X
};

// This is a hook to convert an opcode to its textual name.
const char *opcode_name(int o)
{
    int i;

    for (i = 0; i < CARD(the_insts); i++)
        if (the_insts[i].opcode == o)
            return the_insts[i].name;

    return "INVALID";
}

// This converts an opcode name to its numeric form. Returns -1 if not found.
int find_opcode(char *name)
{
    int i;

    for (i = 0; i < CARD(the_insts); i++)
        if (!strcmp(the_insts[i].name, name))
            return the_insts[i].opcode;
    return -1;
}

/* FOOTNOTES

N:  (Applies to old assembler only.)
    For many instructions, the assembler does not rely solely
    on assembler name to determine the opcode to emit. For
    example, "++" can result in any ac.* instruction. So to
    "rename" instructions, changes to symbols in syntax.c may be
    needed in addition to or instead of changes to this field.

*/
