/* zeta (flags) SRAM definition

INPUT BITS

    0       previous N
    1       previous R
    2       previous T
    3       previous Z
    4-8     alpha5 encoded range
    9       theta carry summary -OR- flags' view of carry decision bit 5
    10      theta propagate summary
    11-15   flag slot from control decoder
    16      gamma_35 (gamma's idea of N)
    17      AND of z_i from gamma (gamma's idea of Z)

OUTPUT BITS

    0       N'egative (per additive functions), or 0 if using gamma's N
    1       R'ange
    2       T'emporal range
    3       Z'ero
    4       I'nterrupt
    5       branch decision (zeta_j_* only)

*/

// XXX There may be "safer" ways to write this routine, maybe table driven,
//     to avoid some types of oversights like not preserving Z.
static int zeta_operation(
    int prev_n,                     // previous N(egative) flag
    int prev_r,                     // previous R(ange) flag
    int prev_t,                     // previous T(emporal) flag
    int prev_z,                     // previous Z(ero) flag
    int encoded_range,              // see alpha.c RANGE ENCODING comment
    int c_summ,                     // theta's summary of alpha carry outputs
    int p_summ,                     // theta's summary of alpha prop. outputs
    int slot,                       // zeta's operation (from control decoder)
    int gamma_n,                    // naive sign bit from gamma output bit 35
    int gamma_z                     // AND is gamma's six naive Z output bits
) {
    int range =                     // [0, 4] per alpha.c RANGE ENCODING
        c_summ ? encoded_range / 5 : encoded_range % 5;
    enum {                          // flags can be 0, 1, or the following:
        maybe = 1024, keep = 2048,
    };

    // output bits
    int n;                          // new N(egative) flag
    int r;                          // new R(ange) flag
    int t;                          // new T(emporal) flag
    int z;                          // new Z(ero) derivation
    int interrupt = 0;              // new I(nterrupt) flag
    int jump = 0;                   // new branch decision

    // The control decoder decides among these cases:
    switch (slot) {
        default:                    // Don't spend NOR boot flash or
            return 0;               // boot time on unused slots.

        case zeta_nop:
            n = r = t = z = keep;
            break;

        case zeta_logic:
            n = gamma_n;
            r = keep;
            t = keep;
            z = gamma_z;
            break;

        case zeta_cmp:
            n = range < 2;
            r = keep;
            t = keep;
            z = range == 2 ? gamma_z : 0;
            break;

        case zeta_add_s:
            n = range < 2;
            r = maybe;
            t = range != 1 && range != 2;
            z = range == 2 ? gamma_z : 0;
            break;

        case zeta_add_u:
            n = range < 2;
            r = maybe;
            t = range != 2 && range != 3;
            z = range == 2 ? gamma_z : 0;
            break;

        case zeta_wrap:
            // Derived from zeta_add_u.
            // TODO Document that result is never negative, or tell assembler
            // not to permit to a signed destination, or both. XXX
            n = 0;
            r = keep;
            t = range != 2 && range != 3;
            z = range == 2 ? gamma_z : 0;
            break;

        case zeta_t_adj:
            n = gamma_n;
            r = keep;

            // Four cases: any P clears T, any C sets it
            // If we see any C with any P, invert T
            // Otherwise do not change T
            if (p_summ && c_summ)
                t = !prev_t;
            else if (p_summ)
                t = 0;
            else if (c_summ)
                t = 1;
            else
                t = 0;

            z = gamma_z;
            break;

        case zeta_bound:            // can only change I'nterrupt flag
            n = keep;
            r = keep;
            t = keep;
            z = keep;
            interrupt = range > 1;  // array subscript out of bounds
            break;

        case zeta_clear_r:          // copy R to T, then clear R
            n = keep;
            r = 0;
            t = prev_r;
            z = keep;
            break;

        case zeta_asr_uu:
            n = 0;
            r = keep;
            t = 0;
            z = gamma_z;
            break;

        case zeta_asr_us:
            n = 0;
            r = maybe;
            t = (encoded_range & 3) == 3;
            z = gamma_z;
            break;

        case zeta_asr_su:
            n = encoded_range & 1;
            r = maybe;
            t = n;
            z = gamma_z;
            break;

        case zeta_asr_ss:
            n = encoded_range & 1;
            r = maybe;
            t = 0;
            z = gamma_z;
            break;

        case zeta_asl_uu:
            n = 0;
            r = maybe;
            t = !!p_summ;
            z = t ? 0 : gamma_z;
            break;

        case zeta_asl_us:
            // 1 encroaches bit 35 or input is negative?
            n = 0;
            r = maybe;
            t = (p_summ || encoded_range & 1);
            z = t ? 0 : gamma_z;
            break;

        case zeta_asl_su:
            // 1 encroaches bit 36 or input is negative?
            n = encoded_range & 1;
            r = maybe;
            t = (p_summ || encoded_range & 1);
            z = t ? 0 : gamma_z;
            break;

        case zeta_asl_ss:
            // 1 encroaches bit 35 with non-negative input, or
            // 1 encroaches bit 35 with negative input?
            n = encoded_range & 1;
            r = maybe;
            t =    p_summ && !(encoded_range & 1)
                || c_summ && encoded_range & 1;
            z = t ? 0 : gamma_z;
            break;

        case zeta_j:
            n = r = t = z = keep;
            jump = 1;
            break;

        case zeta_j_lt:
            n = r = t = z = keep;
            jump = prev_n;
            break;

        case zeta_j_le:
            n = r = t = z = keep;
            jump = prev_n || prev_z;
            break;

        case zeta_j_eq:
            n = r = t = z = keep;
            jump = prev_z;
            break;

        case zeta_j_ne:
            n = r = t = z = keep;
            jump = !prev_z;
            break;

        case zeta_j_ge:
            n = r = t = z = keep;
            jump = !prev_n;
            break;

        case zeta_j_gt:
            n = r = t = z = keep;
            jump = !(prev_n || prev_z);
            break;

        case zeta_j_r:
            n = r = t = z = keep;
            jump = prev_r;
            break;

        case zeta_j_nr:
            n = r = t = z = keep;
            jump = !prev_r;
            break;

        case zeta_j_t:
            n = r = t = z = keep;
            jump = prev_t;
            break;

        case zeta_j_nt:
            n = r = t = z = keep;
            jump = !prev_t;
            break;
    }

    // Pack up and return the result.
    int out_bits = 0;

    // I(nterrupt) flag
    if (interrupt)
        out_bits |= bit4;

    // branch decision
    if (jump)
        out_bits |= bit5;

    // N(egative) flag
    if (n == keep)
        n = prev_n;
    if (n == 1)
        out_bits |= bit0;
    else if (n == 0)
        ;
    else
        ng("!zeta_operation n");

    // R(ange) flag
    if (r == maybe)
        r = prev_r ? 1 : t;
    if (r == keep)
        r = prev_r;
    if (r == 1)
        out_bits |= bit1;
    else if (r == 0)
        ;
    else
        ng("!zeta_operation r");

    // T(emporal) flag
    if (t == keep)
        t = prev_t;
    if (t == 1)
        out_bits |= bit2;
    else if (t == 0)
        ;
    else
        ng("!zeta_operation t");

    // Z(ero) flag
    if (z == keep)
        z = prev_z;
    if (z == 1)
        out_bits |= bit3;
    else if (z == 0)
        ;
    else
        ng("!zeta_operation z");

    return out_bits;
}

static int zeta_calc(int subword_unused, int i)
{
    int prev_n, prev_r, prev_t, prev_z,
        enc_ra, c_summ, p_summ, slot, gamma_n, gamma_z;

    prev_n = i       &  1;
    prev_r = i >>  1 &  1;
    prev_t = i >>  2 &  1;
    prev_z = i >>  3 &  1;
    enc_ra = i >>  4 & 31;
    c_summ = i >>  9 &  1;
    p_summ = i >> 10 &  1;
    slot = i >> 11 & 31;
    gamma_n = i >> 16 & 1;
    gamma_z = i >> 17 & 1;

    return zeta_operation(prev_n, prev_r, prev_t, prev_z, enc_ra,
        c_summ, p_summ, slot, gamma_n, gamma_z);
}
