/*
    c_1g74.c

    This is the simulation's implementation of a Texas Instruments
    SN74AUC1G74 flip-flop.
*/

typedef struct {
    Prefix pre;             // fields common to all components
    in_clock clk;           // clock input
    in_pin d;               // clocked data input
    in_pin pre_;            // active-low async preset input
    in_pin clr_;            // active-low async clear input
    out_pin q;              // output
    out_pin q_;             // inverted output
    int halts_sim;          // if nonzero, halt sim'n when 1 is clocked
    int pvio;               // paravirtualized I/O strobe
    int memset;             // simulated memset
} c_1g74;

// Here are names and byte offsets of input pins within the struct.
in_namer in_namer_1g74[] = {
    "clk", &((c_1g74 *) NULL)->clk.pin,
    "d", &((c_1g74 *) NULL)->d,
    "pre#", &((c_1g74 *) NULL)->pre_,
    "clr#", &((c_1g74 *) NULL)->clr_,
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
out_namer out_namer_1g74[] = {
    "q", &((c_1g74 *) NULL)->q,
    "q#", &((c_1g74 *) NULL)->q_,
    NULL, NULL,
};

// These are the datasheet timing parameters for SN74AUC1G74 (picoseconds).
enum {
    t_w_1g74 = 1000,     // min clock hi/lo, pre#/clr# lo
    t_su_1g74 = 400,     // setup time for data, pre#/clr# inactive
    t_h_1g74 = 300,      // hold time
    t_pd_q_1g74 = 1200,  // max propagation delay (to Q)  XXX min 800
    t_pd_qi_1g74 = 1100, // max propagation delay (to Q#) XXX min 700
    t_pd_pc_1g74 = 1500, // max propagation delay pre#/clr# to Q/Q# XXX min 900
};

// This is used by the paravirtualized I/O subsystem. See pvio.c for more.
int pvio_1g74_was_set;

// This is used for simulated memsets.
int memset_1g74_was_set;

// The simulation's power-up state for this component is not
// guaranteed to match the real world.
void init_1g74(UNUSED void *v, UNUSED int z)
{
    static int once;

    if (!once) {
        // The anticipated design does not use clr# or pre# after system
        // reset on any 1G74s, so we probably do not need to implement
        // width checking on these pins.
        pf("e.warn: width of 1G74 clr# and pre# pulses never checked");
        once = 1;
    }
}

// Called one hold time after a clock rise, this queues either the
// indicated state (if stable) or 'uncert' (if not stable)
// to the output pins.
static void hold_time_callback_1g74(c_1g74 *f, void *unused, int new_state)
{
    if (f->d.last_change + t_su_1g74 > f->clk.last_rise) {
        new_state = uncert;
        pf("e.setup/hold time violation by %li ps on %s",
            f->d.last_change + t_su_1g74 - f->clk.last_rise, f->pre.name);
    }

    // normal output
    pin_output_after_delay(&f->q, new_state, t_pd_q_1g74 - t_h_1g74);
    if (f->pre.trace)                   // user can infer q# is also clocked
        pf(".%s.q clocking %s\n", f->pre.name, pin_val_text(new_state));

    // compute inverted output (preserve 'uncert')
    if (new_state == lo)
        new_state = hi;
    else if (new_state == hi)
        new_state = lo;

    pin_output_after_delay(&f->q_, new_state, t_pd_qi_1g74 - t_h_1g74);
}

static void preset_or_clear(c_1g74 *f, int q, int q_)
{
#if 0
    // MWA propose to remove a/o Mar 2023
    pin_output_after_delay(&f->q, q, t_pd_pc_1g74 - 1);
    pin_output_after_delay(&f->q_, q_, t_pd_pc_1g74 - 1);
#else
    // MWA Mar 2023: I see no reason to shorten these delays 1 ps!
    pin_output_after_delay(&f->q, q, t_pd_pc_1g74);
    pin_output_after_delay(&f->q_, q_, t_pd_pc_1g74);
#endif
}

// Preset or clear from .ns script.
void tweak_1g74(void *v, char *parm)
{
    c_1g74 *f = v;

    if (!strcmp(parm, "clear")) {
        preset_or_clear(f, lo, hi);
        if (f->pre.trace)
            pf(".script cleared %s\n", f->pre.name);
    }

    else if (!strcmp(parm, "preset")) {
        preset_or_clear(f, hi, lo);
        if (f->pre.trace)
            pf(".script preset %s\n", f->pre.name);
    }

    else if (!strcmp(parm, "halts"))
        f->halts_sim = 1;

    else if (!strcmp(parm, "pvio"))
        f->pvio = 1;

    else if (!strcmp(parm, "memset"))
        f->memset = 1;

    else
        pf("e.unknown tweak %s on %s\n", parm, f->pre.name);
}

// This emulates a 1G74's activity after a change to one or more input pins.
void be_1g74(void *v)
{
    c_1g74 *f = v;
    int d;

    do_in_clock(&f->clk, t_w_1g74);

    // pre# and clr# have unconditional effect
    if (f->pre_.changed || f->clr_.changed) {
        f->pre_.changed = 0;
        f->clr_.changed = 0;

        switch (f->clr_.val) {
            case lo:
                // -1 resolves a corner case
                preset_or_clear(f, lo, hi);
                if (f->pre.trace)
                    pf(".%s.clr# asserted\n", f->pre.name);
                return;                 // nothing further thank you
            case hi:
                break;                  // look at pre# next
            default:
                pf("e.noisy clr# on %s", f->pre.name);
                return;                 // cannot determine this ff's state
        }

        switch (f->pre_.val) {
            case lo:
                // -1 resolves a corner case
                preset_or_clear(f, hi, lo);
                if (f->pre.trace)
                    pf(".%s.pre# asserted\n", f->pre.name);
                return;                 // nothing further thank you
            case hi:
                break;                  // look at clock next
            default:
                pf("e.noisy pre# on %s", f->pre.name);
                return;                 // cannot determine this ff's state
        }
    }

    if (!f->clk.rising)
        return;

    // clock is rising
    f->clk.rising = 0;

    if (f->clr_.val != hi || f->pre_.val != hi) {
        pf("e.pre# or clr# swallowed clock on %s", f->pre.name);
        return;
    }

    // compute new state
    d = f->d.val;

    if (f->pre.trace) {
        pf(".flop %s input pin %s\n", f->pre.name, pin_val_text(d));
    }

    if (f->clr_.last_change + t_su_1g74 > now 
            || f->pre_.last_change + t_su_1g74 > now) {
        pf("e.recovery time violation on %s", f->pre.name);
        d = uncert;
    }

    // provision for a 1G74 pseudo-component to implement HALT
    if (d == hi && f->halts_sim)
        queue_quit(1, "HALT via pseudo-1G74");

    // Removal time for the 1G74 happens to be 0, so we don't check.
    // Setup time is checked together with hold time in the callback.
    queue_add(t_h_1g74, (qcall) hold_time_callback_1g74, f, NULL, d);

    // If the control decoder turns on bit dc.d170, we set this bit
    // for the paravirtualized I/O subsystem's use. It doesn't go
    // through the ordinary signal queue, but it's timing would be
    // ok if we output this bit on click 1 or click 2. See pvio.c.
    if (f->pvio && d == hi)
        pvio_1g74_was_set = 1;

    // Likewise, control decoder bit dc.dc169 is a hook to do a
    // large memset in M0, M1, or P in order to speed execution.
    if (f->memset && d == hi)
        memset_1g74_was_set++;
}
