/*
    c_osc.c

    This oscillator spec is being tightened down in late-March 2023.
    The basic idea is that it will power up 'uncert' and stay that way
    until tweaked to start. The tweak sequence is one of:

    1. tweak the_osc 80_000_000f        ; 'f'requency 80 MHz
    1. tweak the_osc 6_250h             ; 'h'alf-period 6250 ps
    1. tweak the_osc 12_500p            ; 'p'eriod 12500 ps

    Then tweak the time for the first rising edge. There will be a falling
    edge (from uncert) a half-period prior. This means the first rising
    edge must not preceed a half-period into the simulation.

    If this tweak is skipped, starting the oscillator is an error.

    2. tweak the_osc 25_000i            ; 'i'nitial rise at absolute ps

    If this tweak is skipped, the initial rise is one full period after
    the dawn of simulation.

    A tweak may be needed to postpone the oscillator breakpoint to
    correspond to the clock tree propagation delay. Use e.g.

    3. tweak the_osc 14_200b            ; 'b'reakpoint delay 14.2 ns

    The 'b' tweak doesn't work as well as we may hope, because by
    breakpoint time the oscillator may already be ahead of us,
    causing "skipped beats." A better method might be to not use 'b'
    (zero breakpoint delay), but slow the clock enough that the
    entire circuit's outcome can be examined between clicks during
    debugging. That "better method" may too come under fire later if
    race conditions need debugged.

    Then tweak the number of rising edges (clicks) to generate. The
    oscillator will run continuously without interruption, but the
    simulation will quit (pause) 1 ps before the n+1st rising edge.

    4. tweak the_osc 20c                ; 20 'c'licks = 5 instructions
    4. tweak the_osc c                  ; 'c'ontinuous (no pause)

    Tweaking the oscillator does not start or continue simulation; use
    the newly-named 'run' (previously 'settle') command for that.

    The first 'c' tweak will do queue insertions for the first falling
    and rising edges. The falling edge location is rounded to a picosecond
    boundry. The oscillator callback handles reschedules and quits. The
    first 'c' tweak is also the point where any calculations from prior
    tweaks are realized. Once a 'c' tweak happens, none of the other
    tweaks are permitted for the oscillator.

    In order to keep oscillator semantics no more complex than actually
    needed for minicomputer development, there is no provision for
    changing oscillator frequency, pausing the oscillator once started,
    or handling the lo/hi/uncert states more flexibly.

    There is one more oscillator tweak for regression testing, and it's
    a huge kludge, because the oscillator isn't actually involved in the
    tweak. It uses the good() and fail() functions for counting how
    many regression tests were ok. It's implemented in the oscillator,
    because that's the component that actually stops the clock right now,
    and I wanted a regression test for that.

    5. tweak the_osc 1_002_003t         ; verify current 't'ime is this
*/

typedef struct {
    Prefix pre;             // all components have this
    int period_ps;          // reciprocal of frequency in THz
    int first_rise_ps;      // absolute time of first rising clock edge
    int delay_ps;           // additional breakpoint delay
    int started;            // true if oscillator has started
    long clicks;            // total count of clicks
    long stop_after_click;  // this click will set a queue_quit()
    out_pin out;            // square-wave output
    in_pin unknown;         // use TBD
} c_osc;

// Here is the name and byte offset of the struct's one output pin.
out_namer out_namer_osc[] = {
    "out", &((c_osc *) NULL)->out,
    NULL, NULL,
};

// The datasheet does not say what this pin does.
in_namer in_namer_osc[] = {
    "UNKNOWN", &((c_osc *) NULL)->unknown,
    NULL, NULL,
};

// The initialization has an internal connection to make so that
// oscillation can occur.
void init_osc(void *v, int truth)
{
    // There is nothing to do; the c_osc was zeroed by trackmem
    // and then the prefix was filled in. Other initialization is
    // waiting until the first 'c' tweak.
}

// This is called from the queue for every lo and hi transition,
// with one lo and one hi always kept in the queue.
static void osc_callback(c_osc *g, void *unused, int state)
{
    // Queue the next one like this.
    queue_add(g->period_ps, (qcall) osc_callback, g, NULL, state);

    // Flip the output pin.
    // XXX Unusually, I'm queuing with delay 0. Will anything break?
    pin_output_after_delay(&g->out, state, 0);

    // Just consider rising edges.
    if (state != hi)
        return;
    ++g->clicks;

    // Indicate rising edges in output.
    if (g->pre.trace)
        pf("." YEL "****** click %i rising edge ******" BAW "\n",
                g->clicks % 4);

    // If a breakpoint is requested, it goes 1 ps before the next rising edge.
    if (g->clicks == g->stop_after_click)
        queue_quit(now + g->period_ps - 1 + g->delay_ps,
                "oscillator breakpoint");
}

// This lets us change the half-period of the oscillator.
// 'parm' is a decimal integer string with units of picoseconds.
void tweak_osc(void *v, char *parm)
{
    c_osc *g = v;
    long number = 0;                    // from digits in 'parm'
    int half_ps;
    char *bad_tweak = parm;

    for (; *parm; ++parm) {

        if (isdigit(*parm)) {           // digits forming number
            number = 10 * number + *parm - '0';
            continue;
        }

        if (*parm == '_')               // group separator for digits
            continue;

        switch (*parm) {
            case 'f':                   // frequency
                if (!number || g->started)
                    goto bad;                
                g->period_ps = 1000000000000L / number;
                break;

            case 'h':                   // half-period
                number *= 2;
                nobreak;
            case 'p':                   // period
                if (!number || g->started)
                    goto bad;                
                g->period_ps = number;
                break;

            case 'b':                   // breakpoint delay
                if (g->started)
                    goto bad;
                g->delay_ps = number;
                break;

            case 'i':                   // initial rise at time
                if (g->started)
                    goto bad;
                g->first_rise_ps = number;
                break;

            case 't':                   // time check
                if (now == number)
                    pf(".%s - script knows the correct time\n", good());
                else
                    pf(".%s - script thinks time is %li ps\n", fail(), number);
                break;

            case 'c':                   // clicks, or continuous

                // The first 'c' will start the oscillator. Here's how.
                if (!g->started) {
                    g->started = 1;

                    // Ensure the oscillator was explicitly given a frequency.
                    if (!g->period_ps) {
                        pf("e!.%s frequency left unspecified", g->pre.name);
                        break;
                    }

                    // Ensure the first rise time is sane.
                    if (!g->first_rise_ps)
                        g->first_rise_ps = g->period_ps;
                    half_ps = g->period_ps / 2;
                    if (g->first_rise_ps < half_ps) {
                        pf("e!.%s first rise is too early", g->pre.name);
                        break;
                    }

                    // Queue the first transition to lo and hi.
                    queue_add(g->first_rise_ps - half_ps, (qcall) osc_callback,
                            g, NULL, lo);
                    queue_add(g->first_rise_ps, (qcall) osc_callback, g, NULL, hi);
                }

                // Insert a simulation breakpoint if requested after a specific
                // number of clock rises.
                g->stop_after_click = number ? g->clicks + number : 0;
                break;

            default:
                goto bad;
        }

        // If we get here, a lettered option was just taken.
        // Clear any entered number in order to support catenated tweaks.
        number = 0;
    }

    // All done.
    return;

bad:
    pf("e!.bad tweak %s for %s", bad_tweak, g->pre.name);
}

// Here is the simulation itself.
void be_osc(void *v)
{
    // Because there are no input pins, there is nothing to do here.
}
