/*
    c_pulse.c

    This simulates a non-existent device that marks clicks in the
    simulator output. Here's how you would use this:

    kind pulse
    name heartbeat
    from giant7
    to heartbeat
    signal y15 clk 1    ; sim can't handle 0 ps delay, so use 1 ps
    trace heartbeat

    This will cause ****** click 0 rising edge ****** messages to
    be displayed every fourth giant7.y15 transitions high, and the
    click 1, 2, 3 messages at their appropriate points.

    The reason this was implemented as a non-existent device is that
    it was too kludgy and inflexible to hardcode these messsages
    into the SN74AUC374 (c_374.c) simulation.
*/

typedef struct {
    Prefix pre;             // fields common to all components
    int click;              // 0, 1, 2, 3, repeat

    in_clock clk;           // just a clock pin, no output pins
} c_pulse;

// Here are names and byte offsets of input pins within the struct.
in_namer in_namer_pulse[] = {
    "clk", &((c_pulse *) NULL)->clk.pin,
    NULL, NULL,
};

// There aren't any output pins.
out_namer out_namer_pulse[] = {
    NULL, NULL,
};

void init_pulse(UNUSED void *v, UNUSED int parm)
{
    // No action is needed.
}

/*
    The only pulse tweak is to advance the click count in order to
    synchronize it with ground truth. One click per parameter character.
*/
void tweak_pulse(void *v, char *parm)
{
    c_pulse *g = v;
    g -> click = (g -> click + strlen(parm)) % 4;
}

// Here is the simulation for this "virtual" part.
void be_pulse(void *v)
{
    c_pulse *p = v;

    do_in_clock(&p->clk, 100);      // 100 ps minimum width very arbitrary
    if (!p->clk.rising)
        return;

    // clock is rising
    p->clk.rising = 0;
    p->click = (p->click + 1) % 4;
    if (p->pre.trace)
        pf("." YEL "****** %s %i rising edge ******" BAW "\n",
                p->pre.name, p->click % 4);
}
