/*
    c_sram.c

    This is the simulation's implementation of several 18-wide and 36-wide
    zero bus turnaround synchronous static RAMs made by GSI, including

          GS840Z18CGT-250I   256k x 18
          GS880Z18CGT-250I   512k x 18
         GS8160Z18DGT-250I     1M x 18
         GS8320Z18AGT-250I     2M x 18
         GS8640Z18 GT-250I     4M x 18
        GS81280Z18 GT-250I     8M x 18

          GS840Z36CGT-250I   128k x 36
          GS880Z36CGT-250I   256k x 36
         GS8160Z36DGT-250I   512k x 36
         GS8320Z36AGT-250I     1M x 36
         GS8640Z36 GT-250I     2M x 36
        GS81280Z36 GT-250I     4M x 36

    A qwirk of the architecture is that address pins that don't
    apply for a particular memory size are to be grounded. This
    permits part substitutions with larger sizes in some cases,
    as well as simplifies the simulation.

    In order to make address pin wiring errors easier to detect 
    without explicitly checking the width of all parts, this
    module intentionally reads any value other than 'lo' on an
    address pin as 'hi'. This will show up as an out-of-range
    error on the chip in the event unused address inputs are not
    grounded as intended.
*/

#include "pvio.c"       // simulator-provided paravirtualized filesystem

enum {
    max_addr_pins = 24, // number of pins we read to obtain address

    // diagonstic settings
    check_constant_inputs = 1,

    // t_ name conventions are from GSI's GS8320Z36AGT-250I datasheet.
    t_KH = 1300,        // min clock high time
    t_KL = 1500,        // min clock low time
    t_KQX = 2000,       // min clock to output invalid
    t_KQ = 5500,        // max clock to output valid
    t_HZ = 2500,        // max clock to output in hi-Z

    t_S = 1500,         // setup time
    t_H = 500,          // hold time

    // These state names are from the same datasheet, rev. 1.03. Try page 8.
    // Keep these in sync with the state_names[] strings below.
    deselect = 0,
    new_read,
    burst_read,
    new_write,
    burst_write,
};

// synchronous static RAM
typedef struct {
    Prefix pre;             // fields common to all components

    // Configuration
    int log2;               // log of memory size; number of 'a' used
    int rows;               // 2 ** 'log2'
    int width;              // width in bits (18 or 36); number of 'd' used
    int                     // ONLY for early simulations
        simulator_write_protect;
    int force_read_tweak;   // ONLY for early simulations
    int memset_tweak;       // simulate fast memset
    int detailed_trace;     // extra diagnostic output :)
    long *pvio_operand;     // NULL / &pvio_data / &pvio_cmd
    int *pvio_operand_src;  // NULL / &pvio_data_src / &pvio_cmd_src
    long last_val_stored;   // last value written
    int last_val_stored_at; // address of last value of written (for memsets)

    // Storage
    void *mem;              // int * or long *, depending on 'width'

    int rw_state;           // r/w control state
    int addr;               // most recent r/w address

    // Clock input
    in_clock ck;            // clock (use ck.pin to connect!)

    // Clocked inputs
    in_pin                  // address
        a[max_addr_pins];
    in_pin d[36];           // data (d[] and q[] are same physical pins)
    in_pin ft_;             // flow-through mode (high not implemented)
    in_pin lbo_;            // linear byte order ((high not implemented)
    in_pin adv;             // lo: load, hi: advance
    in_pin cke_;            // clock enable
    in_pin w_;              // write enable
    in_pin b_[4];           // byte write enable (high not implemented)
    in_pin e1_;             // active-low enable
    in_pin e2;              // active-high enable
    in_pin e3_;             // another active-low enable

    // Unclocked inputs
    in_pin zz;              // sleep (high not implemented)
    in_pin g_;              // output enable (high not implemented)

    // Clocked outputs
    out_pin q[36];          // data (d[] and q[] are same physical pins)
} c_sram;

// Here are names and byte offsets of input pins within the struct.
// Note that d0-d35 are also output pins.
in_namer in_namer_sram[] = {
    "a0", &((c_sram *) NULL)->a[0],
    "a1", &((c_sram *) NULL)->a[1],
    "a2", &((c_sram *) NULL)->a[2],
    "a3", &((c_sram *) NULL)->a[3],
    "a4", &((c_sram *) NULL)->a[4],
    "a5", &((c_sram *) NULL)->a[5],
    "a6", &((c_sram *) NULL)->a[6],
    "a7", &((c_sram *) NULL)->a[7],
    "a8", &((c_sram *) NULL)->a[8],
    "a9", &((c_sram *) NULL)->a[9],
    "a10", &((c_sram *) NULL)->a[10],
    "a11", &((c_sram *) NULL)->a[11],
    "a12", &((c_sram *) NULL)->a[12],
    "a13", &((c_sram *) NULL)->a[13],
    "a14", &((c_sram *) NULL)->a[14],
    "a15", &((c_sram *) NULL)->a[15],
    "a16", &((c_sram *) NULL)->a[16],
    "a17", &((c_sram *) NULL)->a[17],
    "a18", &((c_sram *) NULL)->a[18],
    "a19", &((c_sram *) NULL)->a[19],
    "a20", &((c_sram *) NULL)->a[20],
    "a21", &((c_sram *) NULL)->a[21],
    "a22", &((c_sram *) NULL)->a[22],
    "a23", &((c_sram *) NULL)->a[23],
    "d0", &((c_sram *) NULL)->d[0],
    "d1", &((c_sram *) NULL)->d[1],
    "d2", &((c_sram *) NULL)->d[2],
    "d3", &((c_sram *) NULL)->d[3],
    "d4", &((c_sram *) NULL)->d[4],
    "d5", &((c_sram *) NULL)->d[5],
    "d6", &((c_sram *) NULL)->d[6],
    "d7", &((c_sram *) NULL)->d[7],
    "d8", &((c_sram *) NULL)->d[8],
    "d9", &((c_sram *) NULL)->d[9],
    "d10", &((c_sram *) NULL)->d[10],
    "d11", &((c_sram *) NULL)->d[11],
    "d12", &((c_sram *) NULL)->d[12],
    "d13", &((c_sram *) NULL)->d[13],
    "d14", &((c_sram *) NULL)->d[14],
    "d15", &((c_sram *) NULL)->d[15],
    "d16", &((c_sram *) NULL)->d[16],
    "d17", &((c_sram *) NULL)->d[17],
    "d18", &((c_sram *) NULL)->d[18],
    "d19", &((c_sram *) NULL)->d[19],
    "d20", &((c_sram *) NULL)->d[20],
    "d21", &((c_sram *) NULL)->d[21],
    "d22", &((c_sram *) NULL)->d[22],
    "d23", &((c_sram *) NULL)->d[23],
    "d24", &((c_sram *) NULL)->d[24],
    "d25", &((c_sram *) NULL)->d[25],
    "d26", &((c_sram *) NULL)->d[26],
    "d27", &((c_sram *) NULL)->d[27],
    "d28", &((c_sram *) NULL)->d[28],
    "d29", &((c_sram *) NULL)->d[29],
    "d30", &((c_sram *) NULL)->d[30],
    "d31", &((c_sram *) NULL)->d[31],
    "d32", &((c_sram *) NULL)->d[32],
    "d33", &((c_sram *) NULL)->d[33],
    "d34", &((c_sram *) NULL)->d[34],
    "d35", &((c_sram *) NULL)->d[35],
    "ck", &((c_sram *) NULL)->ck.pin,
    "ft#", &((c_sram *) NULL)->ft_,
    "lbo#", &((c_sram *) NULL)->lbo_,
    "adv", &((c_sram *) NULL)->adv,
    "cke#", &((c_sram *) NULL)->cke_,
    "w#", &((c_sram *) NULL)->w_,
    "b0#", &((c_sram *) NULL)->b_[0],
    "b1#", &((c_sram *) NULL)->b_[1],
    "b2#", &((c_sram *) NULL)->b_[2],
    "b3#", &((c_sram *) NULL)->b_[3],
    "e1#", &((c_sram *) NULL)->e1_,
    "e2", &((c_sram *) NULL)->e2,
    "e3#", &((c_sram *) NULL)->e3_,
    "zz", &((c_sram *) NULL)->zz,
    "g#", &((c_sram *) NULL)->g_,
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
// Note that d0-d35 are also input pins.
out_namer out_namer_sram[] = {
    "d0", &((c_sram *) NULL)->q[0],
    "d1", &((c_sram *) NULL)->q[1],
    "d2", &((c_sram *) NULL)->q[2],
    "d3", &((c_sram *) NULL)->q[3],
    "d4", &((c_sram *) NULL)->q[4],
    "d5", &((c_sram *) NULL)->q[5],
    "d6", &((c_sram *) NULL)->q[6],
    "d7", &((c_sram *) NULL)->q[7],
    "d8", &((c_sram *) NULL)->q[8],
    "d9", &((c_sram *) NULL)->q[9],
    "d10", &((c_sram *) NULL)->q[10],
    "d11", &((c_sram *) NULL)->q[11],
    "d12", &((c_sram *) NULL)->q[12],
    "d13", &((c_sram *) NULL)->q[13],
    "d14", &((c_sram *) NULL)->q[14],
    "d15", &((c_sram *) NULL)->q[15],
    "d16", &((c_sram *) NULL)->q[16],
    "d17", &((c_sram *) NULL)->q[17],
    "d18", &((c_sram *) NULL)->q[18],
    "d19", &((c_sram *) NULL)->q[19],
    "d20", &((c_sram *) NULL)->q[20],
    "d21", &((c_sram *) NULL)->q[21],
    "d22", &((c_sram *) NULL)->q[22],
    "d23", &((c_sram *) NULL)->q[23],
    "d24", &((c_sram *) NULL)->q[24],
    "d25", &((c_sram *) NULL)->q[25],
    "d26", &((c_sram *) NULL)->q[26],
    "d27", &((c_sram *) NULL)->q[27],
    "d28", &((c_sram *) NULL)->q[28],
    "d29", &((c_sram *) NULL)->q[29],
    "d30", &((c_sram *) NULL)->q[30],
    "d31", &((c_sram *) NULL)->q[31],
    "d32", &((c_sram *) NULL)->q[32],
    "d33", &((c_sram *) NULL)->q[33],
    "d34", &((c_sram *) NULL)->q[34],
    "d35", &((c_sram *) NULL)->q[35],
    NULL, NULL,
};

// Keep these names in sync with the corresponding enums above.
static char *state_names[] = {
    "deselect", "new_read", "burst_read", "new_write", "burst_write"
};

/*
    The datasheet does an excellent job documenting the SRAM transitions,
    so this module was strikingly easy to write despite its appearance
    of complexity. This table encapsulates most of the SRAM's conduct.
*/
static struct fsm {
    int prev;                   // previous state
    int x;                      // transition
    int load;                   // load address from A pins
    int adv;                    // advance previous address
    int read;                   // read SRAM at this clock's address to Q pins
    int write;                  // store D pins at last clock's address
    int loz;                    // go from high to low impedance
    int hiz;                    // go from low to high impedance
    int current;                // (new) current state
} fsm_table[] = {
//  prev          x   L  A  R  W lZ hZ  current
    deselect,    'B', 0, 0, 0, 0, 0, 0, deselect,
    deselect,    'D', 0, 0, 0, 0, 0, 0, deselect,
    deselect,    'R', 1, 0, 1, 0, 1, 0, new_read,
    deselect,    'W', 1, 0, 0, 0, 0, 0, new_write,

//  prev          x   L  A  R  W lZ hZ  current
    new_read,    'B', 0, 1, 1, 0, 0, 0, burst_read,
    new_read,    'D', 0, 0, 0, 0, 0, 1, deselect,
    new_read,    'R', 1, 0, 1, 0, 0, 0, new_read,
    new_read,    'W', 1, 0, 0, 0, 0, 1, new_write,

//  prev          x   L  A  R  W lZ hZ  current
    burst_read,  'B', 0, 1, 1, 0, 0, 0, burst_read,
    burst_read,  'D', 0, 0, 0, 0, 0, 1, deselect,
    burst_read,  'R', 1, 0, 1, 0, 0, 0, new_read,
    burst_read,  'W', 1, 0, 0, 0, 0, 1, new_write,

//  prev          x   L  A  R  W lZ hZ  current
    new_write,   'B', 0, 1, 0, 1, 0, 0, burst_write,
    new_write,   'D', 0, 0, 0, 1, 0, 0, deselect,
    new_write,   'R', 1, 0, 1, 1, 1, 0, new_read,
    new_write,   'W', 1, 0, 0, 1, 0, 0, new_write,

//  prev          x   L  A  R  W lZ hZ  current
    burst_write, 'B', 0, 1, 0, 1, 0, 0, burst_write,
    burst_write, 'D', 0, 0, 0, 1, 0, 0, deselect,
    burst_write, 'R', 1, 0, 1, 1, 1, 0, new_read,
    burst_write, 'W', 1, 0, 0, 1, 0, 0, new_write,
};

// Here's a pointer to the end of the above table.
static struct fsm *fsm_end = fsm_table + CARD(fsm_table);

// This reads from a memory location in the SRAM.
long peek(c_sram *r, int addr)
{
    if (addr < 0 || addr >= r->rows) {
        pf("e.illegal peek %s[%i]", r->pre.name, addr);
        return 0;
    }

    if (r->width == 36)
        return ((long *) r->mem)[addr];
    else
        return ((int *) r->mem)[addr];
}

// This writes to a memory location in the SRAM.
void poke(c_sram *r, int addr, long data)
{
    if (r->simulator_write_protect) {
        if (r->pre.trace)
            pf(".%s is write protected, not really storing\n", r->pre.name);
        return;         // This does NOT work on physical implementations.
    }

    if (addr < 0 || addr >= r->rows) {
        pf("e.illegal poke %s[%i] = %li", r->pre.name, addr, data);
        return;
    }

    if (r->width == 36)
        ((long *) r->mem)[addr] = data;
    else
        ((int *) r->mem)[addr] = data;
}

// This is only useful called on the L or R (register) RAMs.
// It copies all of user 0's registers to the other users' registers.
// Its purpose is to make assembler constants available to multiple
// users when testing multitasking.
static void fork_tweak(c_sram *r)
{
    int i, j;
    long copy;

    for (i = 0; i < 512; ++i) {
        copy = peek(r, i);
        for (j = i + 512; j < r->rows; j += 512)
            poke(r, j, copy);
    }

    pf(".fork tweak on %s\n", r->pre.name);
}

/*
    The simulation's power-up state for this component surely
    will not match the real world. For one thing, the simulation's
    SRAM is initialized with zeros.
*/
void init_sram(void *v, int parm)
{
    c_sram *r = v;
    int i;
    static int tattled;

    // Make sure we know word sizes.
    if (sizeof(long) != 8)
        pf("e!.long is %zi bytes", sizeof(long));
    if (sizeof(int) != 4)
        pf("e!.int is %zi bytes", sizeof(int));

    // Advise operator as to extra checking.
    if (!tattled) {
        tattled = 1;
        pf("e.constant SRAM inputs %s checked this run",
             check_constant_inputs ? "ARE" : "ARE NOT");
    }

    // Firmware comes from the same source as the VM.
    get_firmware(r->pre.name, &r->width, &r->rows, &r->mem);

    // Compute number of address lines in use.
    for (r->log2 = 1; 1 << r->log2 < r->rows; r->log2++);
    if (1 << r->log2 != r->rows)
        pf("e.strange RAM size %ix%i", r->rows, r->width);

    // D and Q are the same physical pins. Connect them internally.
    for (i=0; i < r->width; i++)
        connect_pin_to_pin(&r->q[i], &r->d[i], 1);

    // SRAM is not doing anything at power-up.
    r->rw_state = deselect;

    // The datasheet says that I/O pins power up with high impedance.
    output_impedance_after_delay(r->q, r->width, 1);
}

/*
    TODO This whole comment block is out of date 10 July 2023.

    The only tweaks are to enable/disable write protection, which exists
    only in simulation.

    An ability to query or change specific locations would be useful,
    but has not been implemented.
*/
void tweak_sram(void *v, char *parm)
{
    c_sram *r = v;

    if (!strcmp(parm, "ro"))                // read-only SRAM
        r->simulator_write_protect = 1;
    else if (!strcmp(parm, "rw"))           // read-write SRAM
        r->simulator_write_protect = 0;
    else if (!strcmp(parm, "forceread"))
        r->force_read_tweak++;              // 25Oct2022 hack to boot code RAM
    else if (!strcmp(parm, "detailed"))
        r->detailed_trace++;                // extra diagnostic output
    else if (!strcmp(parm, "fork"))
        fork_tweak(r);                      // for testing multitasking
    else if (!strcmp(parm, "memset"))
        r->memset_tweak++;                  // virtual memsets
    else if (!strcmp(parm, "pvio")) {       // "L" register for PVIO
        r->pvio_operand = &pvio_data;
        r->pvio_operand_src = &pvio_data_src;
    }
    else if (!strncmp(parm, "pvio=", 5)) {  // "R" register for PVIO
        r->pvio_operand = &pvio_cmd;
        r->pvio_operand_src = &pvio_cmd_src;
        pvio_basepath(parm + 5);
    }
    else if (*parm == '@') {
        long want = number(parm+1, "-0lk"); // regression: last value stored
        long mask = r->width == 18 ? 0777777L : 0777777777777L;

        if ((want ^ r->last_val_stored) & mask)
            pf(".%s - %s last value was %'li instead of %li\n",
                    fail(), r->pre.name, r->last_val_stored, want);
        else
            pf(".%s - %s last value stored is %'li\n",
                    good(), r->pre.name, want);
    }
    else
        pf("e.strange tweak '%s' on '%s'", parm, r->pre.name);
}

/*
    This provides an assertion that pins providing features that
    we don't use (and that the simulator does not support) are
    connected to ground.
*/
void be_lo(in_pin *p)
{
    if (p->val != lo)
        pf("e.constant input pin %s has wrong value", p->name);
}

/*
    This checks to see that input pins are receiving valid values.
    The most likely problem would be a setup/hold issue, but another
    thing that can happen is that a valid high or low might not be
    avaiable at a pin. During board power-up, legions of these
    notices may be generated, but they are supposed to go away once
    the system has stabilized.
*/
static void check_setup_hold_times(
    c_sram *r,
    in_pin *p,
    time_ps cutoff,
    char **misc                     // concise list of failing pin names
) {
//  int warn = r->pre.trace;       // FIXME warn all, not just traced 25Oct2022
    int warn = 1;
    int pin_problem = 0;

    if (p->last_change > cutoff) {
        ++pin_problem;
        if (r->detailed_trace)
            pf("e.setup/hold time violation on %s by %li ps",
                    p->name, p->last_change - cutoff);
    }

    // This MIGHT give unwanted warnings on D pins, but I don't think so
    // because of all the lockouts.
    if (now >= stable_at && warn && p->val != lo && p->val != hi) {
        ++pin_problem;
        if (r->detailed_trace)
            pf("e.invalid clocked value %i on %s", p->val, p->name);
    }

    if (pin_problem)
        *misc += sprintf(*misc, "%s ", in_pin_name(p));
}

// Called one hold time after a clock rise, confirm inputs are stable
// and emulate the SRAM.
static void hold_time_callback(c_sram *r, void *unused, int ignored)
{
    struct fsm *fsm;
    int i;              // loop counter
    int x;              // transition: R)ead W)rite B)urst D)eselect
    int addr;           // address read from A pins
    long word;          // word read from D pins
    int bit;            // 'lo' or 'hi'
    time_ps cutoff = r->ck.last_rise - t_S;

    // names of misclocked pins
    // XXX unchecked buffer
    char misclocked_pins[1024], *misc = misclocked_pins;

    *misc = '\0';

    // If clock enable is not low, ignore this clock pulse.
    check_setup_hold_times(r, &r->cke_, cutoff, &misc);
    if (r->cke_.val != lo)
        goto summarize;

    // Check most setup and hold times.
    // We don't check the data inputs yet, because we don't know
    // if they apply.
    for (i = 0; i < max_addr_pins; i++)
        check_setup_hold_times(r, &r->a[i], cutoff, &misc);
    check_setup_hold_times(r, &r->adv, cutoff, &misc);
    check_setup_hold_times(r, &r->w_, cutoff, &misc);
    check_setup_hold_times(r, &r->e1_, cutoff, &misc);
    check_setup_hold_times(r, &r->e2, cutoff, &misc);
    check_setup_hold_times(r, &r->e3_, cutoff, &misc);

    // Determine what transition applies.
    // See synchronous truth table in datasheet.
    if (r->force_read_tweak != 0) {
        x = 'R';                // diagnostic to help CPU boot in simulation
        r->force_read_tweak--;
    }
    else if (r->adv.val == hi)
        x = 'B';                // B)urst; see FOOTNOTE 1
    else if (r->e1_.val != lo || r->e2.val != hi || r->e3_.val != lo)
        x = 'D';                // D)eselect
    else if (r->w_.val == lo)
        x = 'W';                // W)rite
    else
        x = 'R';                // R)ead

    for (fsm = fsm_table; fsm < fsm_end; ++fsm)
        if (fsm->prev == r->rw_state && fsm->x == x)
            goto yes;
    pf("e!.SRAM fsm table incomplete");
yes:

    if (r->detailed_trace && r->pre.trace)
        pf(".%s %s + %c will become %s\n", r->pre.name,
            state_names[r->rw_state], x, state_names[fsm->current]);

    // Store D pins at the last clock's address.
    if (fsm->write) {

        // It's at this point that data setup and hold times need validated.
        for (i = 0; i < r->width; i++)
            check_setup_hold_times(r, &r->d[i], cutoff, &misc);

        // Then we can read the data pins.
        for (word = 0, i = r->width - 1; i >= 0; --i) {
            word <<= 1;
            if (r->d[i].val == hi)
                ++word;
        }

        // This write hook for register-mapped paravirtualized I/O
        // may override the word that is being written when PVIO is on.
        if (r->pvio_operand)
            word = pvio_write(word);

        if (r->addr >= r->rows) {   // This should not happen after startup.
            // MWA Actually, this can happen while the OS is starting
            //     up and probing for memory size. I changed this to no
            //     longer be an error 14 July 2023.
            pf(".out of range address %i for %s", r->addr, r->pre.name);
            r->addr &= (r->rows - 1);
        }

        // This is a simulated memset. The base address is r->addr,
        // except the rightmost 1 is replaced with 0. The length is
        // twice the place value of the replaced 1.
        if (r->memset_tweak && memset_1g74_was_set) {
            if (r->pre.trace)
                pf("." GGRE "MEMSET %s[%i..%i] = %li" BAW "\n",
                        r->pre.name, r->last_val_stored_at, r->addr, word);
            for (i=r->last_val_stored_at; i<=r->addr; i++)
                poke(r, i, word);
            r->last_val_stored_at = r->addr;
            r->last_val_stored = word;
            memset_1g74_was_set--;
        }

        // This is a regular store.
        else {
            // Then we write to the SRAM.
            if (r->pre.trace)
                pf("." GGRE "STORING %li" BAW " (%lx`h %012lo`o) to %s[%i]\n",
                        word, word, word, r->pre.name, r->addr);
            poke(r, r->addr, word);
            r->last_val_stored_at = r->addr;
            r->last_val_stored = word;
        }
    }

    // Advance the previous address. Only the two lowest bits change.
    if (fsm->adv) {
        r->addr = r->addr & ~3 | 3 & (r->addr + 1);
        if (r->detailed_trace && r->pre.trace)
            pf(".%s advancing address to %i\n", r->pre.name, r->addr);
    }

    // Load an address from the A pins.
    else if (fsm->load) {
        for (addr = 0, i = max_addr_pins - 1; i >= 0; --i) {
            addr <<= 1;
            if (r->a[i].val != lo)  // Only lo is guaranteed to mean lo.
                ++addr;
        }

        if (addr >= r->rows) {      // This should not happen after startup.
            // MWA Actually, this can happen while the OS is starting
            //     up and probing for memory size. I changed this to no
            //     longer be an error 14 July 2023.
            pf(".out of range address %i for %s", addr, r->pre.name);
            addr &= (r->rows - 1);
        }

        if (r->detailed_trace && r->pre.trace)
            pf(".%s setting address to %i\n", r->pre.name, addr);
        r->addr = addr;
    }

    // Read out the memory's value at this clock's address to the Q pins.
    if (fsm->read) {
        word = peek(r, r->addr);

        // If this is one of the register files and paravirtualized I/O
        // has been turned on, save the value read for pvio.c
        if (r->pvio_operand) {
            *r->pvio_operand = word;
            *r->pvio_operand_src = r->addr;
        }

        // Here is some diagnostic output.
        if (r->pre.trace) {
            // For the code RAM, break the instruction into a legible form.
            if (!strcmp(r->pre.name, "C") && boredom - anti_boredom <= 0) {
                pf(".fetched C[%i] ", r->addr);
                asm_disassemble(RED, word, BAW "\n");
            }
            // The other RAMs display their contents as-is.
            else
                pf(".reading %li (%lx`h %lo`o) from %s[%i]\n",
                        word, word, word, r->pre.name, r->addr);
        }

        // Go low-impedance on new reads. One call does many bits.
        if (fsm->loz) {
            if (r->pre.trace && r->detailed_trace)
                pf(".%s outputs going lo-Z\n", r->pre.name);
            output_impedance_after_delay(r->q, -r->width, t_KQX - t_H);
        }

        // Before the output is valid, the previous output becomes invalid.
        // Note also we're not on the clock pulse, but one hold time later.
        for (i = 0; i < r->width; i++) {
            bit = word & 1L << i ? hi : lo;
            pin_output_after_delay(&r->q[i], uncert, t_KQX - t_H);
            pin_output_after_delay(&r->q[i], bit, t_KQ - t_H);
        }

#if 0
        // XXX This was moved to a dedicated pseudo-1G74 for better flexibility.
        // The HALT instruction is simulated here (for now).
        if (!strcmp(r->pre.name, "C") && word >> 27 == oc_HALT)
            // HALT must not just be fetched, but have started executing.
            queue_quit(now + 25000, "HALT instruction");
#endif
    }

    // Output pins shut off when a read is complete.
    if (fsm->hiz) {
        if (r->pre.trace && r->detailed_trace)
            pf(".%s outputs going hi-Z\n", r->pre.name);
        output_impedance_after_delay(r->q, r->width, t_HZ - t_H);
    }

    // Don't forget there's a new state now.
    r->rw_state = fsm->current;

    // Concisely indicate if any pins have clock problems.
summarize:
    if (misclocked_pins != misc) {
        pf("ew.glitch on " YEL "%s" ERC " pin(s) %s",
                r->pre.name, misclocked_pins);
    }
}

// Here we simulate an SRAM part.
void be_sram(void *v)
{
    UNUSED c_sram *r = v;
    int i;

    // Enforce longer of t_KH and t_KL to avoid writing code to check both.
    do_in_clock(&r->ck, t_KL);
    if (!r->ck.rising)
        return;                         // nothing to do
    r->ck.rising = 0;

    // The constant inputs don't concern us, because they concern features
    // not used in our hardware design, and therefore not implemented in
    // this simulation. But sometime XXX before making the circuit board,
    // we want to make sure these pins are grounded. (It just turns out
    // all the correct inputs are low. It didn't have to turn out that way,
    // but it did.) And while ZZ and G# are asynchronous inputs, it's
    // adequate to check when the clock is rising as done here.
    if (check_constant_inputs) {
        for (i = 0; i < 4; i++)
            be_lo(&r->b_[i]);
        be_lo(&r->zz);
        be_lo(&r->g_);
        be_lo(&r->ft_);
        be_lo(&r->lbo_);
    }

    // Rather than look at setup times right away, it's simpler to
    // defer processing until the hold time has passed, because we
    // can do both checks in one step.

    // It's simplest to defer processing until the hold time has passed.
    queue_add(t_H, (qcall) hold_time_callback, r, NULL, 0);
}

/*

FOOTNOTE 1:

    The datasheet says B, except in the case of continuing a deselect cycle,
    where it says D. I think the semantics should be regarded as a "burst of
    deselect," which better represents the control state diagram which
    follows. So I've said B in all cases where CKE# is low and ADV is high.
    It's more straightforward, and the departure in specification does not
    change any behavior.

*/
