/*
    pvio.c

    This is included by c_sram.c and provides a paravirtualized I/O
    subsystem via the electrical simulation. The Dauug House has two
    articles about how this works; they're under

        Code samples -> Paravirtualized I/O
        User instructions -> Simulation

    so I'm not going to write too much about what this does here.

    MWA 8 July 2023
*/

// These command numbers are their tribble values, i.e., clock`t, in8`t
#define PVIO_clock  206930708
#define PVIO_close  206931726
#define PVIO_dec    54156
#define PVIO_demux  221865889
#define PVIO_hex    70561
#define PVIO_in8    75208
#define PVIO_in36   4812998
#define PVIO_len8   5563848
#define PVIO_len36  356085958
#define PVIO_oct    99101
#define PVIO_out8   6416200
#define PVIO_out36  410636486
#define PVIO_read   7135885
#define PVIO_sdec   7394188
#define PVIO_seteof 30307313167L
#define PVIO_setidl 30307328853L
#define PVIO_tet    119709
#define PVIO_unlink 32603710932L
#define PVIO_write  544024398
#define PVIO_nsleep 25171387289L    // consider "not for use programs"

// For each filename mapped to do PVIO, there is one instance of this struct.
typedef struct PVIO_ctx {
    long filename;                  // 
    struct PVIO_ctx *next;
    int fd;                         // open file descriptor or -1
    long eof_marker;                // what an EOF reads as
    long idl_marker;
    int word_mode;                  // 1 (words) or 0 (bytes)
} PVIO_ctx;

static PVIO_ctx *pvio_files;        // linked list of open files
static char *pvio_dirname;          // directory where the files are kept
static long pvio_data, pvio_cmd;    // operands from left and right registers
static int                          // registers these operands come from
        pvio_data_src, pvio_cmd_src;
static long pvio_eof_marker         // EOF marker for future files opened
    = 68719476735L;
static long pvio_idl_marker         // idle marker for future files opened
    = 68719476734L;

static char *tet_digit =
        "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.";

static long read_byte(PVIO_ctx *c)
{
    unsigned char buf[1];

    switch (read(c->fd, buf, 1)) {
        case 1:
            return *buf;
        case 0:
            return c->eof_marker;
        case -1:
        default:
            return c->idl_marker;
    }
}

static long read_word(PVIO_ctx *c)
{
    unsigned char buf[5];
    int i;
    long word;

    switch (read(c->fd, buf, 5)) {
        case 5:
            word = 0;
            for (i=0; i<5; i++)
                word = word << 8 | buf[i];
            return word & 68719476735L;
        case 0:
            return c->eof_marker;
        case -1:
        default:
            return c->idl_marker;
    }
}

static void write_byte(PVIO_ctx *c, long byte)
{
    unsigned char buf[1];

    *buf = byte;
    write(c->fd, buf, 1);
}

static void write_word(PVIO_ctx *c, long word)
{
    unsigned char buf[5];

    word = word & 68719476735L;
    buf[0] = word >> 32;
    buf[1] = word >> 24;
    buf[2] = word >> 16;
    buf[3] = word >> 8;
    buf[4] = word;
    write(c->fd, buf, 5);
}

// Intended for byte mode.
static void write_number(PVIO_ctx *c, long word, long cmd)
{
    char buf[16], *b = buf;

    switch (cmd) {
        default:
            strcpy(buf, "?base?");
            break;

        case PVIO_dec:
            sprintf(buf, "%'lu", word);
            break;

        case PVIO_sdec:
            if (word & 1L << 35)
                word |= comp36;
            sprintf(buf, "%'li", word);
            break;

        case PVIO_oct:
            sprintf(buf, "%012lo`o", word);
            break;

        case PVIO_hex:
            sprintf(buf, "%09lx`h", word);
            break;

        case PVIO_tet:
            buf[0] = tet_digit[word >> 30 & 63];
            buf[1] = tet_digit[word >> 24 & 63];
            buf[2] = tet_digit[word >> 18 & 63];
            buf[3] = tet_digit[word >> 12 & 63];
            buf[4] = tet_digit[word >>  6 & 63];
            buf[5] = tet_digit[word       & 63];
            buf[6] = '`';
            buf[7] = 't';
            buf[8] = '\0';
            break;
    }

    while (*b)
        write_byte(c, *b++);
}

// Called by a "tweak R = pvio=/path/to/pvio_files", this stores said
// path in 'pvio_dirname'. A buffer is allocated for this.
static void pvio_basepath(char *path)
{
    if (pvio_dirname) {
        trackmem(pvio_dirname, 0);
        pvio_dirname = NULL;
    }

    pvio_dirname = trackmem(pvio_dirname, 1 + strlen(path));
    strcpy(pvio_dirname, path);
}

// List the open PVIO register numbers.
static void pvio_diagnostic(void)
{
    PVIO_ctx *c;

    printf("Open PVIO files:");
    for (c = pvio_files; c; c = c->next)
        printf(" %li", c->filename);
    printf("\n");
}

// Shut down and deallocate a PVIO_ctx. It is already unlinked.
static void pvio_done_with(PVIO_ctx *c)
{
    if (c->fd >= 0)
        close(c->fd);
    c->fd = -1;
    trackmem(c, 0);
}

// Find a PVIO-mapped register if it exists.
static PVIO_ctx *pvio_find(long filename)
{
    PVIO_ctx *c;

    for (c = pvio_files; c; c = c->next)
        if (c->filename == filename)
            return c;
    return NULL;
}

// Unmap a PVIO-mapped register. It is not an error to unmap one
// that does not exist. If a file is open, it will be closed.
static void pvio_unmap(long filename)
{
    PVIO_ctx *bye, **byep;

    // This loop makes up to one deletion from a linked list.
    for (byep = &pvio_files, bye = *byep; bye; byep = &bye->next, bye = *byep) {
        if (bye->filename != filename)
            continue;

        // Pull 'bye' out of the list.
        *byep = bye->next;

        // Now clean up and deallocate 'bye'.
        pvio_done_with(bye);
        break;
    }
}

// Create and zero a context for a PVIO mapping onto a new register.
// The caller is responsible for ensuring the register is not already mapped.
// The caller must also fill in the context.
static PVIO_ctx *pvio_add(long filename)
{
    PVIO_ctx *c = trackmem(NULL, sizeof(*c));

    c->filename = filename;
    c->next = pvio_files;
    return pvio_files = c;
}

// Free all PVIO memory for malloc/free tracking purposes at end of program.
void discard_pvio(void)
{
    PVIO_ctx *c;

    while (c = pvio_files) {
        pvio_files = c->next;
        pvio_done_with(c);
    }

    if (pvio_dirname) {
        trackmem(pvio_dirname, 0);
        pvio_dirname = NULL;
    }
}

// This does most of the paravirtualized I/O work. It's called when a
// register is being written after the control decoder turned on the PVIO bit.
long pvio_write(long normal_value)
{
    PVIO_ctx *c;
    int div_len_by = 1;
    long result = 0;
    struct stat st;
    struct timespec ts;

    static int just_copy;
    static long what_to_copy;
    static long write_fname;

    if (!pvio_1g74_was_set)
        return normal_value;

    // Both register files need a copy of whatever PVIO does, so both
    // will call this function when something is happening. Whoever is
    // second to call will simply receive a copy of what the first did.
    if (just_copy) {
        just_copy = 0;
        pvio_1g74_was_set = 0;
        return what_to_copy;
    }

    // For most commands, the data provided is a filename. Most of
    // these will need a full path built from the right register file's
    // "tweak R pvio=/some_path" tweak. The left register file needs
    // a plain "tweak L pvio". The full 'path' is an alloca buffer, and the
    // format is "/some_path/file-0clock" assuming 'pvio_data' is clock`t.
    char *dirname = pvio_dirname ? pvio_dirname : ".";
    char *path = alloca(strlen(dirname) + 32);
    *path = '\0';

    switch (pvio_cmd) {
        case PVIO_in8:
        case PVIO_in36:
        case PVIO_len8:
        case PVIO_len36:
        case PVIO_out8:
        case PVIO_out36:
        case PVIO_unlink: {
            char *p;

            strcpy(path, dirname);
            strcat(path, "/file-");
            p = path + strlen(path);
            *p++ = tet_digit[pvio_data >> 30 & 63];
            *p++ = tet_digit[pvio_data >> 24 & 63];
            *p++ = tet_digit[pvio_data >> 18 & 63];
            *p++ = tet_digit[pvio_data >> 12 & 63];
            *p++ = tet_digit[pvio_data >>  6 & 63];
            *p++ = tet_digit[pvio_data       & 63];
            *p = '\0';
        }
    }

    // Here are the command implementations.
    switch (pvio_cmd) {
        case PVIO_len36:
also_len36:
            div_len_by = 5;
        case PVIO_len8:
also_len8:
            if (stat(path, &st))
                // Treat stat error as if file doesn't exist (-1).
                result = 0xfffffffff;
            else
                result = st.st_size / div_len_by;
            goto done;

        case PVIO_unlink:
            unlink(path);
            goto done;

        case PVIO_clock:
            result = time(NULL);
            goto done;

        case PVIO_nsleep:
            if (pvio_data_src < 512) {      // superuser only
                ts.tv_sec = pvio_data / 1000000000;
                ts.tv_nsec = pvio_data % 1000000000;
                nanosleep(&ts, NULL);
            }
            goto done;

        case PVIO_demux:
also_demux:
            write_fname = pvio_data_src;
            goto done;

        case PVIO_write:
            c = pvio_find(write_fname);
            if (!c || c->fd < 0)
                goto done;
            if (c->word_mode)
                write_word(c, pvio_data);
            else
                write_byte(c, pvio_data);
            goto done;

        case PVIO_read:
            c = pvio_find(pvio_data_src);
            if (!c)
                goto done;
            else if (c->fd < 0)
                result = c->eof_marker;
            else if (c->word_mode)
                result = read_word(c);
            else
                result = read_byte(c);
            goto done;

        case PVIO_close:
            pvio_unmap(pvio_data_src);
            goto done;

        case PVIO_in8:
        case PVIO_in36:
            pvio_unmap(pvio_data_src);
            c = pvio_add(pvio_data_src);
            c->fd = open(path, O_RDONLY | O_NONBLOCK);
            c->eof_marker = pvio_eof_marker;
            c->idl_marker = pvio_idl_marker;
            c->word_mode = pvio_cmd == PVIO_in36;
            goto also_len8;

        case PVIO_out8:
        case PVIO_out36:
            pvio_unmap(pvio_data_src);
            c = pvio_add(pvio_data_src);
            c->fd = open(path, O_WRONLY | O_CREAT | O_APPEND,
                S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
            c->word_mode = pvio_cmd == PVIO_out36;
            goto also_demux;

        case PVIO_seteof:
            pvio_eof_marker = pvio_data;
            goto done;

        case PVIO_setidl:
            pvio_idl_marker = pvio_data;
            goto done;

        case PVIO_dec:
        case PVIO_sdec:
        case PVIO_oct:
        case PVIO_hex:
        case PVIO_tet:
            c = pvio_find(write_fname);
            if (!c || c->fd < 0)
                goto done;
            write_number(c, pvio_data, pvio_cmd);
            goto done;
    }

done:
    just_copy = 1;
    what_to_copy = result;
    if (result == PVIO_seteof)          // never true
        pvio_diagnostic();              // avoid "unused" compiler msg
    return result;
}
