/*
    load.c
    This is the parser for netsim (netlist simulation) scripts.
*/

static void *asm_ctx;                   // assembler context (limit one)

void load(char *fName)
{
    enum { max_bytes = 80, max_args = 5 };

    static char *cmds[] = {
        "- off",
        "+ on",
        "? ?", "? trace",
        "! notrace",
        "a asm",
        "A oldasm",
        "b boring",
        "c c", "c cls",
        "C check",
        "f f", "f from",
        "F fussy",
        "k k", "k kind",
        "L load",
        "n n", "n name",
        "O note",
        "R run",
        "s s", "s signal",
        "t t", "t to",
        "T tweak",
        NULL
    };

    int n_args;                         // correct number of arguments
    int n_have;                         // number of arguments scanned
    int lineNo = 0;                     // for error reports
    int off = 0;                        // in multi-line comment
    int save_boredom;                   // see comments for pf() in util.c

    char buf[max_bytes];                // small line buffer
    char *blanks = "\n\t ";             // token delimiters
    char *loading;                      // strtok_r scratch memory

    char *cmd, *arg[max_args], *a;      // command and argument pointers
    int oper;                           // operation specified by command

    // These have to be preserved from line to line. Be sure to init.
    char cKind[max_bytes];              // component kind
    char cFrom[max_bytes];              // from-component name
    char cTo[max_bytes];                // to-component name

    // These are consumed immediately and don't need copied.
    char *pFrom, *pTo;                  // from-pin and to-pin names
    char *note;                         // text of note
    int ps;                             // from-to delay in picoseconds

    FILE *f;                            // stream being loaded

    long *image;                        // code RAM contents
    int n_image;                        // code RAM size

    *cKind = *cFrom = *cTo = '\0';      // sane init

    f = fopen(fName, "r");
    if (!f)
        goto h;

    while (fgets(buf, max_bytes, f)) {
        ++lineNo;

        // delete anything from comment symbol ';' rightward
        buf[strcspn(buf, ";")] = '\0';

        // get command from line; only first char matters
        cmd = strtok_r(buf, blanks, &loading);
        if (!cmd)
            continue;                   // blank line or comment

        // fold command from keyword to a single character
        oper = *wesKeyword(cmd, cmds, "2");

        // multi-line comment?
        if (off && oper != '+')
            continue;

        // There is a free-form "note" command to display test info.
        // We just grab the rest of the input string (without breaking).
        if (oper == 'O') {
            note = strtok_r(NULL, "\n", &loading);
            if (note)
                pf("." GGRE "=== %s ===" BAW "\n", note);
            else
                pf(".\n");
            continue;
        }

        // Get args after command up to some limit.
        for (n_have = 0; n_have < max_args; ++n_have)
            ifnot (arg[n_have] = strtok_r(NULL, blanks, &loading))
                break;

        // See how many args the command calls for, and verify we have 
        // that number. The wacky string below is simply commands at
        // even bytes and number of arguments on the odd bytes.
        // n_args == 10 means there is one optional argument.
        // n_args == 11 means the command does not exist.
        n_args = wesCharSub(oper, "b1f1t1s3k1n1?1!1R:C3T2U1-0+0c0L1a1A1F0b;") - '0';
        if (n_args == 11) {
            pf("e.%s : %i unknown command", fName, lineNo);
            continue;
        }
        else if (n_args == 10 && (n_have == 0 || n_have == 1)
                || n_args == n_have)
            ;
        else {
            pf("e.%s : %i wrong # of args", fName, lineNo);
            continue;
        }

        switch (oper) {

            default:
                pf("e.%s : %i known but unimplemented command", fName, lineNo);
                continue;

            case 'k':                   // k <component-kind>
                strcpy(cKind, arg[0]);
                continue;

            case 'n':                   // n <new-component-name>
                new_component(arg[0], cKind);
                continue;

            case '?':                   // ? <component-name>
                enable_trace(arg[0], 1);
                continue;

            case '!':                   // ! <component-name>
                enable_trace(arg[0], 0);
                continue;

            case 'f':                   // f <from-component-name>
                strcpy(cFrom, arg[0]);
                continue;

            case 'F':                   // fussy
                fussy = 1;  
                continue;

            case 't':                   // t <to-component-name>
                strcpy(cTo, arg[0]);
                continue;

            case 's':                   // s <from-pin> <to-pin> <delay-ps>
                pFrom = arg[0];
                pTo = arg[1];
                if (strlen(arg[2]) != strspn(arg[2], "0123456789")) {
                    pf("e.%s : %i invalid delay", fName, lineNo);
                    continue;
                }
                ps = atoi(arg[2]);
                if (ps < 1) {
                    pf("e.%s : %i positive delay required", fName, lineNo);
                    continue;
                }
                connect_pins_by_name(cFrom, pFrom, cTo, pTo, ps);
                continue;

            case 'R':                   // run (with optional quit time)
                if (n_have)
                    queue_quit(number(arg[0], "lk0"), ".ns script");
                queue_run();
                continue;

            case 'C':                   // check <comp> <out-pin> <state>
                check_output_pin_state(arg[0], arg[1], arg[2]);
                continue;

            case 'T':                   // tweak
                tweak_component(arg[0], arg[1]);
                continue;

            case '-':                   // comment out
                off = 1;
                continue;

            case '+':                   // comment in
                off = 0;
                continue;

            case 'c':                   // clear display and scrollback
                printf(".\x1b" "c");    // no pf(), or may erase a timestamp
                fflush(stdout);
                continue;

            case 'L':                   // 'load' does NOT reinitialize,
                save_boredom = boredom; // but is a continuation of
                load(arg[0]);           // what already is happening.
                boredom = save_boredom;
                continue;              

            case 'a':                   // asm
                // Free whatever was assembled previously (if anything)
                // and assemble a new code RAM from the file specified.
                // Then find the code RAM's memory (you have to have
                // already created the code RAM in the script) and load
                // it with what the assembler did.
                as_dealloc();

                asm_dealloc(asm_ctx);
                pf(".assembling %s (new assembler)\n", arg[0]);
                if (as_assemble_from_filename(arg[0]))
                    pf("e!.assembly failed");
                else {
                    get_firmware("C", NULL, &n_image, (void **) &image);
                    as_output_program_code((void *) image, n_image);
                }
                continue;

            case 'A':                   // oldasm
                // Free whatever was assembled previously (if anything)
                // and assemble a new code RAM from the file specified.
                // Then find the code RAM's memory (you have to have
                // already created the code RAM in the script) and load
                // it with what the assembler did.
                asm_dealloc(asm_ctx);
                pf(".assembling %s (old assembler)\n", arg[0]);
                asm_ctx = asm_assemble(arg[0], "fl", NULL);
                get_firmware("C", NULL, &n_image, (void **) &image);
                asm_fetch_code(asm_ctx, image, n_image);
                continue;

            case 'b':                   // boring
                // This is a kind of "debug level" that can suppress
                // output that I ordinarily would not want to read.
                // The higher the "boring" argument, the less I want to
                // read what may be printed. Negative arguments, including
                // -0, are used offset boredom during debugging.
                a = arg[0];
                if (*a == '-')
                    anti_boredom = number(a+1, "0");
                else
                    boredom = number(a, "0");
                continue;
        }
    }

    if (ferror(f) || fclose(f))
        goto h;

    return;
h:
    pf("e.error '%s' reading '%s'", strerror(errno), fName);
}

// This cleans up memory after use of the assembler.
void discard_assembly(void)
{
    asm_dealloc(asm_ctx);
    asm_ctx = NULL;
}
