/*
    ns.c
    This is the root source file for the netlist simulation (netsim).

    To simplify reading this source code by newcomers and finding
    things by gurus, this netsim is divided into many files according
    to plausible but maybe not optimal concept boundaries.

    Those files are catenated into one translation unit, which speeds
    compilation and eliminates much need for dependency checking. I
    recognize that some maintainers would prefer a more "traditional"
    build scenario. Every approach has its benefits and drawbacks.
*/

#include <sys/types.h>              // pvio.c uses stat(), open(), etc.
#include <sys/stat.h>               // pvio.c uses stat(), open(), etc.
#include <fcntl.h>                  // pvio.c uses open()
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>                   // for pvio.c: time() and nanosleep()
#include <unistd.h>                 // symlink()
#include <locale.h>                 // setlocale()
#include <alloca.h>                 // pvio.c uses alloca()

#include "../consts/opcode.h"       // c_sram.c uses oc_HALT
#include "../old-asm/asm.h"         // how to access old assembler
#include "../cross-asm/as.h"        // how to access new assembler
#include "../misc/misc.h"           // shared with VM and assembler
#include "../firmware/firmware.h"   // get_firmware()
#include "compo.h"                  // building blocks for pins and components

#include "../misc/misc.c"           // shared with VM and assembler
#include "util.c"                   // netsim-specific tidbit functions
#include "pin.c"                    // simulates pins
#include "comps/c_1g74.c"           // part: single D flip-flop
#include "comps/c_244.c"            // part: 16-bit (quad quad) buffer
#include "comps/c_2gxx.c"           // part: dual BUF INV AND OR NAND NOR XOR
#include "comps/c_374.c"            // part: 16-bit (dual octal) D flip-flop
#include "comps/c_io.c"             // part: bus I/O for simulations only
#include "comps/c_osc.c"            // part: 80 MHz crystal oscillator
#include "comps/c_pull.c"           // part: 10k resistor (not used or sim'd)
#include "comps/c_reset.c"          // part: generic voltage monitor
#include "comps/c_sram.c"           // part: synchronous static RAM
#include "comps/c_switch.c"         // part: switch for simulations only
#include "comps/c_pulse.c"          // part: clock phase output for sim'ns only
#include "connect.c"                // code to hook components together
#include "queue.c"                  // container for priority queue
#include "quser.c"                  // simulator's interface to priority queue
#include "load.c"                   // read and perform a *.ns script

static void help() {
    printf("%s",
"Usage: ns [OPTION]... [SCRIPT]\n"
"Notable uses and options:\n\n"
"  ns -h        display this help and exit\n"
"  ns -l        list opcodes by number and exit\n"
"  ns SCRIPT    open and run SCRIPT (usually has .ns suffix)\n"
"  ns           re-run most recent script\n"
    );
}

static void list_opcodes() {
    int i;
    for (i=0; i<128; i++)
        printf("%3i %7s    %3i %7s    %3i %7s    %3i %7s\n",
                i, opcode_name(i), i+128, opcode_name(i+128),
                i+256, opcode_name(i+256), i+384, opcode_name(i+384));
}


/*
    The netsim executable 'ns' takes at most one argument, which is
    the name of a script to run. If no script is specified, the
    most recently run script is re-run, as determined by a symlink
    'last-test.ns' that is managed by this program.
*/
int main(int argc, char **argv)
{
    char *fname = "last-test.ns";
    int opt;

    while ((opt = getopt(argc, argv, "lh")) != -1) {
        switch (opt) {
            case 'l':
                list_opcodes();
                return 2;
            case 'h':
            default:
                help();
                return 1;
        }
    }

    if (argc == optind + 1) {       // keep link to most recently run script
        unlink(fname);
        symlink(argv[optind], fname);
        fname = argv[optind];
    }

    setlocale(LC_NUMERIC, "");      // pf ' modifier in queue_run()

    if (sizeof(int) != 4 || sizeof(long) != 8)
        pf("e!.word sizes are not as thought");

    if (oc_J == 0)
        pf("e!.J opcode must not be 0 due to Osmin code pool terminator");

    queue_setup();                  // create queue for simulation's events
    load(fname);                    // start reading the script
    fussy = 0;                      // the fire report is not fatal
    fire_report();                  // write out short circuit durations

    discard_queue();                // free the event queue's memory
    discard_assembly();             // free memory used by assembler
    discard_all_firmware();         // free the firmware's memory
    discard_netlist();              // free the netlist's memory
    discard_pvio();                 // free all PVIO mappings
    trackmem_report();              // account for all heap memory

    // Some scripts test for expected outcomes. Were there failures?
    if (n_fail) {
        pf(".%s - ", fail());    // exaggerates n_fail by 1
        pf(".%i good and %i failed tests\n", n_good, n_fail - 1);
        return 1;
    }

    // Some scripts test for expected outcomes. Were there *no* failures?
    if (n_good) {
        pf(".%s - ", good());    // exaggerates n_good by 1
        pf(".all %i tests were successful\n", n_good - 1);
    }

    return 0;
}
