/*
    quser.c

    This provides a time-based priority queue for the netlist simulation
    with picosecond granularity. Everything transpires using a single
    queue, with callbacks and arguments determining what happens as
    entries expire out of the queue. The underlying container is in
    queue.c, which was adapted from weslib.
*/

wesQueue Q;                 // the queue itself
static int queue_running;   // cleared by quit_callback()

// Here is a single priority queue entry for the time element of simulation.
// See also queue.c's documentation.
typedef struct q_ent {
    time_ps when;           // when item comes out of queue
    qcall callback;         // function to call when we come out
    void *v;                // pointer data for called function
    void *w;                // more pointer data for called function
    int i;                  // integer data for called function
} q_ent;

// This function determines time order within the priority queue.
// See also queue.c's documentation.
static int q_ent_less_than(void *vl, void *vr)
{
    q_ent *l = vl, *r = vr;

    return l->when < r->when;
}

// Initialize the simulation queue.
void queue_setup(void)
{
    wesQueueInit(&Q, sizeof(q_ent), q_ent_less_than);
}

// Free the simulation queue.
void discard_queue(void)
{
    trackmem(Q.first, 0);
    memset(&Q, 0, sizeof(Q));
}

static void quit_callback(char *reason, void *unused, int ignored)
{
    pf(".quit for %s\n", reason);
    queue_running = 0;
}

// Schedule something to happen some (simulated) picoseconds from now.
void queue_add(int ps_delay, qcall callback, void *v, void *w, int i)
{
    q_ent e;

    if (ps_delay < 0)
        pf("e!.queue request for past action at %i ps", ps_delay);

    e.when = now + ps_delay;
    e.callback = callback;
    e.v = v;
    e.w = w;
    e.i = i;
    wesQueueInsert(&Q, &e);
}

// Schedule a quit to happen.
// Right now, this happens at an absolute time. May prefer relative time later.
// `when` will be silently clamped to be in the future; otherwise, the
// 'time did not advance' error can appear when a script ends with 'quit 0'.
// If other events are to happen the at the same time as a quit, the semantics
// of queue_run() will process all other events prior to the quit.
void queue_quit(time_ps when, char *reason)
{
    q_ent e;

    e.when = when <= now ? now + 1 : when;
    e.callback = (qcall) quit_callback;
    e.v = reason;
    e.w = NULL;
    e.i = 0;
    wesQueueInsert(&Q, &e);
}

/*
    Run the queue by popping events out in their time sequence and
    simulating them. Stop when the queue runs out of events, because
    nothing will ever happen again. Also stop when the user's time
    limit is reached, as specified with the 'quit' keyword in the
    *.ns script.
*/
void queue_run(void)
{
    q_ent *e;

    every_pin_is_dirty();
    queue_running = 1;

    while (queue_running) {             // loop over timesteps
        all_components_respond();

        if (!Q.n) {
            pf(".simulation queue empty\n");
            break;
        }

        // Advance to NEXT timestep.
        e = Q.first;

        if (e->when < now)              // XXX remove sanity check someday
            pf("e!.time is not monotonic");
        if (e->when == now)             // XXX remove sanity check someday
            pf("e.time did not advance");
        now = e->when;

        while (Q.n) {                   // loop over events within timestep

            e = Q.first;

            // stop after this timestep is finished
            if (e->when != now)
                break;

            // do whatever the first element says
            e->callback(e->v, e->w, e->i);

            // remove the first element (there should be more)
            wesQueueRemoveHead(&Q);
        }
    }
}
