/*
    util.c
    These minor tidbit functions are used only by netsim.
*/

#define ERC TEA                         // "error color"

// Here is the simulator's elapsed time, measured as whole picoseconds,
// and a provision to end the simulation.
time_ps now;

// XXX deprecated
time_ps stable_at = 0;                  // to quiet stablization msgs

int boredom, anti_boredom;              // selective message suppression
int fussy;                              // any error is visible and fatal

char *watching = "";                    // not watching anything yet

static int n_good, n_fail;              // regression test counters

/*
    This wraps long text to suit me. It doesn't consider whitespace
    either to determine where to split lines or to avoid duplicating
    a trailing \n.
*/
static void wrap(
    char *s,
    int first_line_indent,              // suppressed for empty strings
    int other_line_indent,
    int wrap_width,
    int wrap_marker                     // %c before and after line breaks
) {
    int width_left;
    int active_marker = 0;

    // May take a few printings.
    while (*s) {
        width_left = wrap_width;

        if (first_line_indent) {        // indent for (non-)first line
            printf("%*s", first_line_indent, "");
            width_left -= first_line_indent;
        }

        // If requested, show delimeter at start of continuation lines.
        if (active_marker)
            width_left -= printf("%c", active_marker);

        // Print as much as would fit.
        s += printf("%.*s", width_left, s);

        // If the whole string has been printed, stop.
        if (!*s)
            break;

        // If requested, show delimeter at end of breaking lines.
        if (wrap_marker) {
            active_marker = wrap_marker;
            printf("%c", active_marker);
        }

        // Break the line.
        printf("\n");

        first_line_indent = other_line_indent;
    }
}

/*
    This is a general output routine that wraps printf() with supplemental
    functionality. The format string REQUIRES a prefix of characters that
    map to added capabilities, as follows:

    e   Indicate this as an error via color highlighting. A newline is added.
    !   Exit the program after printing.
    w   Wrap lines to suit 'c_sram.c'.
    .   End of prefix. All format strings must have this period.

    Furthermore, if 'now' is not negative (indicating that the simulation
    has indeed begun), a picosecond timestap is displayed to the right of the
    first line with any changed time. This is accomplished by printing a line
    with the time that terminates with a carriage return but no newline.

    We don't count the number of characters output due to semantic questions.
    (For example, would escape sequences count?)

    BOREDOM
    -------

    Another functionality of this routine is suppression of "boring" output
    such as known startup errors. This uses the 'boring' keyword without
    a - sign to adjust the 'boredom' zero (typically zero), or with a -
    sign to adjust 'anti_boredom' (typically zero). pf() output is suppressed
    if the net boredom (boredom - anti_boredom) is positive.

    The 'load' keyword saves and restores boredom, allowing a .ns file that
    generates unimportant messages to suppress them without affecting any files
    that load it. But anti_boredom is not saved or restored, allowing an outer
    .ns file to add verbosity throughout when debugging.

    NOTE
    ----

    I set the default 'wrap_width' to 2**20 effective 21 July 2023.
    It used to be INT_MAX, but I ran into a libc somewhere that has
    a "%.*s" printf conversion (see wrap() above) that doesn't output
    any characters if the width is INT_MAX. No one else should have
    to figure that out.
*/
void pf(char *fmt, ...)
{
    static time_ps last_stamp = LONG_MIN;

    char *copy_fmt = fmt, *buf = NULL;
    int is_err = 0;
    int is_fatal = 0;
    int buf_size = 1024, needed_len;
    int first_line_indent = 0;
    int other_line_indent = 0;
    int wrap_width = 1 << 20;       // See NOTE above.
    int wrap_marker = 0;
    va_list ap;

    // Scan format string for anticipated capabilities.
    while (1) {
        switch (*fmt++) {
            case '.':
                goto prefix_ended;

            case 'e':
                is_err = 1;
                if (fussy)
                    is_fatal = 1;
                break;

            case '!':
                is_fatal = 1;
                break;

            case 'w':
                first_line_indent = 0;
                other_line_indent = 4;
                wrap_width = 72;
                wrap_marker = '|';
                break;

            default:
                // Can't continue this string, because 'fmt' may be past a null.
                pf("e.illegal pf(\"%s\", ...) encountered", copy_fmt);
                return;
        }
    }
prefix_ended:

    // Suppress non-fatal messages if they are boring.
    if (!is_fatal && boredom - anti_boredom > 0)
        return;

    if (now >= 0) {
        if (now != last_stamp) {
            fflush(stdout);     // occasionally helpful
            printf("%'88li ps\r", now);
            last_stamp = now;
        }
    }

    if (is_err)
        printf(ERC);           // color highlight for errors

    // Assume we don't have aprintf. We need to put this string into a
    // buffer, because line wrapping may happen.
    while (1) {
        buf = trackmem(buf, buf_size);
        va_start(ap, fmt);
        needed_len = 1 + vsnprintf(buf, buf_size, fmt, ap);
        va_end(ap);
        if (needed_len > buf_size) {
            buf_size = needed_len;
            continue;
        }
        break;
    }

    wrap(buf, first_line_indent, other_line_indent, wrap_width, wrap_marker);
    trackmem(buf, 0);

    if (is_err)
        printf(BAW "\n");      // for errors, end color and add newline

    if (is_fatal) {
        printf(ERC "terminating due to immediately preceding error" BAW "\n");
        exit(1);
    }
}

/*
    Works a little like "assq" in Scheme.  The list is an alternating
    series of "from" and "to" bytes. Zeros may appear anywhere.

    IF you care about catching overruns when 'from' does not appear in the
    list, terminate the list using whatever you used first for 'from', and
    put the default value in the 'to'.

    Example:  wesCharSub('b', "a#b&c)a?") returns '&' and would have had
    default value '?'.
*/
int wesCharSub(int from, const char *associationList)
{
    const char *s = associationList;
    int stopAt = *s;

    while (1) {
        if (from == *s || *s == stopAt && s != associationList)
            return s[1];
        s += 2;
    }
}

/*
	Find 'word' in 'list', which must end with NULL.  Example:

		char **list = { "@ amp", "# hash", "^ carat", NULL };

		switch (*wesKeyword("HASH", list, "2i")) {
			case '\0':		// word is not in list
			case '@':		// word is "amp"
			case '#':		// word is "hash"
			case '^':		// word is "carat"
			default:		// word is something you added to list
		}

	'options' is a string of character flags, may be:

	'i'  make comparison case-insensitive
	'n'  return NULL instead of "" if there is no match
	'0'
	'1'
	...  any digit (default 1) indicates byte offset of search string;
	     caller may use prior bytes to specify behavior when matched
*/
char *wesKeyword(char *word, char **list, char *options)
{
	int prefixBytes = 1;	// bytes to skip in each list item
	int null = 0;			// return NULL instead of "" if not found?
	int (*cmp)				// comparison function
		(const char *, const char *) = strcmp;

	for (; *options; ++options) {
		if (isdigit(*options))
			prefixBytes = *options - '0';
		else if (*options == 'i')
			cmp = strcasecmp;
		else if (*options == 'n')
			null = 1;
	}

	for (; *list; ++list)
		if (!cmp(word, prefixBytes + *list))
			return *list;

	return null ? NULL : "";
}

/*
    Allocate and copy some memory.
    Pass -1 to get length from strlen().
*/
void *copy(void *mem, int len)
{
    void *buf;

    if (len < 0)
        len = 1 + strlen(mem);

    buf = trackmem(NULL, len);
    memcpy(buf, mem, len);
    return buf;
}

/*
    Catenate and copy three strings. You may not free this.
    FIXME 9 Nov 2022 it uses trackmem, should be quite ok to free.
*/
char *cat(char *s, char *t, char *u)
{
    int len;
    char *buf;

    if (!s) s = "";
    if (!t) t = "";
    if (!u) u = "";

    len = strlen(s) + strlen(t) + strlen(u) + 1;
    buf = trackmem(NULL, len);

    strcpy(buf, s);
    strcat(buf, t);
    strcat(buf, u);

    return buf;
}

/*
    Watch a byte to see when it changes.
    It's like a memory breakpoint, for callback debugging.
*/
void watch(char *alarm)
{
    static int old;

    if (*watching != old) {
        old = *watching;
        pf("e.%li ps: %s (byte is now %i)", now, alarm, *watching);
    }
}

/*
    This is a range-checking convenience function for parsing a string
    that represents an integer. The string must be \0 terminated.
    
    If 'parms' is empty, the number must be positive and fit in an int.
    If 'parms' contains '-', the number may be negative.
    If 'parms' contains '0', the number may be zero.
    If 'parms' contains 'l', the number may fit in a long.
    If 'parms' contains 'k', a kludgy in-place deletion of _'s precedes parsing.

    This does not detect an out-of-range long, where strtol()
    sets 'errno' to ERANGE. Providing such a test is not a priority.
*/
long number(char *s, char *parms)
{
    long l;
    char *endp;

    if (!parms)
        parms = "";

    // XXX This in-place deletion of _ is begging for trouble.
    if (strchr(parms, 'k')) {
        char *from = s, *to = s;
        do {
            if (*from != '_')
                *to++ = *from;
        } while (*from++);
    }

    l = strtol(s, &endp, 10);
    if (endp == s || *endp)
        pf("e!.invalid number '%s'", s);

    if (!strchr(parms, '-') && l < 0)
        pf("e!.negative number not allowed '%s'", s);

    if (!strchr(parms, '0') && l == 0)
        pf("e!.zero number not allowed '%s'", s);

    if (!strchr(parms, 'l') && l != (int) l)
        pf("e!.number does not fit in int '%s'", s);

    return l;
}

/*
    This is a constant "OK" string within a green ANSI terminal
    control sequence, with a side effect of incrementing a counter.
    It's a convenience function for diagnostic printfs.
*/
char *good(void)
{
    ++n_good;
    return GRE "OK" BAW;
}

/*
    This is a constant "NG" string within a red ANSI terminal
    control sequence, with a side effect of incrementing a counter.
    It's a convenience function for diagnostic printfs.
*/
char *fail(void)
{
    ++n_fail;
    return RED "NG" BAW;
}
