/*
    asm.c
    This assembles Dauug|36 instructions and optionally interprets
    the machine code in a small VM for software testing.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "../consts/opcode.h"
#include "../consts/slots.h"
#include "../firmware/firmware.h"
#include "../misc/misc.h"
#include "./context.h"

#include "errors.c"
#include "syms.c"
#include "lex.c"
#include "syntax.c"
#include "alu.c"

// Reset program. Provided automatically after assembly only.
void asm_reset(void *context)
{
    X = context;

    preset_regs(X->sym_root);   // zero registers (no longer preset constants)
    X->ip = 0;
    X->flags = 0;
    X->n_stack = 0;
}

// Assemble from source already attached to X.
// If any errors, eMsg_head will have the list.
static void assemble_from_context(void)
{
    int pass, parsing, i;
    long code;
    char *s;
    struct eMsg **last_err;

    X->n_regs = 0;
    X->code_size = 4096;                // MWA arbitrary, should be adjustable
    X->preamble_size = 100;             // MWA arbitrary, should be adjustable

    for (pass = 1; pass < 3; ++pass) {
        parsing = 1;
        X->cursor = X->preamble_size;
        s = X->src;

        if (pass == 2) {
            // Now that we have code memory, go ahead and write the
            // immediate preamble at the start of it.
            immediate_preamble(X);
        }

        while (parsing) {
            last_err = X->eMsg_tail;    // Notice if error message issues.
            code = -1;                  // This only changes if parse succeeds.

            if (newline(&s)) {          // Blank lines are legal and frequent.
                if (!*s) {
                    code = (long) oc_HALT << 27;
                    parsing = 0;        // how parsing terminates
                }
            }

            else if (label(&s, pass))
                ;                       // keep parsing after label

            else if (assignment_statement(&s, pass, &code))
                ;                       // something = ...

            else if (keyword_statement(&s, pass, &code))
                ;                       // keyword ...

            else {
                if (last_err == X->eMsg_tail)
                    note_parse_error(s, "syntax error", NULL, NULL);
                flush_eol(&s);          // An error message has issued.
            }

            emit_code(X, code);         // add a word to code memory
        }

        if (pass == 1) {

            // Guarantee that a "zero register" has been allocated.
            // This will provide the constant zero while the program is
            // running, but in the immediate preamble will provide a
            // temporary register for forming 36-bit integer constants.
            struct reg_name zeroReg = { "0" };
            int ignored = 0;
            reg_to_sym(&zeroReg, "nowhere", &ignored);

            // No more registers will be allocated. Number them all
            // in lexicographic order.
            number_regs(X->sym_root);

            // Ok to allocate registers now.
            X->regs = trackmem(NULL, X->n_regs * sizeof(*X->regs));

            // Allocate code and fill with NOLIST.
            // MWA Kludge: This allocation is used to detect pass 2 elsewhere.
            X->code = trackmem(NULL, X->code_size * sizeof(*X->code));
            for (i=0; i<X->code_size; i++)
                X->code[i] = (long) oc_NOLIST << 27;

            if (X->eMsg_head)
                break;
        }
    }

    if (!X->eMsg_head)
        asm_reset(X);                   // ready to execute immediately
}

// Deallocate assembler context.
void asm_dealloc(void *context)
{
    X = context;
    if (!X) return;
    clear_symbol_table(&X->sym_root);
    clear_parse_errors();
    trackmem(X->code, 0);
    trackmem(X->regs, 0);
    if (X->free_src) {
        trackmem(X->src, 0);
        X->src = NULL;
    }
    trackmem(X, 0);
    X = NULL;
}

// Run program (with lots of sanity checking).
// Note this is a *logical* simulation of the CPU. The 'netsim' discrete
// event simulator can do an *electrical* simulation of the CPU that can
// execute programs more realistically, but does not involve this function.
void asm_run(void *context)
{
    long inst, y, l, r;                 // instruction, register *values*
    int opcode, yn, ln, rn;             // opcode, register *numbers*
    int maxreg;                         // a bound

    X = context;
    maxreg = X->n_regs;

    // Make sure some hooligan didn't set invalid register bits.
    for (ln = 0; ln < maxreg; ++ln)
        if (X->regs[ln] & comp36)
            ng("!user goofed up register %i", ln);

    // run now
    while (1) {
        // Make sure IP points to code (no crazy branch).
        if (X->ip < 0 || X->ip >= X->code_size) {
            ng("asm_run branch outside code (%i)", X->ip);
            return;                     // bad, but no exit()
        }

        inst = X->code[X->ip++];        // read inst
        opcode = inst >> 27 & 511;      // decode opcode

        if (0) {
            printf("%4i  ", X->ip - 1);
            asm_disassemble(RED, inst, BAW "\n");
        }

        if (opcode < oc_NO_MORE_ALU) {  // This is an ALU operation.

            yn = inst >> 18 & 511;      // decode register numbers
            ln = inst >> 9 & 511;
            rn = inst & 511;

            if (yn >= maxreg || ln >= maxreg || rn >= maxreg)
                ng("!asm_run decoded invalid register");

            l = X->regs[ln];            // fetch register values
            r = X->regs[rn];

            // Do the ALU operation here.
            y = alu(opcode, l, r, &X->flags);
            if (y & comp36)
                ng("!asm_run ALU returned extraneous bits");

            // Write the result back.
            switch (opcode) {
/*
    TODO This switch block needs to be replaced by using the decoded
    instruction. But as of 29 April the hardware design does not
    accommodate reading G without writing the result to L and R.
    The gamma lockout has to be expanded, and simulation will be
    needed to ensure nothing breaks when that change is made.
*/
                case oc_CRF:
                case oc_BOUND_u:
                case oc_BOUND_s:
                case oc_CMP_uu:
                case oc_CMP_us:
                case oc_CMP_su:
                case oc_CMP_ss:
                    break;
                default:
                    X->regs[yn] = y;
            }

            // Trace the XL operation as diagnostic output.
            if (opcode == oc_XL)
                trace_reg(X->sym_root, ln);

            continue;
        }

        // This is a non-ALU operation.
        switch (opcode) {
            case oc_HALT:                       // HALT
                return;

            case oc_NOLIST:                     // NOLIST
            case oc_NOP:                        // NOP
                break;

            case oc_C:                          // CALL
                if (X->n_stack >= max_call_depth) {
                    ng("call depth too deep");  // bad, but no exit()
                    return;
                }
                X->stack[X->n_stack++] = X->ip;
                X->ip = inst & 0777777777;
                break;

            case oc_R:                          // RETURN
                if (X->n_stack <= 0) {
                    ng("return without call");  // bad, but no exit()
                    return;
                }
                X->ip = X->stack[--X->n_stack];
                break;

            // immediate loads reverified 28 March 2023, see file "imm.test"
            case oc_IMP:                        // load immediates
            case oc_IMN:
            case oc_IMH:
            case oc_IMB:
                yn = inst >> 18 & 511;      // decode register numbers
                rn = inst & 511;

                l = inst;
                r = X->regs[rn] & mask36;   // fetch register values

                // Do the ALU operation here.
                y = alu(opcode, l, r, &X->flags);
                if (y & comp36)
                    ng("!asm_run ALU returned extraneous bits");

                // Write the result back.
                X->regs[yn] = y;
                break;

            // (un)conditional jumps
            case oc_J:
            case oc_J_lt:
            case oc_J_le:
            case oc_J_eq:
            case oc_J_ne:
            case oc_J_ge:
            case oc_J_gt:
            case oc_J_r:
            case oc_J_nr:
            case oc_J_t:
            case oc_J_nt:
                y = alu(opcode, 0, 0, &X->flags);
                if (X->flags & bit5)
                    X->ip = inst & 0777777777;
                break;

            default:                            // any of several JUMPs
                ng("!bonkers opcode %i in asm_run()", opcode);
                break;
        }
    }
}

/*
    Assemble from source code. Returns an opaque pointer.

    flags:

        b   Print a banner message before doing anything.
        c   Show changes to registers after execution (requires 'r').
        d   Deallocate everything and return NULL when done.
        f   Supplied 'source' is a filename, not a buffer.
        l   Provide a listing.
        r   Run what is assembled.

    TODO If assembly fails in Pass 1, memory won't be attached to
    the context, and execution will segfault.
    XXX DON'T FIX UNTIL PROGRAM CAN BE RUN FOR TESTING.
*/
void *asm_assemble(char *source, char *flags, char *banner)
{
    char *fName;

    if (!flags)
        flags = "";

    // set up context to work in
    X = trackmem(NULL, sizeof(*X));
    X->eMsg_tail = &X ->eMsg_head;
    X->src = source;
    X->list = !!strchr(flags, 'l');

    // print signon message if called for
    if (strchr(flags, 'b'))
        puts(banner);

    // load file if called for
    if (strchr(flags, 'f')) {
        if (X->src = slurp(fName = X->src))
            X->free_src = 1;
        else {
            ng("cannot open assembler source '%s'", fName);
            X->src = "";                        // bad, but no exit()
        }
    }

    // assemble and reset
    if (X->list)
        printf("\n--- code ---\n");
    assemble_from_context();

    // list symbols
    if (X->list && X->sym_root) {
        printf("\n--- symbols ---\n");
        show_symbol_table(X->sym_root);
    }

    // list assembly errors
    if (X->eMsg_head) {
        printf("\n--- errors ---\n");
        show_parse_errors(X->src);

        // XXX DON'T FIX UNTIL PROGRAM CAN BE RUN FOR TESTING.
        // XXX Kludge: Assembly errors terminate the program.
        // Otherwise, this would segfault due to regs & code
        // not attached to the context. Want to straighten all
        // this out someday.
        ng("!terminating");
    }

    // run
    else if (strchr(flags, 'r')) {
        printf("\n--- running ---\n");
        asm_run(X);
        if (strchr(flags, 'c')) {       // show regs that changed
            printf("\n--- changed registers ---\n");
            show_changed_regs(X->sym_root);
            printf("\n");
        }
    }

    // dealloc
    if (strchr(flags, 'd')) {
        asm_dealloc(X);
        X = NULL;
    }

    return X;
}

// Return pointer to specified register. Will ng() if no such name.
// asm_run() will ng() if the register is abused by caller.
long *asm_get_register(void *context, char *name)
{
    static long bogus_register;

    X = context;
    struct sym *s = find_symbol(name, NULL, 0, 1);
    if (!strchr("su", s->what)) {
        ng("%s is not a register", name);
        return &bogus_register;                 // bad, but no exit()
    }
    if (s->number < 0 || s->number >= X->n_regs) {
        ng("%s is for invalid register %i", name, s->number);
        return &bogus_register;                 // bad, but no exit()
    }
    return X->regs + s->number;
}

// Returns pointer to flags.
int *asm_get_flags(void *context)
{
    X = context;
    return &X->flags;
}

/*
    Given a code RAM in 'out' and its size in 'rows' from netsim, this
    function copies assembled code to the RAM so it can be tested.
    Returns the number of words copied.
*/
int asm_fetch_code(void *context, long *out, int rows)
{
    int i;

    X = context;

    if (X->code_size > rows)
        ng("too much code: %i pigeons into %i holes", rows, X->code_size);
    else
        rows = X->code_size;

    for (i = 0; i < rows; i++)
        out[i] = X->code[i];

    return rows;
}
