/*
    lex.c
    Assembler terminals and low-level parsing
*/

const char *tetrasexagesimal = "0123456789abcdefghijklmnopqrstuvwxyz"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ@$";

/*
    This swallows spaces, tabs, and comments from ; to end of line.
    It does NOT eat wild newlines, as we are an assembler.

    It also swallows parenthesized text, including newlines, where
    the number of consecutive closing parentheses must match the
    number which opened. So (in-line comment) as well as (multi-line
    comment) and ((comment with encapsulated parenthetical remarks
    (and other interesting discourses) get swallowed too)).

    This routine only needs to be called within terminal symbols;
    e.g., by code that examines **s.
*/
static void eatBlanks(char **s)
{
    int lParens, rParens, saw_semi = 0;
    char *began_on;

    while (**s && **s != '\n') {

        if (**s == ' ' || **s == '\t' || saw_semi) {
            ++*s;
            continue;
        }

        if (**s == ';') {
            saw_semi = 1;                   // detect ; comment
            continue;
        }

        lParens = 0;                        // detect stream of (
        while (**s == '(')
            ++*s, ++lParens;
        if (!lParens)
            break;                          // NOT blank, tab, semi, (
        began_on = *s;                      // line comment began on

        while (1) {
            while (**s && **s != ')')       // skip to first )
                ++*s;

            rParens = 0;                    // count number of )
            while (**s == ')')
                ++*s, ++rParens;

            if (rParens == lParens)         // end of (( comment ))
                break;

            if (!**s) {                     // error: EOF in (( comment
                note_parse_error
                    (began_on, "end of file in ( comment", NULL, NULL);
                break;
            }
        }
    }
}

/*
    Detect newline and advance parser past it.
    An implicit newline is at the end of the input.
*/
static int newline(char **ss)
{
    char *s = *ss;

    eatBlanks(&s);
    if (*s) {
        if (*s != '\n')
            return 0;
        ++s;
    }
    *ss = s;
    return 1;
}

/*
    Flush through newline. Used to continue assembly after bad syntax.
*/
static void flush_eol(char **ss)
{
    while (**ss) {
        eatBlanks(ss);          // pass multiline comment after bad syntax

        if (isgraph(**ss))
            ++*ss;              // pass suspicious character

        if (**ss == '\n') {
            ++*ss;              // pass newline and stop
            break;
        }
    }                           // also stop at null
}

/*
    Advance past an arbitrary symbol, like "<>", if present.

    By "symbol" here we mean a short non-empty string of characters,
    probably without any numbers or letters. So this word "symbol" doesn't
    have the same meaning as, say, an ASCII symbol (which is not a string),
    nor does it mean an entry in the assembler's symbol table (which
    provides a means to look up registers by name).
*/
static int symbol(char **ss, char *sym)
{
    char *s = *ss;
    eatBlanks(&s);
    while (*sym && *s == *sym)
        ++s, ++sym;
    if (*sym) return 0;
    *ss = s;
    return 1;
}

/*
    This detects an identifier, returns its start position, and
    advances 's' just past its end.

    Identifiers start with a letter and may contain letters, digits,
    apostrophes, and underscores.
*/
static char *identifier(char **ss)
{
    char *s = *ss, *start;

    eatBlanks(&s);
    start = s;
    if (!isalpha(*s))
        return NULL;

    while (isalnum(*s) || *s == '\'' || *s == '_' || *s == '.')
        ++s;
    *ss = s;
    return start;
}

// Advance past a keyword if present.
static int keyword(char **ss, char *kw)
{
    char *s = *ss;  
    char *id = identifier(&s);
    int len;

    if (!id) return 0;                  // no word here
    len = strlen(kw);
    if (len != s - id) return 0;        // wrong length
    if (memcmp(id, kw, len)) return 0;  // not the word we seek

    *ss = s;
    return 1;
}

/*
    Find the end of a substring that could conceivably
    be an unsigned radix-64 number.
*/
char *delimit_number_segment(char *s)
{
    while (*s == '_' || *s && strchr(tetrasexagesimal, *s))
        ++s;
    return s;
}

/*
    Evaluate a substring as an unsigned long with some customizations.
    To simplify overflow logic, max_allowed may not exceed 2**36 - 1.
    To disable group separators, use one that does not fit in a char.
*/
static int multiradix_range_convert(
    char *from,             // start of substring
    char *to,               // byte after end of substring
    int radix,              // must be in range 1 through 64
    ulong min_allowed,      // valid number must be at least this much
    ulong max_allowed,      // valid number may not exceed this
    int group_sep,          // like a thousands separator; e.g., ',' or '_'
    ulong *result           // output if conversion succeeds
) {
    int digit, saw_digit = 0;
    char *s, *d;
    ulong n = 0;

    if (max_allowed > mask36) {
        ng("!multiradix_range_convert unsupported range");
    }

    for (s = from; s < to; ++s) {
        if (*s == group_sep)
            continue;       // ignore this symbol
        d = strchr(tetrasexagesimal, *s);
        if (!d) {
            // possible identifier, because attempted conversion of non-digit
            return 0;
        }
        saw_digit = 1;      // trying to make a number
        digit = d - tetrasexagesimal;
        if (digit >= radix) {
            // possible identifier, because digit not within radix
            return 0;
        }
        n = n * radix + digit;
        if (n > max_allowed) {
            // possible identifier, because number or radix too large
            return 0;
        }
    }

    if (!saw_digit) {
        note_parse_error(s, "number without any digits", NULL, NULL);
        return 0;
    }

    if (n < min_allowed) {
        note_parse_error(s, "number or radix too small", NULL, NULL);
        return 0;
    }

    *result = n;
    return 1;
}

/*
    Parse a number radix, which is one of:

    `[digits]       where [digits] is an integer from 1 through 64 in decimal
    `u              1 (unary)
    `b              2 (binary)
    `o              8 (octal)
    `d              10 (decimal)
    `h              16 (hexadecimal)
    `t              64 (tetrasexagesimal or tribble)

    If no radix appears, parsing does not advance and the radix is 10.

    For first-letter consistency `x is not a supported abbreviation for `16.

    This function always succeeds with a result between 1 and 64,
    but it may generate parse errors and leave an unconverted suffix that
    will mess stuff up later.
*/
static int scan_number_radix(char **ss)
{
    char *s = *ss, *to;
    int radix = 10;
    ulong r;

    if (*s++ != '`')
        goto done;

    to = delimit_number_segment(s);
    if (to == s)
        goto done;

    if (to == s + 1) {
        switch (*s) {
            case 'u': radix =  1; goto ok;
            case 'b': radix =  2; goto ok;
            case 'o': radix =  8; goto ok;
            case 'd': radix = 10; goto ok;
            case 'h': radix = 16; goto ok;
            case 't': radix = 64; goto ok;

            default:
                goto done;
        }

ok:
        ++s;
        *ss = s;
        goto done;
    }
        
    ifnot (multiradix_range_convert(s, to, 10, 1, 64, 999, &r))
        goto done;

    radix = r;
    *ss = s;

done:
    return radix;
}

/*
    Parse an integer within the 36-bit signed and unsigned range,
    which is -34_359_738_368 through +68_719_476_735. The format is
    in section A.3 of the dissertation.
*/
static int scan_number(char **ss, long *num)
{
    char *s = *ss;
    char *from, *to;

    int neg = 0, radix;
    ulong n, limit = 68719476735;

    // Accommodate a single optional + or -.
    if (*s == '+')
        ++s;
    else if (*s == '-')
        ++s, neg = 1, limit = 34359738368;

    // Look for possible digits.
    to = delimit_number_segment(from = s);
    if (to == from) {
        return 0;
    }

    // Look for a backtick and number radix.
    s = to;
    radix = scan_number_radix(&s);

    // Convert the possible digits according to the radix.

    ifnot (multiradix_range_convert(from, to, radix, 0, limit, '_', &n)) {
        return 0;
    }

    *ss = s;
    if (num)
        *num = neg ? -n : n;
    return 1;
}

// Look for a number and report it back as a constant-valued register.
// The lexing portion is factored out for reuse by reg_to_sym().
static int number(char **ss, struct reg_name *r)
{
    char *s = *ss;

    eatBlanks(&s);
    r->vi = s;
    if (!scan_number(&s, NULL))
        return 0;
    r->de = s;
    r->cast = 0;
    *ss = s;
    return 1;
}

// Look for "label :" at the beginning of a line.
static int label(char **ss, int pass)
{
    char *s = *ss, *id, *end;
    struct sym *l;

    ifnot (id = identifier(&s))
        return 0;
    end = s;
    ifnot (symbol(&s, ":"))
        return 0;

    // Syntax is good. Note line parsing may continue after the colon.
    *ss = s;
    if (pass != 1) return 1;

    // In pass 1, label goes in symbol table.
    l = find_symbol(id, end, 1, 0);
    if (l->what)
        note_parse_error(s, "label defined more than once", id, end);
    else {
        l->what = 'l';
        l->number = X->cursor;
    }
    return 1;
}

// A cast may be either "<unsigned>" or "<signed>",
// or abbreviated <u.> or <s.>. Returns 'u', 's', or 0 if no cast.
static int cast(char **ss)
{
    int c = 0;
    char *s = *ss;

    do {
        ifnot (symbol(&s, "<"))
            break;

        if (keyword(&s, "signed") || keyword(&s, "s."))
            c = 's';
        else if (keyword(&s, "unsigned") || keyword(&s, "u."))
            c = 'u';
        else
            break;

        ifnot (symbol(&s, ">"))
            break;
        *ss = s;

    } while (0);

    return c;
}

// Lvalues are identifiers, preceded optionally by a cast.
static int lvalue(char **ss, struct reg_name *r)
{
    char *s = *ss;

    r->cast = cast(&s);
    ifnot (r->vi = identifier(&s))
        return 0;

    r->de = *ss = s;
    return 1;
}

// Registers can be numbers or lvalues.
// A number is a constant register containing that number.
// An lvalue is something that is allowed left of an '=' sign.
static int parse_register(char **ss, struct reg_name *r)
{
//read_print_eval();
//    return lvalue(ss, r) || number(ss, r);
    return number(ss, r) || lvalue(ss, r);
}
