/*
    syms.c
    Symbol table support for assembler
*/

/*
    XXX General note on numeric constants

    At present, numeric constants live in registers that have
    numeric names. For example, a register named "53" will be
    pre-loaded with the value 53 during VM initialization.
    Two big transformations are likely to change this:

    1)  IMM, IMMU, IMMB (the load immediate instructions) will
        get constants into the system. The assembler may also
        automatically generate these instructions.

    2)  The toolchain (assembler and not-yet-existing linker)
        will probably do register optimization; that is, reuse
        registers where permissible. That process may involve
        some magic to limit the number of automatically generated
        IMM, IMMU, and IMMB.
*/

/*
    This is a parsed register name. It does not contain any buffer
    for the text of this name, but instead points into the source
    code being assembled. So this is a temporary value coming
    from the parser/lexer, as opposed to a symbol table entry,
    which does more and contains a copy of the name.

    The parsed output is the interval [vi, de), and an optional
    'cast' which is 0, 's', or 'u' indicating one of
    { no cast, <signed>, <unsigned> } prior to the name.

    If 'de' is NULL, then we consider de == vi + strlen(vi).
*/
struct reg_name {
    char *vi;
    char *de;
    int cast;
};

/*
    The symbol table is an unbalanced binary tree indexed by name.
*/
struct sym {
    int what;                       // 's'igned reg, 'u'nsigned reg, 'l'abel
    int number;                     // register # or label address
    long init_val;                  // register value when program starts
    struct sym *left, *right;       // children
    char name[1];
};

static int scan_number(char **ss, long *num);       // in lex.c

/*
    Compare two names that might not be \0 terminated.
    Returns < 0, == 0, > 0 based on l0 < r0, l0 == r0, l0 > r0.
*/
static int namecmp(char *l0, char *l1, char *r0, char *r1)
{
    int len_l = l1 ? l1 - l0 : strlen(l0);
    int len_r = r1 ? r1 - r0 : strlen(r0);
    int len_common = len_l < len_r ? len_l : len_r;
    int c;

    if (c = memcmp(l0, r0, len_common))
        return c;

    return len_l - len_r;
}

/*
    Find or create symbol in table.
    Its name may use 'end_name' in place of \0 termination.
    If not found and 'create' is true, make a place for it.
    If 'exclusive' is true, confirm pre-existence is opposite of 'create'.
*/
static struct sym *find_symbol
    (char *name, char *end_name, int create, int exclusive)
{
    int c, len;
    struct sym **s = &X->sym_root;

    // 's' gets where the pointer to this symbol belongs,
    // whether or not that pointer exists.
    while (*s) {
        c = namecmp(name, end_name, (*s)->name, NULL);
        if (!c)
            break;
        s = c <= 0 ? &(*s)->left : &(*s)->right;
    }

    if (*s) {                           // symbol was found
        if (create && exclusive)
            ng("!symbol at head of '%.20s' was not supposed to exist", name);
        return *s;
    }

    // symbol was not found
    if (!create) {
        if (exclusive)
            ng("!symbol at head of '%.20s' was supposed to exist", name);
        return NULL;
    }

    // symbol needs created
    len = end_name ? end_name - name : strlen(name);
    *s = trackmem(NULL, sizeof(struct sym) + len);
    memcpy((*s)->name, name, len);
    (*s)->name[len] = '\0';

    return *s;
}

// Erase symbol table. Pass &X->sym_root to call this.
static void clear_symbol_table(struct sym **root)
{
    if (!*root) return;
    clear_symbol_table(&(*root)->left);
    clear_symbol_table(&(*root)->right);
    trackmem(*root, 0);
    *root = NULL;
}

/*
    Assign register numbers (range can be 0-511) to all
    symbol table entries. No register optimization is done.
    Pass X->sym_root to call this.
*/
static void number_regs(struct sym *t)
{
    if (!t) return;
    number_regs(t->left);
    if (t->what == 's' || t->what == 'u')
        t->number = X->n_regs++;
    number_regs(t->right);
}

/*
    Preset registers to ZERO for execution.
    We no longer preset register constants to their initial values,
    because the assembler inserts runtime code for that.

    During execution, registers containing numeric constants 
    will not be modified. Pass X->sym_root to call this.
*/
static void preset_regs(struct sym *t)
{
    if (!t) return;
    preset_regs(t->left);
    if (t->what == 's' || t->what == 'u') {
        if (t->number < 0 || t->number >= X->n_regs)
            ng("!preset_regs logic fail");
        X->regs[t->number] = 0;         // no longer use t->init_val;
    }
    preset_regs(t->right);
}

/*
    Show registers that are no longer at their
    initial values after execution.
*/
static void show_changed_regs(struct sym *t)
{
    long n;

    if (!t) return;
    show_changed_regs(t->left);

    if (t->what == 's' || t->what == 'u') {
        n = X->regs[t->number];
        if (n != t->init_val) {
            if (t->what == 's' && (n & pole))
                n |= comp36;            // negative value
            printf("%12li %s\n", n, t->name);
        }
    }

    show_changed_regs(t->right);
}

/*
    Show a particular register's name and value as a diagnostic.
    This does a linear-time reverse lookup, so it isn't fast.
    pass X->sym_root to call this.
*/
static void trace_reg(struct sym *t, int regnum)
{
    if (!t) return;
    trace_reg(t->left, regnum);

    if (t->number == regnum) {
        if (t->what == 's')
            printf("%s = %li\n", t->name, X->regs[regnum]);
        if (t->what == 'u')
            printf("%s = %lu\n", t->name, X->regs[regnum]);
    }

    trace_reg(t->right, regnum);
}

// Print out the symbol table. Pass X->sym_root to call this.
static void show_symbol_table(struct sym *t)
{
    char *what;

    if (!t) return;
    switch (t->what) {
        case 's': what = "s."; break;
        case 'u': what = "u."; break;
        case 'l': what = "l."; break;
        default:  what = "? "; break;
    }

    show_symbol_table(t->left);
    if (isdigit(*t->name))
        printf(" %4i  %s  %li\n", t->number, what, t->init_val);
    else
        printf(" %4i  %s  %s\n", t->number, what, t->name);
    show_symbol_table(t->right);
}

// Convert register name to symbol table entry and computed signage.
static struct sym *reg_to_sym(struct reg_name *r, char *place, int *signage)
{
    struct sym *s;
    char num_buf[15], *nump;
    long n, n_unsigned;
    char *de = r->de ? r->de : r->vi + strlen(r->vi);

    // Numeric constants can be written multiple ways, but we want
    // only their canonical unsigned decimal symbols in the table.
    nump = r->vi;
    if (scan_number(&nump, &n) && nump == de) {
        // Chop to 36 bits for its name in the symbol table.
        // A symbol is a numeric constant iff isdigit(first_char_of_name).
        n_unsigned = n & mask36;
        sprintf(num_buf, "%011li", n_unsigned);
        s = find_symbol(num_buf, NULL, 1, 0);

        // Put the numeric contents in the symbol table.
        if (!s->what) {
            s->init_val = n_unsigned;
            s->what = 'u';
        }

        // If the value is in fact negative (the constant started with
        // a minus sign), pass this back as a signed register. The symbol
        // table will SAY it's unsigned, but encode_instruction() will
        // generate code for the real signage. This kludge simplifies
        // our symbol table and lets signed and unsigned values with
        // the same 36-bit representation share an entry.
        *signage = n < 0 ? 's' : 'u';
        return s;
    }

    // This is an ordinary register name, and a possible cast.
    // Find the symbol.
    s = find_symbol(r->vi, de, 1, 0);

    // Pass 1 should have marked the symbol's signedness.
    if (s->what != 's' && s->what != 'u') {
        note_parse_error(place, "register must be declared signed or unsigned",
                s->name, NULL);
    }

    // <signed> and <unsigned> casts can override,
    // but can't substitute for, register declarations.
    // All registers must be declared signed or unsigned.
    *signage = r->cast ? r->cast : s->what;

    return s;
}

// Mini-disassembler for listings
void asm_disassemble(char *prefix, long code, char *suffix)
{
    int opc = code >> 27 & 511;
    int d = code >> 18 & 511;
    int l = code >> 9 & 511;
    int r = code & 511;

    printf("%s%7s %i %i %i%s", prefix, opcode_name(opc), d, l, r,
            suffix);
}

/*
    Append an instruction to code memory. Suppress if 'code' is -1.
    Also suppress if we're in pass 1, but count anyway.
*/
static void emit_code(struct context *x, long code)
{
    int opc = code >> 27 & 511;

    if (code == -1)
        return;

    if (x->code) {          // pass 2?
        if (x->code[x->cursor] >> 27 != oc_NOLIST)
            ng("!assembler writing to location %i twice", x->cursor);

        x->code[x->cursor] = code;
        if (x->list && opc != oc_NOLIST) {
            printf("%4i  ", x->cursor);
            asm_disassemble("", code, "\n");
        }
    }
    ++x->cursor;            // Keep count during both passes.
}

static void imm_pre_work(struct context *x, struct sym *symbol, int zero_regno)
{
    if (symbol->what != 'u' && symbol->what != 's')
        return;

    int dest = symbol->number;
    if (dest < 0 || dest > 511)
        ng("!too many registers: register %i in imm_pre_recurse", dest);

    // Only operate on numeric constant registers. Exclude the zero register.
    if (dest == zero_regno || !isdigit(*symbol->name))
        return;

    int imm_l = symbol->init_val & mask18;
    int imm_h = symbol->init_val >> 18 & mask18;
    long instr;

    // case 1: negative and within [-262143, -1]
    if (symbol->init_val >= -262143 && symbol->init_val < 0) {
        instr = (long) oc_IMN << 27 | imm_l;
    }

    // case 2: [0, 262143]
    else if (imm_h == 0) {
        instr = (long) oc_IMP << 27 | imm_l;
    }

    // case 3: positive multiples of 262144
    else if (imm_l == 0) {
        instr = (long) oc_IMH << 27 | imm_h;
    }

    // case 4: positive multiples of 262145,
    // meaning the high and low 18 bits are identical
    else if (imm_h == imm_l) {
        instr = (long) oc_IMB << 27 | imm_l;
    }

    else
        goto hard_case;

    // This is still the easy case. Include the destination register.
    instr |= dest << 18;

    // Write the instruction.
    emit_code(x, instr);
    goto either_case;

hard_case:
    // If we get here, the 36-bit bit immediate value has different
    // and non-trivial upper and lower halves, meaning we can't
    // fit the immediate value, destination register, and opcode into
    // a single 36-bit instruction. We'll need three instructions.
    emit_code(x, (long) oc_IMH << 27 | dest << 18 | imm_h);
    emit_code(x, (long) oc_IMP << 27 | zero_regno << 18 | imm_l);
    emit_code(x, (long) oc_OR << 27 | dest << 18 | dest << 9 | zero_regno);

either_case:;
}

static void imm_pre_recurse(struct context *x, struct sym *symbol, int zero_regno)
{
    // Inorder recursion
    if (!symbol)
        return;
    imm_pre_recurse(x, symbol->left, zero_regno);
    imm_pre_work(x, symbol, zero_regno);
    imm_pre_recurse(x, symbol->right, zero_regno);
}

/*
    Given a context with a numbered symbol table and allocated code memory,
    generate a program preamble that loads immediate values into registers.
    This will start at offset 0 and hopefully stop before preamble_size.
*/
static void immediate_preamble(struct context *x)
{
    // The "0" immediate value will be created last. Its register
    // is used for temporary storage by 3-instruction immediate values.
    struct reg_name zeroReg = { "0" };
    int ignored = 0;
    int zero_regno = reg_to_sym(&zeroReg, "nowhere", &ignored)->number;

    // Write the immediate value instructions for the non-zero registers.
    x->cursor = 0;
    imm_pre_recurse(x, x->sym_root, zero_regno);

    // I held back the zero register as a temporary register for
    // imm_pre_recurse. It's now available for its normal function.
    emit_code(x, (long) oc_IMP << 27 | zero_regno << 18);

    // Give us a jump over any unused preamble space.
    if (x->cursor < x->preamble_size)
        emit_code(x, (long) oc_J << 27 | x->preamble_size);
        // May now be equal, which is okay,

    // but greater is an overrun.
    else if (x->cursor > x->preamble_size)
        ng("!preamble of length %i overran space of %i",
                x->cursor, x->preamble_size);

    // Here's where we'll be writing code.
    x->cursor = x->preamble_size;
}
